/******************************************************************************
 CButton.c

							The Button Class
							
	A standard Mac push-button control. It can issue a command to its
	supervisor after it is clicked.
			
	SUPERCLASS = CControl
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	TCL 1.1 CHANGES
	[
		- implemented alternate initialization method INewButton and
		  IButtonX method.
		- implemented GetClickCmd, SetDefault methods
	]
 ******************************************************************************/

#include "Global.h"
#include "CButton.h"
#include "Commands.h"
#include "CError.h"
#include "CPaneBorder.h"

#define 	kDefaultBtnRounding	16		/* rounding diameter for default button border */

/**** C O N S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IButton

		Initialize a Button object
 ******************************************************************************/

void	CButton::IButton(
	short			CNTLid,				/* CNTL resource ID					*/	
	CView			*anEnclosure,		/* Enclosing view					*/
	CBureaucrat		*aSupervisor)		/* Boss in chain of command			*/
{
    long			enclOriginH, enclOriginV;
    CNTLtemplateHndl cntl;  
    short			savedVis;        
	
	CView::IView(anEnclosure, aSupervisor);
	
	itsBorder = NULL;

	macPort = itsEnclosure->GetMacPort();
	SetPort( macPort);
	SetOrigin( 0, 0);
	ForceNextPrepare();
	
	// get the CNTL resource and clear the visible flag so we create the
	// control invisible
	
	cntl = (CNTLtemplateHndl) GetResource( 'CNTL', CNTLid);
	FailNILRes( cntl);
	savedVis = (**cntl).visible;
	(**cntl).visible = 0;
	
	macControl = GetNewControl(CNTLid, macPort);
	FailNIL( macControl);
			
	visible = FALSE;
	active = TRUE;
	autoRefresh = TRUE;
	hSizing = vSizing = sizFIXEDSTICKY;
    printClip = clipFRAME;
    printing = FALSE;
    

	hEncl = (**macControl).contrlRect.left;
	vEncl = (**macControl).contrlRect.top;
	width = (**macControl).contrlRect.right - hEncl;
	height = (**macControl).contrlRect.bottom - vEncl;
	   
    anEnclosure->GetOrigin(&enclOriginH, &enclOriginV);
    hEncl += enclOriginH;
    vEncl += enclOriginV;
    
	CalcFrame();						/* Set up Frame coordinates			*/
	CalcAperture();						/* Determine visible portion		*/
	
	itsEnvironment = NULL;
	itsEnclosure->AddSubview(this);		/* This is a subview of enclosure	*/
	
	IButtonX();
	
										/* if CNTL specified visible control */
										/* then show it now					*/
	if (savedVis) 
	{
		Show();	
		cntl = (CNTLtemplateHndl) GetResource( 'CNTL', CNTLid);
		FailNILRes( cntl);
		(**cntl).visible = savedVis;
	}
}
	
/******************************************************************************
 INewButton
 
 	Initialize a button from dynamic parameters, rather than from a
 	CNTL resource.
******************************************************************************/

void CButton::INewButton( short aWidth, short aHeight, short aHEncl, short aVEncl, 
					StringPtr title, Boolean fVisible, short procID,
					CView *anEnclosure, CBureaucrat	*aSupervisor)
{
	Rect r;	
	
	CPane::IPane( anEnclosure, aSupervisor, aWidth, aHeight,
				aHEncl, aVEncl, sizFIXEDSTICKY, sizFIXEDSTICKY);

	FrameToWindR( &frame, &r);
	SetPort( macPort);
	SetOrigin( 0, 0);
	ForceNextPrepare();
	
	macControl = NewControl( macPort, &r, title,
					fVisible, 0, 0, 1, procID, 0);
	FailNIL( macControl);
	visible = fVisible;
	
	IButtonX();

}	/* CButton::INewButton */

/******************************************************************************
 IButtonX
******************************************************************************/

void CButton::IButtonX( void)
{
	clickCmd = cmdNull;
	wantsClicks = TRUE;
}

/**** C L I C K  H A N D L I N G   M E T H O D S ****/


/******************************************************************************
 DoGoodClick {OVERRIDE}

		A Button responds to a good click (the mouse being pressed
		and released within the button) by passing a particular command
		to its supervisor.
 ******************************************************************************/

void	CButton::DoGoodClick(
	short		whichPart)				/* Will always be inButton			*/
{
	if (clickCmd != cmdNull)			/* Issue the appropriate command	*/
		itsSupervisor->DoCommand(clickCmd);
}


/******************************************************************************
 SetClickCmd

		Specify the command which is sent to a Button's supervisor
		after a confirmed click
 ******************************************************************************/

void	CButton::SetClickCmd(
	long		aClickCmd)				/* Specified command number			*/
{
	clickCmd = aClickCmd;				/* Set instance variable			*/
}

/******************************************************************************
 GetClickCmd
 
 	Return the command which is sent to a Button's supervisor after a
 	confirmed click
******************************************************************************/
long CButton::GetClickCmd( void)
{
	return clickCmd;
}

/******************************************************************************
 SetDefault
 
 	Specify whether or not the button is the default button. The default
 	button has a 3 pixel thick rounded rectangle as its border.
******************************************************************************/

void CButton::SetDefault( Boolean fDefault)
{
	CPaneBorder	*border = NULL;
	Rect	margin;
	
	if (fDefault && !itsBorder)
	{
		border = new( CPaneBorder);
		border->IPaneBorder( kBorderRoundRect);
		border->SetRounding( kDefaultBtnRounding, kDefaultBtnRounding);
		border->SetPenSize( 3, 3);
		SetRect( &margin, -1, -1, 1, 1);
		border->SetMargin( &margin);
		SetBorder( border);
		
	}
	else if (!fDefault && itsBorder)
	{
		SetBorder( NULL);
	}
}

/******************************************************************************
 SimulateClick
 
 	Simulate a click in the button, usually requested in response to
 	a keyboard shortcut
******************************************************************************/

void CButton::SimulateClick( void)
{
	#define kFlashDelay		8	/*  number of ticks to hilite button */
	long	ticks;
	
	if (IsActive())				/* TCL 1.1.2 DLP 10/28/91			*/
	{
		Prepare();
		HiliteControl( macControl,1);
		Delay( kFlashDelay, &ticks);
		HiliteControl( macControl,0);
		DoGoodClick( inButton);
	}
}