/******************************************************************************
 CPopupMenu.c

		A PopupMenu is an object that manages a menu and can pop it up
		on request. Its menu is known to the Bartender and can send
		command messages.
		
	SUPERCLASS = CBureacrat
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CPopupMenu.h"
#include "CBartender.h"
#include "Commands.h"
#include "Global.h"
#include "TBUtilities.h"

extern CBartender	*gBartender;

/******************************************************************************
 IPopupMenu
 
 	Initialize a PopupMenu object. aMenuID specifies the resource ID of
 	the 'MENU' resource. The fAutoSelect flag specifies whether the
 	menu should automatically check and uncheck menu items when selected.
 	The fMultiSelect flag specifies if multiple menu items may be selected.
******************************************************************************/

void CPopupMenu::IPopupMenu( short aMenuID, CBureaucrat *aSupervisor,
					Boolean fAutoSelect, Boolean fMultiSelect)
{	
	CBureaucrat::IBureaucrat( aSupervisor);
		
	menuID = aMenuID;
	multiSelect = fMultiSelect;
	autoSelect = fAutoSelect;


	/*
	 *	Has it previously been registered by bartender?
	 *	We only want to register it with the bartender once.
	 */
	macMenu = gBartender->FindMacMenu( menuID);
	
	if (!macMenu)	/* no it hasn't */
	{
				/* add the menu, but don't install into the menu list */
		gBartender->AddMenu( menuID, FALSE, 0);
		macMenu = gBartender->FindMacMenu( menuID);
		
		/* it better have been found now */
		FailNILRes( macMenu);
	}
	
	/* NOTE: we never call DisposeMenu( macMenu) because the Bartender		*/
	/* has a copy of it AND because there is no way to tell if this			*/ 
	/* menu handle is shared with other objects. We could allocate a		*/
	/* private copy, but then the menu would be hidden from the Bartender	*/ 
	
	firstSelection = GetCheckedItem();	/* TCL 1.1.1 DLP 9/18/91 */
	lastCmd = cmdNull;
	
	PreparePopup();
		
}	/* CPopupMenu::IPopupMenu */

/******************************************************************************
 PreparePopup
 
 	Set the menu options and initially selected item, if any. Override for
 	additional preparation.
******************************************************************************/

void CPopupMenu::PreparePopup( void)
{
	
	gBartender->SetDimOption( menuID, dimNONE);	/* Never dim menu items		*/
	
	/* When autoSelect is TRUE, menu items are automatically checked	*/
	/* when selected and selections are remembered. If it is FALSE, then */
	/* menu items should be checked via the UpdateMenus mechanism		*/
	
	gBartender->SetUnchecking( menuID, !autoSelect);
	
	if (autoSelect && GetNumItems() > 0)
		SelectItem( 1, pmForceOn);

}	/* CPopupMenu::PreparePopup */


/******************************************************************************
 AppendResNames
 
 	Append the names of all available resources of a given type.
******************************************************************************/

void CPopupMenu::AppendResNames( ResType aResType)
{
	AddResMenu( macMenu, aResType);

}	/* CPopupMenu::AppendResNames */

/******************************************************************************
 CheckCmdKey
 
 	Check if the keypress information corresponds to a command key
 	for the menu. Returns the command and item number if any,
 	othwise returns cmdNull and 0.
******************************************************************************/

void CPopupMenu::CheckCmdKey( char theChar, Byte keyCode, EventRecord *macEvent,
					long *theCmd, short *theItem)
{
	short currItem, numItems;
	short cmdChar;
	unsigned char buf[10];
	
	*theCmd = cmdNull;
	*theItem = 0;
	if ((macEvent->modifiers & cmdKey) && (theChar != hMenuCmd))
	{
		buf[0] = 1; buf[1] = theChar;
		UprString( buf, FALSE);
		theChar = buf[1];
		
		numItems = GetNumItems();
		for (currItem = 1; currItem <= numItems; currItem++)
		{
			GetItemCmd(macMenu, currItem, &cmdChar);
			if (theChar == cmdChar)
			{
				*theCmd = gBartender->FindCmdNumber( menuID, currItem);
				*theItem = currItem;
				break;
			}
		}
	}
	
}	/* CPopupMenu::CheckCmdKey */

/******************************************************************************
 GetMacMenu
 
 	Return the MenuHandle
******************************************************************************/

MenuHandle CPopupMenu::GetMacMenu( void)
{
	return macMenu;

}	/* CPopupMenu::GetMacMenu */

/******************************************************************************
 GetNumItems
 
 	Return the number of items in the menu.
******************************************************************************/

short CPopupMenu::GetNumItems( void)
{
	return CountMItems( macMenu);

}	/* CPopupMenu::GetNumItems */

/******************************************************************************
 GetTitle
 
 	Return the menu's title
******************************************************************************/

void CPopupMenu::GetTitle( StringPtr aTitle)
{
	CopyPString( (StringPtr) (**macMenu).menuData, aTitle);

}	/* CPopupMenu::GetTitle */

/******************************************************************************
 GetWidth
 
 	Return the menu's width when popped up.
******************************************************************************/

short CPopupMenu::GetWidth( void)
{
	CalcMenuSize( macMenu);
	return (**macMenu).menuWidth;

}	/* CPopupMenu::GetWidth */

/******************************************************************************
 SelectItem
 
 	Applys the action requested in actionType to the item. The valid
 	actions are:
 		pmForceOn	- always check the item
 		pmForceOff	- always uncheck the item
 		pmToggle	- check if unchecked, uncheck if checked.
******************************************************************************/

void CPopupMenu::SelectItem( short itemNum,tPMSelectAction actionType)
{
	Rect	r;
	MenuHandle menu = macMenu;
	
	if ((itemNum <= 0)||(itemNum > GetNumItems())) return;
	
	if (multiSelect)
	{
		if (actionType == pmToggle) 
			CheckMenuItem( itemNum, !ItemIsChecked( itemNum));
		else CheckMenuItem( itemNum, actionType == pmForceOn);
		
		firstSelection = GetCheckedItem(); 
	}
	else	
	{
		/*
		 *	For single selection menus, we must modify the requested action to
		 *	ensure that there is always a selected item.
		 *	Note that we must ignore pmForceOff completely for this reason, 
		 *	and pmForceOn accomplishes turning off the old selection and
		 *	turning on the new one.
		 */
		if ((actionType == pmForceOff)||(itemNum == firstSelection)) return;
		
		CheckMenuItem(  firstSelection, FALSE);
		CheckMenuItem( itemNum, TRUE);
		firstSelection = itemNum;
	}

}	/* CPopupMenu::SelectItem */


/******************************************************************************
 SelectItemName
 
 	Select an item by its item text. Does nothing if the item is not found 
 	in the menu.
 	
******************************************************************************/

void  CPopupMenu::SelectItemName( StringPtr name, tPMSelectAction actionType)
{	
	short	i, itemCount = GetNumItems();
	Str255	itemString;

	for (i = 1; i <= itemCount; i++)
	{
		GetItem( macMenu, i, itemString);
		if (EqualString( name, itemString, FALSE, FALSE))
		{
			SelectItem( i, actionType);
			break;
		}
	}
}	/* CPopupMenu::SelectItemName */

/******************************************************************************
 ItemIsChecked

	Return TRUE if item in menu has checkmark.
******************************************************************************/

Boolean CPopupMenu::ItemIsChecked(short item)
{
	short	markChar;
	
	GetItemMark( macMenu, item, &markChar);
	return (markChar == checkMark);

}	/* CPopupMenu::ItemIsChecked */

/******************************************************************************
 PopupSelect
 
 	Pop up the menu and return the resulting command. If the resulting command
 	is not cmdNull, then the command message is also sent to itsSupervisor.
 	Where specifies the location, in global coordinates, for the menu
 	to pop up. aCmd specifies the command the should be selected when the
 	menu pops up. cmdNull may be passed for a suitable default command
 	to be calculated.
******************************************************************************/

long CPopupMenu::PopupSelect( Point where, long aCmd)
{
	long		menuResult;
	short		selectedItem, mID;
	MenuHandle	mHandle;
	
	gBartender->UpdateAllMenus();

	/* The caller can specify the initially selected item by passing	*/
	/* in its command number. If none is specified, then one is 		*/
	/* calculated. If autoSelection is on, then the first selected item	*/
	/* is used, otherwise the last item chosen is used. Finally, if all	*/
	/* else fails, the first menu item is used.							*/
	
	if (aCmd != cmdNull)
		gBartender->FindMenuItem( aCmd, &mID, &mHandle, &selectedItem);
	else
	{
		if (autoSelect)
		{
			selectedItem = firstSelection;
			aCmd = gBartender->FindCmdNumber( menuID, selectedItem);
		}
		else
		{
			aCmd = lastCmd;
			gBartender->FindMenuItem( aCmd, &mID, &mHandle, &selectedItem);
		}
	}
	if (selectedItem == 0)
		aCmd = gBartender->FindCmdNumber( menuID, 1);
		
	InsertMenu( macMenu, hierMenu);	/* temporarily insert into menu list	*/
	CalcMenuSize( macMenu);
	menuResult = PopUpMenuSelect( macMenu, where.v, where.h, selectedItem);
	DeleteMenu( menuID);			/* remove from menu list				*/
	
	selectedItem = LoShort( menuResult);
	mID = HiShort(menuResult);  /* could be submenu */

	aCmd = gBartender->FindCmdNumber( mID, selectedItem);
	
	/*	FindCmdNumber will return a positive integer if 		*/
	/*	there is a command for this item or a negative number	*/
	/*	if the menu was registered but the item has no command.	*/
	/*	If the menu was not registered it will return cmdNull.	*/

	if (aCmd != cmdNull)
		itsSupervisor->DoCommand( aCmd);
		
	/* notify dependents and supervisor of the new selection	*/
	
	BroadcastChange( popupMenuNewSelection, &selectedItem);
		
	lastCmd = aCmd;
	
	if (autoSelect && (mID == menuID))
		SelectItem( selectedItem, pmToggle);
		
	/* return the last resulting command */
	
	return aCmd;
	
}	/* CPopupMenu::PopupSelect */

/******************************************************************************
 GetCheckedItem
 
 	Return item number if any item in menu has a checkmark, else 
 	returns 0.
******************************************************************************/

short CPopupMenu::GetCheckedItem( void)
{	short	i, count;
	MenuHandle menu = macMenu;

	if (!menu) return 0;
	
	count = CountMItems( menu);
	for (i = 1; i <= count; i++)
	{
		if (ItemIsChecked( i)) return i;
	}
	return 0;

}	/* CPopupMenu::GetCheckedItem */

/******************************************************************************
 GetItemString
 
 	Return the text of the specified item
******************************************************************************/

void CPopupMenu::GetItemString( short item, StringPtr string)
{
	GetItem( macMenu, item, string);

}	/* CPopupMenu::GetItemString */


/******************************************************************************
 GetCurrItemString
 
 	Return text of the first checked item if there is one.
******************************************************************************/

void CPopupMenu::GetCurrItemString( StringPtr string)
{
	short	checkedItem = GetCheckedItem();
	
	if (checkedItem)
		GetItemString( checkedItem, string);
	else
		string[0] = 0;

}	/* CPopupMenu::GetCurrItemString */

/******************************************************************************
 GetItemCommand
 
 	Return the command number of an item
******************************************************************************/

long CPopupMenu::GetItemCommand( short item)
{
	return gBartender->FindCmdNumber( menuID, item);
}

/******************************************************************************
 CheckMenuItem 
 
 	Private method to check a menu item. Clients use SelectItem
******************************************************************************/

void CPopupMenu::CheckMenuItem( short item, Boolean fCheckIt)
{
	short cmdChar;
		
	/*	make sure that item does not have hierarchical menu,	*/
	/*	otherwise clearing item mark would lose the submenu		*/
	
	GetItemCmd( macMenu, item, &cmdChar);
	if (cmdChar != hMenuCmd)
	{
		SetItemMark( macMenu, item, fCheckIt? checkMark: 0);
	}

}	/* CPopupMenu::CheckMenuItem */

