/******************************************************************************
 CPopupPane.c

		A PopupPane is a pane that owns a CPopupMenu and pops up the
		menu when clicked. This class does not draw. It is up to
		the subclasses to draw a visual representation of the menu
		and its selections. See CStdPopupPane and CArrowPopupPane
		for examples.
		
	SUPERCLASS = CPane
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CPopupPane.h"
#include "CPopupMenu.h"
#include "Commands.h"
#include "Global.h"

/******************************************************************************
 IPopupPane
 	
 	Initialize a CPopupPane. Creates a CPopupMenu object using the first three
 	parameters. The supervisor of the CPopupMenu is the CPopupPane.
******************************************************************************/

void CPopupPane::IPopupPane( short menuID, Boolean fAutoSelect, Boolean fMultiSelect,
			CView *anEnclosure, CBureaucrat *aSupervisor,
			short aWidth, short aHeight, short aHEncl, short aVEncl,
			SizingOption aHSizing, SizingOption aVSizing)
{
	CPane::IPane( anEnclosure, aSupervisor, aWidth, aHeight, aHEncl, aVEncl,
				aHSizing, aVSizing);
				
	SetWantsClicks( TRUE);
	MakePopupMenu( menuID, fAutoSelect, fMultiSelect);

}	/* CPopupPane::IPopupPane */

/******************************************************************************
 IViewTemp {OVERRIDE}
 
 	Initialize a PopupPane from a view template resource.
******************************************************************************/

void CPopupPane::IViewTemp( CView *anEnclosure, CBureaucrat *aSupervisor,
							Ptr	viewData)
{
	PopupTempP	template = (PopupTempP) viewData;
	
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	
	MakePopupMenu( template->menuID, template->fAutoSelect != 0,
					template->fMultiSelect != 0);

}	/* CPopupPane::IViewTemp */

/******************************************************************************
 MakePopupMenu
 
 	Creates the PopupMenu owned by the PopupPane. Override to create
 	a subclass of CPopupMenu.
******************************************************************************/

void CPopupPane::MakePopupMenu( short menuID, Boolean fAutoSelect,
					Boolean fMultiSelect)
{
	itsMenu = new( CPopupMenu);
	itsMenu->IPopupMenu( menuID, this, fAutoSelect, fMultiSelect);

}	/* CPopupPane::MakePopupMenu */

/******************************************************************************
 GetMenu
 
 	Return the CPopupMenu.
******************************************************************************/

CPopupMenu *CPopupPane::GetMenu( void)
{
	return itsMenu;

}	/* CPopupPane::GetMenu */

/******************************************************************************
 SetMenu
 
 	Sets the PopupMenu to a new CPopupMenu. Dispose of the previous one, 
 	if any.
******************************************************************************/

void CPopupPane::SetMenu( CPopupMenu *aMenu)
{
	ForgetObject( itsMenu);
	itsMenu = aMenu;

}	/* CPopupPane::SetMenu */

/******************************************************************************
 DoClick
 
 	Handle a click by popping up the menu. The CalcPopupPoint method is used 
 	to get the location and the CalcPopupCmd method is used to get the
 	initially selected command.
******************************************************************************/

void CPopupPane::DoClick( Point hitPt, short modifierKeys, long when)
{
	long ignoreCmd;
	
	if (itsMenu)
	{
		ignoreCmd = itsMenu->PopupSelect( CalcPopupPoint( hitPt, modifierKeys, when),
					CalcPopupCmd());
	}

}	/* CPopupPane::DoClick */

/******************************************************************************
 CalcPopupPoint
 
 	Return the location to popup the menu in global coordinates. This
 	method uses the click location.
******************************************************************************/

Point CPopupPane::CalcPopupPoint( Point hitPt, short modifierKeys, long when)
{
	
	LocalToGlobal( &hitPt);
	
	return hitPt;

}	/* CPopupPane::CalcPopupPoint */

/******************************************************************************
 CalcPopupCmd
 
 	Return the command to use as the initially selected command when
 	the menu pops up. This version returns cmdNull, which signals
 	to CPopupMenu to determine the appropriate item itself.
******************************************************************************/

long CPopupPane::CalcPopupCmd( void)
{
	return cmdNull;

}	/* CPopupPane::CalcPopupCmd */

/******************************************************************************
 NewMenuSelection
 
 	Called whenever a new selection is made in the menu. Override to
 	do something special whenever this happens, e.g. update some
 	text that reflects the current item.
******************************************************************************/

void CPopupPane::NewMenuSelection( short itemSelected)
{

}	/* CPopupPane::NewMenuSelection */

/******************************************************************************
 ProviderChanged {OVERRIDE}
 
 	A Provider has changed. The only provider this class has is
 	itsMenu. itsMenu notifies the popup pane when a new selection was made,
 	and the pane sends itself the NewMenuSelection message.
******************************************************************************/

void CPopupPane::ProviderChanged( CCollaborator *aProvider, long reason,
					void* info)
{
	if ((aProvider == itsMenu) && (reason == popupMenuNewSelection))
	{
		NewMenuSelection( *(short*) info);
	}
	else
		inherited::ProviderChanged( aProvider, reason, info);

}	/* CPopupPane::ProviderChanged */

/******************************************************************************
 Draw {OVERRIDE}
 
 	This class does not draw any representation of the menu or its selection,
 	but does provide the default "greyed-out" appearance when the menu
 	is disabled. Subclasses can call inherited::Draw to get this behavior.
******************************************************************************/

void CPopupPane::Draw( Rect *area)
{
	if (!IsActive())	/* TCL 1.1.2 DLP 12/16/91 */
	{
		Rect r = *area;
		
		PenPat( gray);
		PenMode(patBic);
		PaintRect( &r);
		PenNormal();
	}

}	/* CPopupPane::Draw */

/******************************************************************************
 Activate {OVERRIDE}
 
 	A PopupPane is being activated. Indicate that it wants clicks.
******************************************************************************/

void CPopupPane::Activate( void)
{
	inherited::Activate();
	SetWantsClicks( TRUE);

}	/* CPopupPane::Activate */

/******************************************************************************
 Deactivate {OVERRIDE}
 
 	A PopupPane is being deactivated. Indicate that it doesn't want clicks
 	when it is deactivated.
******************************************************************************/

void CPopupPane::Deactivate( void)
{
	inherited::Deactivate();
	SetWantsClicks( FALSE);

}	/* CPopupPane::Deactivate */

/******************************************************************************
 Dispose {OVERRIDE}
 
 	Release its CPopupMenu and everything else.
******************************************************************************/

void CPopupPane::Dispose( void)
{
	ForgetObject( itsMenu);
	inherited::Dispose();

}	/* CPopupPane::Dispose */