/******************************************************************************
 CStdPopupPane.c

		A subclass of CPopupMenu that provides the Apple Human Interface
		Guidelines look and feel for a popup menu.
		
	SUPERCLASS = CPopupMenuPane
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CStdPopupPane.h"
#include "CPaneBorder.h"
#include "CPopupMenu.h"
#include "TBUtilities.h"
#include "TCLUtilities.h"
#include "Constants.h"
#include "Commands.h"
#include "CBartender.h"
#include "Global.h"
#include <Script.h>
#include "OSChecks.h"	/* TCL 1.1.1 DLP 9/25/91 */
									
#define __INSTALL_MENU_HOOK__		// if defined the menu  defproc will be patched
									// to provide the standard popup menu width
									

static pascal void MenuProc( short message, MenuHandle menu, Rect *menuRect,
					Point hitPt, short *whichItem);

extern CBartender	*gBartender;

/******************************************************************************
 IStdPopupPane
 
 	Create a standard popup menu pane. The constant kAutoSize may be passed
 	for both aWidth and aHeight to request that the popup dimensions
 	be calculated to the best size.
******************************************************************************/

void CStdPopupPane::IStdPopupPane( short menuID, CView *anEnclosure, 
			CBureaucrat *aSupervisor, short aWidth, short aHeight, 
			short aHEncl, short aVEncl)
{
	Boolean fFirstInstall;
	
		// We need to know if this is the first time we're installing
		// this menu so we install the menu hook only once.
		
	fFirstInstall = gBartender->FindMacMenu( menuID) == NULL;
	
	CPopupPane::IPopupPane( menuID, TRUE, FALSE, anEnclosure, aSupervisor,
			aWidth, aHeight, aHEncl, aVEncl,
			sizFIXEDSTICKY, sizFIXEDSTICKY);
	IStdPopupPaneX( fFirstInstall);
				
}	/* CStdPopupPane::IStdPopupPane */

/******************************************************************************
 IViewTemp {OVERRIDE}
 
 	Create a standard popup menu pane from a view template
******************************************************************************/

void CStdPopupPane::IViewTemp( CView *anEnclosure, CBureaucrat *aSupervisor,
						Ptr	viewData)
{
	StdPopupTemp	*template = (StdPopupTemp*) viewData;
	Boolean fFirstInstall;

	fFirstInstall = gBartender->FindMacMenu( template->popupTemp.menuID) == NULL;

	if (template->hAutosizing)
		template->popupTemp.paneTemp.width = kAutoSize;
	if (template->vAutosizing)
		template->popupTemp.paneTemp.height = kAutoSize;
		
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	IStdPopupPaneX( fFirstInstall);

}	/* CStdPopupPane::IViewTemp */

/******************************************************************************
 IStdPopupPaneX			TCL 1.1.1 DLP 9/18/91
 
 	Perform common initialization for a standard popup menu pane.
 	Sets the font to the standard system text font and size, and the
 	text style to plain.
******************************************************************************/

void CStdPopupPane::IStdPopupPaneX( Boolean fFirstInstall)
{
	textFont = GetSysFont();
	textSize = GetDefFontSize();
	textStyle = 0;
			
	if (fFirstInstall) 
		InstallHook();

	CalcDimensions();
	MakePopupBox();
	
}	/* CStdPopupPane::IStdPopupPaneX */

/******************************************************************************
 CStdPopupPane::InstallHook
 
 	Install a hook into the menu defproc of the popup menu in order to
 	force the width of the menu to allow for the standard down arrow 
 	within the popup.
 	The technique used is the same as in Apple's SC.006.PopupMenus
 	source code example. The hook can be completely disabled by commenting
 	out the "#define __INSTALL_MENU_HOOK__" at the top of this source file.
******************************************************************************/

void CStdPopupPane::InstallHook( void)
{
#ifdef	__INSTALL_MENU_HOOK__

	tMenuHookHndl	theHook = (tMenuHookHndl) NewHandleCanFail( sizeof( tMenuHook));
	MenuHandle		menu = itsMenu->GetMacMenu();
	
	FailNIL( theHook);
	
	(**theHook).jmp = 0x4EF9;		/* JMP instruction */
	(**theHook).addr = (long) MenuProc;
	(**theHook).realMenuProc = (**menu).menuProc;	
	
	(**menu).menuProc = (Handle) theHook;
	
	FlushCache();	/* TCL 1.1.1 DLP 9/25/91 */
					/* After modifying code, we must flush the cache on 68040s */

#endif

}	/* CStdPopupPane::InstallHook */

/******************************************************************************
 CalcDimensions
 
 	Calculate dimensions and location constants needed to properly draw the
 	popup.
******************************************************************************/

void CStdPopupPane::CalcDimensions()
{
	short		newWidth, newHeight;
	Rect		deltaSize;
	Str255		title;
	FontInfo	fInfo;
	
	Prepare();
	itsMenu->GetTitle( title);
	titleWidth = StringWidth( title);
	GetFontInfo( &fInfo);
	
		/* fontAscent is used when drawing the popup title and selected item */
	fontAscent = fInfo.ascent +1;

	if (width == kAutoSize)
	{
		/* calculte width such that title, the longest menu item, and		*/
		/* the down arrow all fit correctly									*/
		
		newWidth = titleWidth + kTitleGap + itsMenu->GetWidth(); /* TCL 1.1.1 DLP 9/18/91 */
	}
	else newWidth = width;
	
	if (height == kAutoSize)
	{
		/* calculate the height given the chosen font, but make sure		*/
		/* the height is at least an appropriate minimum size				*/
		
		newHeight = fInfo.ascent + fInfo.descent + fInfo.leading;
		newHeight = Max( kStdPopupHeight, newHeight);
	}
	else newHeight = height;
	
	SetRect( &deltaSize, 0, 0, newWidth - width, newHeight - height);
	ChangeSize( &deltaSize, FALSE);
	
		/* calculate the rect to hilite when the menu pops up, as per HIG	*/
		
	SetRect( &hiliteRect, 0, 1, titleWidth + kTitleGap,
				fInfo.ascent + fInfo.descent + fInfo.leading + 1);
				
		/* calculate the position to draw the down arrow SICN				*/
		
	sicnPt.h = width - kArrowWidth - kBorderWidth - kArrowRightInset;
	sicnPt.v = 6;
		
}	/* CStdPopupPane::CalcDimensions */

/******************************************************************************
 MakePopupBox
 
 	Create the pane and border used to draw the popup box and drop shadow
******************************************************************************/

void CStdPopupPane::MakePopupBox( void)
{
	CPaneBorder		*border;
	short			bWidth;
	
	bWidth = width - titleWidth - kTitleGap-2;
	
	popupPt.h = titleWidth + kTitleGap;
	popupPt.v = 1;
	
	itsPopupBox = new( CPane);
	itsPopupBox->IPane( this, this, bWidth, height-kBorderWidth-1, popupPt.h,
					1, sizFIXEDSTICKY, sizFIXEDSTICKY);
	
	border = new( CPaneBorder);
	border->IPaneBorder( kBorderFrame);
	border->SetShadow( 2, 2, 1, 1);

	itsPopupBox->SetBorder( border);
	
}	/* CStdPopupPane::MakePopupBox */

/******************************************************************************
 DoClick {OVERRIDE}
 
 	Hancle a click in the popup. As per HIG, we only popup the menu
 	when the popup box is clicked, not just any part of the pane.
 	We also hilite the title while the menu is popped up.
******************************************************************************/

void CStdPopupPane::DoClick( Point hitPt, short modifierKeys, long when)
{
	LongPt	lpt;
	Point	pt;
	
		/* is the click in the popup box?		*/
	
	QDToLongPt( hitPt, &lpt);
	FrameToWind( &lpt, &pt);
	FrameToWind( &lpt, &pt);
	if (itsPopupBox->Contains( pt))
	{
		InvertTitle();
				
			/* pass the popupPt as the hitPt */
		inherited::DoClick( popupPt, modifierKeys, when);
		
		InvertTitle();
	}

}	/* CStdPopupPane::DoClick */

/******************************************************************************
 NewMenuSelection {OVERRIDE}
 
 	Receive notification of a new menu selection. We need to redraw
 	the selected item text.
******************************************************************************/

void CStdPopupPane::NewMenuSelection( short itemSelected)
{
	LongRect	interior;
	
	/* the text is inside the popup box, so invalide the interior	*/
	/* minus the portion used for the down arrow icon				*/
	
	itsPopupBox->GetInterior( &interior);
	interior.right -= kArrowWidth;
	itsPopupBox->RefreshLongRect( &interior);
	
}	/* CStdPopupPane::NewMenuSelection */

/******************************************************************************
 DoKeyDown {OVERRIDE}
 
 	Handle a keypress by checking for a command key. The popup must be
 	in the chain of command, i.e. gGopher or a supervisor, for this
 	method to be automatically called.
******************************************************************************/

void CStdPopupPane::DoKeyDown(char theChar, Byte keyCode, EventRecord *macEvent)
{	
	#define kMinTicks	8	/* number of ticks to flash title for feedback */
	
	long	cmd;
	short	item;
	long	ticks;
	
	itsMenu->CheckCmdKey( theChar, keyCode, macEvent, &cmd, &item);
	
	if (cmd != cmdNull)
	{
		InvertTitle();
		DoCommand( cmd);
		itsMenu->SelectItem( item, pmToggle);
		Delay( kMinTicks, &ticks);
		InvertTitle();
	}

}	/* CStdPopupPane::DoKeyDown */

/******************************************************************************
 Prepare {OVERRIDE}
 
 	Set the text characteristics for drawing the popup
******************************************************************************/

void CStdPopupPane::Prepare( void)
{
	inherited::Prepare();
	
	TextFont( textFont);
	TextSize( textSize);
	TextFace( textStyle);

}	/* CStdPopupPane::Prepare */

/******************************************************************************
 InvertTitle
 
 	Invert the popup title, usually in response to a click or a command key
******************************************************************************/

void CStdPopupPane::InvertTitle()
{
	Rect r= hiliteRect;
	
	Prepare();
	InvertRect( &r);

}	/* CStdPopupPane::InvertTitle */

/******************************************************************************
 Draw
 
 	Draw the popup title, currently selected item, and down arrow. The
 	popup box itself is a subpane and is drawn automatically.
******************************************************************************/

void CStdPopupPane::Draw( Rect *area)
{
	Str255	str;
	
	itsMenu->GetTitle( str);
	MoveTo( 0, fontAscent);
	DrawString( str);
	DrawSICN( TCL_SICN, POPUP_SICN, sicnPt);
	
	itsMenu->GetCurrItemString( str);
	MoveTo( popupPt.h + kCheckGap, fontAscent);
	DrawString( str);
	
	inherited::Draw( area);

}	/* CStdPopupPane::Draw */

/******************************************************************************
 Activate
 
 	Make sure the popup box is drawn with a solid black pattern so it 
 	appears activated.
******************************************************************************/

void CStdPopupPane::Activate( void)
{
	inherited::Activate();
	itsPopupBox->GetBorder()->SetPattern( black);

}	/* CPopupPane::Activate */

/******************************************************************************
 Deactivate

	Make sure the popup box is drawn with a solid gray pattern so it 
 	appears deactivated.
******************************************************************************/

void CStdPopupPane::Deactivate( void)
{
	inherited::Deactivate();
	itsPopupBox->GetBorder()->SetPattern( gray);

}	/* CPopupPane::Deactivate */


/* This is the custom menu defproc that is called by the Menu Manager	*/
/* All it does is call the original defproc and then check if the		*/
/* message was mSizeMsg. If it was, it patches the menuWidth to allow	*/
/* room for the down arrow												*/

static pascal void MenuProc( short message, MenuHandle menu, Rect *menuRect,
					Point hitPt, short *whichItem)
{
	tMenuHookHndl	menuHook;
	Handle 			realMenuProc;
	SignedByte		state;
	
						/* get handle to the real menu proc */
						
	menuHook = (tMenuHookHndl) (**menu).menuProc;
	realMenuProc = (**menuHook).realMenuProc;
	
	LoadResource( realMenuProc);
	state = HGetState( realMenuProc);
	HLock( realMenuProc);
	CallPascal( message, menu, menuRect, hitPt, whichItem, *realMenuProc);
	HSetState( realMenuProc, state);
	
	if (message == mSizeMsg)
		(**menu).menuWidth += kArrowWidth-2;
}
