/******************************************************************************
 CApplication.c

							The Application Class
		
	Abstract class for a Mac application. Programmers must create a subclass
	of CApplication for their own programs. The application is at the top of
	the chain of command and defines the behavior of the program and its
	interaction with the Mac OS.
	
	SUPERCLASS = CDirectorOwner
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
		written by: Gregory H. Dow (version 1.1 by Dan Podwall)
		
	TCL 1.1 CHANGES
	[
		- IApplication now initializes the gApplication global, so the
	      main() function does not need to (but still can if it wants to).
	    - implemented GetPhase, DoAppleEvents methods.
	    - modified Notify to dispose of tasks that aren't caught by
	      any document
	]

 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "CApplication.h"
#include "Commands.h"
#include "OSChecks.h"
#include "CDirector.h"
#include "CDecorator.h"
#include "CError.h"
#include "TCLUtilities.h"
#include "TBUtilities.h"
#include "CDesktop.h"
#include "CBartender.h"
#include "CClipboard.h"
#include "CChore.h"
#include "CSwitchboard.h"
#include "CWindow.h"
#include "CList.h"
#include "CTask.h"
#include "CAppleEvent.h"
#include <AppleEvents.h>
#include <Traps.h>
#include <Script.h>
#include <GestaltEqu.h>

/*** Global Variables ***/

extern CApplication	*gApplication;		/* Application object				*/
extern CDesktop		*gDesktop;			/* The visible Desktop				*/
extern CBartender	*gBartender;		/* Manages all menus				*/
extern CClipboard	*gClipboard;		/* Copies and Pastes data			*/
extern CBureaucrat	*gGopher;			/* First in line to get commands	*/
extern CError		*gError;			/* Error handler					*/
extern long			gSleepTime;			/* Max time between events			*/
extern Boolean		gInBackground;		/* In background under MultiFinder	*/
extern OSType		gSignature;			/* Creator for Application's files	*/
extern EventRecord	gLastMouseDown;		/* Previous mouse down event		*/
extern EventRecord	gLastMouseUp;		/* Previous mouse up event			*/
extern CView		*gLastViewHit;		/* Last view clicked in				*/
extern CDecorator	*gDecorator;		/* Decorator for arranging windows	*/
extern short		gClicks;			/* Click counter, = 1 single click	*/
										/*                = 2 double click	*/
										/*    etc.							*/
extern CursHandle	gIBeamCursor;		/* I-beam for text views			*/
extern CursHandle	gWatchCursor;		/* Watch cursor for waiting			*/
extern RgnHandle	gUtilRgn;			/* Utility region					*/

/*** Class Constants ***/

#define		GROW_FAILURE	0L			/* Return codes from GrowZoneFunc	*/	
#define		GROW_SUCCESS	1L


	// Initialize cMaxSleepTime to 10 seconds. This is an arbitrary
	// value which gives screen savers, clocks, etc. some time,
	// without stealing too much time from other processes.
	// Applications can override it by simply assigning a new value to it.
	
long	CApplication::cMaxSleepTime = 600;

/**** I N I T I A L I Z A T I O N    M E T H O D S ****/


/******************************************************************************
 IApplication
 
		Initialize an Application.
 ******************************************************************************/

void	CApplication::IApplication(
	short		extraMasters,			/* Number of additional master		*/
										/*   blocks to allocate				*/
	Size		aRainyDayFund,			/* Bytes of memory to reserve for	*/
										/*   a rainy day					*/
	Size aCriticalBalance,				/* Bytes to save for critical 		*/
										/* operations						*/
	Size aToolboxBalance)				/* Bytes to save for the toolbox	*/
{
	gApplication = this;				/* initialize the global pointer */
	phase = appInitializing;			/* indicate initialization phase */

	InitToolbox();						/* Initialize Toolbox Managers		*/
	
										/* Fine tune the Memory Manager		*/
	InitMemory(extraMasters, aRainyDayFund, aCriticalBalance, aToolboxBalance);
	
									/** Instance Variables **/
                                    
    CDirectorOwner::IDirectorOwner(NULL);
 
 	InspectSystem();
 	
 	InstallPatches();

    MakeSwitchboard();
	itsIdleChores = new(CList);
	itsIdleChores->IList();
	itsUrgentChores = new(CCluster);
	itsUrgentChores->ICluster();
	urgentsToDo = FALSE;
	running = TRUE;
	unhandledTask = NULL;
	
									/** Global Variables **/
	gSignature = '???\?';
	gSleepTime = 0;						/* We want an early first Idle		*/
	
	
	MakeError();
	
										/* Cursors							*/
	gIBeamCursor = GetCursor(iBeamCursor);
	HNoPurge( (Handle) gIBeamCursor);
	gWatchCursor = GetCursor(watchCursor);
	HNoPurge( (Handle) gWatchCursor);
	
	gUtilRgn = NewRgn();

	MakeDesktop();
	MakeClipboard();
	MakeDecorator();
	SetUpFileParameters();
	SetUpMenus();
	
	gGopher = this;
	gLastViewHit = NULL;
	gLastMouseUp.when = 0L;
	gClicks = 0;
	
	ForceClassReferences();		/* TCL 1.1.1 DLP 9/27/91 */
}


/******************************************************************************
 InitToolbox

		Initialize the Macintosh Toolbox
 ******************************************************************************/

void	CApplication::InitToolbox()
{
	InitGraf(&thePort);					/* Standard initialization calls	*/
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(NULL);					/* ??? Add a ResumeProc				*/
	InitCursor();
}


/******************************************************************************
 InitMemory
 
		Initialize the Heap and create extra master pointer blocks
 ******************************************************************************/

void	CApplication::InitMemory(
	short		extraMasters,			/* Number of additional master		*/
										/*   blocks to allocate				*/
	Size		aRainyDayFund,			/* Bytes of memory to reserve for	*/
										/*   a rainy day					*/
	Size aCriticalBalance,				/* Bytes to reserve for critical	*/
										/* operations						*/
	Size aToolboxBalance)				/* Bytes to reserve for those toolbox */
										/* routines that don't fail gracefully */
{
	ASSERT( aRainyDayFund >= aCriticalBalance);
	ASSERT( aCriticalBalance >= aToolboxBalance);
	
	MaxApplZone();						/* Grow heap to maximum size to		*/
										/*   help prevent fragmentation		*/
	
		/* Call MoreMasters to allocate extra master pointers.  */
		/* Experiment to determine how many master pointers the */
		/* application uses under session of heavy use.			*/
	
	while (extraMasters-- > 0)
		MoreMasters();

	SetGrowZone(GrowZoneFunc);			/* Traps out of memory conditions	*/
	rainyDay = NULL;					/* Clear rainyDay handle in case	*/
										/*   allocating the rainy day fund	*/
										/*   fail! This is really bad.		*/

	rainyDayFund = aRainyDayFund;		/* Set instance variables			*/
	criticalBalance = aCriticalBalance;
	toolboxBalance = aToolboxBalance;
	rainyDayUsed = FALSE;
	memWarningIssued = FALSE;
    canFail = FALSE;
    inCriticalOperation = FALSE;
    tempAllocation = 0;

	rainyDay = NewHandleCanFail(rainyDayFund);	/* Reserve a block of memory */
	FailNIL( rainyDay);
}

/******************************************************************************
 InstallPatches
 
		Install the TCL trap patches. Patches LoadSeg and ExitToShell
 ******************************************************************************/
 
 	static long origETS, origLoadSeg;
 
 	static pascal void ETS_Patch( void)
 	{
 		long realETS;
 		long oldA5 = SetCurrentA5();
 		
 		if (gApplication)
 			gApplication->RemovePatches();
 		realETS = origETS;
 		
 		SetA5( oldA5);
		CallPascal( realETS);
 	}
 	
 	static void LoadSeg_Patch( void)
 	{
 		asm
 		{
 			clr.l		-(sp)				; reserve some space
			movem.l		a0-a5/d0-d7, -(sp)	; save the regs
 			move.w  	64(sp), d7          ; get seg#
 			
 			
 			movea.l		CurrentA5, a5
 			
 			move.b  	#kAllocCanFail, -(sp) ; this allocation
 			jsr			SetAllocation		; can fail
 			move.b  	d0, d6				; save old canFail flag
 			addq.l		#2, sp
 			
 			clr.l		-(sp)				; do a GetResource on 
 			move.l  	#'CODE', -(sp)      ; the code segment
 			move.w  	d7, -(sp)
 			GetResource
 			movea.l   	(sp)+, a3			; save the handle
 			
 			move.b  	d6, -(sp)			; restore old allocation
 			jsr			SetAllocation
 			addq.l		#2, sp
 			
 			move.l		a3, -(sp)
 			jsr			FailNIL				; returns if OK
 			
 			move.l		origLoadSeg, 56(sp) ; store original LoadSeg on stack
 			movem.l		(sp)+, a0-a5/d0-d7  ; restore regs
 			rts								; return to original LoadSeg
 		}
 	}

void	CApplication::InstallPatches( void)
{
	short trap;
	
	trap = _ExitToShell & 0x3FF;
	origETS = NGetTrapAddress( trap, ToolTrap);
	NSetTrapAddress( (long) ETS_Patch, trap, ToolTrap);

	trap = _LoadSeg & 0x3FF;
	origLoadSeg = NGetTrapAddress( trap, ToolTrap);
	NSetTrapAddress( (long) LoadSeg_Patch, trap, ToolTrap);
}

/******************************************************************************
 RemovePatches
 
		Remove the TCL trap patches
 ******************************************************************************/

void	CApplication::RemovePatches( void)
{
	short trap;

	if (origETS)
	{
		trap = _ExitToShell & 0x3FF;	
		NSetTrapAddress( origETS, trap, ToolTrap);
		trap = _LoadSeg & 0x3FF;
		NSetTrapAddress( origLoadSeg, trap, ToolTrap);
		
	}
}

/******************************************************************************
 ForceClassReferences
 
	If there are classes in your program that are only created via new by
	name and never referenced directly, you must add a dummy reference to
	prevent smart-linking from stripping the classes from your built application.
	Override ForceClassReferences in your application subclasses and reference
	the classes as shown below. CApplication automatically calls 
	ForceClassReferences for you.
	
 ******************************************************************************/

void CApplication::ForceClassReferences( void)
{
	#if 0
	
		// example of referencing classes used only with new_by_name
	Boolean alwaysFalse = FALSE;
	CObject dummy;
	
	if (alwaysFalse)
	{
		member( dummy, CCheckbox);
		member( dummy, CRadioControl);
		member( dummy, CRadioGroupPane);
		.
		.
		.
	}
	#endif
}


/******************************************************************************
 InspectSystem
 
	Gather information about the current OS
 ******************************************************************************/

void CApplication::InspectSystem( void)
{
	OSErr		err;
	SysEnvRec	environs;
	long		response;
	
	gSystem.hasGestalt = TrapAvailable( _GestaltDispatch);
	gSystem.hasWNE = WNEIsImplemented();
	gSystem.hasScriptMgr = TrapAvailable( _ScriptUtil);
	gSystem.scriptsInstalled = 1; 	/* assume only Roman script	*/

	err = SysEnvirons( 1, &environs);
	if (err == noErr)
	{
		gSystem.hasColorQD = environs.hasColorQD;
		gSystem.hasFPU = environs.hasFPU;
		gSystem.systemVersion = environs.systemVersion;
	}
	else
	{
		gSystem.hasColorQD = FALSE;
		gSystem.hasFPU = FALSE;
		gSystem.systemVersion = 0;
	}	
	
	if (gSystem.hasGestalt)
	{
		gSystem.hasAppleEvents = Gestalt( gestaltAppleEventsAttr, &response) == noErr;
		gSystem.hasAliasMgr = Gestalt( gestaltAliasMgrAttr, &response) == noErr;
		gSystem.hasEditionMgr = Gestalt( gestaltEditionMgrAttr, &response) == noErr;
		gSystem.hasHelpMgr = Gestalt( gestaltHelpMgrAttr, &response) == noErr;
		if (gSystem.hasScriptMgr)
		{
			err = Gestalt( gestaltScriptCount, &response);
			if (err == noErr)
				gSystem.scriptsInstalled = (short) response;
		}
	}
	else
	{
		/* If we don't have Gestalt, then we can't have any System 7 features	*/
		
		gSystem.hasAppleEvents = FALSE;
		gSystem.hasAliasMgr = FALSE;
		gSystem.hasEditionMgr = FALSE;
		gSystem.hasHelpMgr = FALSE;
		if (gSystem.hasScriptMgr)
			gSystem.scriptsInstalled = GetEnvirons( smEnabled);			
	}
}

/******************************************************************************
 MakeDesktop
 
		Create the global Desktop object, which is the top level of
		the visual hierarchy. Overrride this method to use a desktop
		which supports floating windows or other extensions to the
		standard desktop.
 ******************************************************************************/

void	CApplication::MakeDesktop()
{
	gDesktop = new(CDesktop);			/* Use a standard Desktop			*/
	gDesktop->IDesktop(this);
}	


/******************************************************************************
 MakeClipboard
 
		Create the global Clipboard object
 ******************************************************************************/

void	CApplication::MakeClipboard()
{
	gClipboard = new(CClipboard);		/* Use standard Clipboard			*/
	gClipboard->IClipboard(this, TRUE);
}	


/******************************************************************************
 MakeDecorator
 
		Create the global Decorator object, which controls the
		arrangement of windows on the screen. Override if a
		different Decorator is desired.
 ******************************************************************************/

void	CApplication::MakeDecorator()
{
	gDecorator = new(CDecorator);		/* Use standard Decorator			*/
	gDecorator->IDecorator();
}	

/******************************************************************************
 MakeSwitchboard
 
 	Create application's switchboard. Override if a different switchboard
 	is desired.
******************************************************************************/

void CApplication::MakeSwitchboard( void)
{
	itsSwitchboard = new(CSwitchboard);
	itsSwitchboard->ISwitchboard();
}

/******************************************************************************
 MakeBartender
 
 	Create global bartender. Override if a different bartender
 	is desired.
******************************************************************************/

void CApplication::MakeBartender( void)
{
	gBartender = new(CBartender);
	gBartender->IBartender(MBARapp);
}

/******************************************************************************
 MakeError
 
 	Create global error handler. Override if a different error handler
 	is desired.
******************************************************************************/

void CApplication::MakeError( void)
{
	gError = new( CError);
}

/******************************************************************************
 SetUpFileParameters
 
		Set values of the parameters used by the Standard File Package.
 ******************************************************************************/

void	CApplication::SetUpFileParameters()
{
									/** Get File Parameters **/
	sfNumTypes = -1;					/* Display all files				*/
	sfFileTypes[0] = '???\?';			/* Just put something here			*/
	sfFileFilter = NULL;				/* No extra filtering of files		*/
	sfGetDLOGHook = NULL;				/* No special handling of items		*/
	sfGetDLOGid = getDlgID;				/* Use built-in get dialog box		*/
	sfGetDLOGFilter = NULL;				/* No special event processing		*/
}	


/******************************************************************************
 SetUpMenus
 
		Set up the menus used by the application
 ******************************************************************************/

void	CApplication::SetUpMenus()
{
	MakeBartender();
	AddResMenu(GetMHandle(MENUapple), 'DRVR');
	gBartender->SetDimOption(MENUapple, dimNONE);
}

/**** A C T I O N   M E T H O D S ****/

/******************************************************************************
 Notify {OVERRIDE}

		A subordinate is notifying the Application that a task has been
		completed.
 ******************************************************************************/

void	CApplication::Notify(
	CTask		*theTask)				/* The completed task				*/
{
	unhandledTask = theTask;
}										/* Null Method						*/


/******************************************************************************
 DoKeyDown {OVERRIDE}

		Respond to a keystroke. Application has no supervisor to whom to
		pass the buck. Default action is to ignore keystrokes.
 ******************************************************************************/

void	CApplication::DoKeyDown(
	char			theChar,			/* The associated character			*/
	Byte			keyCode,			/* Code for the associated key		*/
	EventRecord		*macEvent)			/* Key down event record			*/
{
}										/* Null Method						*/


/******************************************************************************
 DoAutoKey {OVERRIDE}

		Respond to a key being held down. By default, keystrokes are ignored.
 ******************************************************************************/

void	CApplication::DoAutoKey(
	char			theChar,			/* The associated character			*/
	Byte			keyCode,			/* Code for the associated key		*/
	EventRecord		*macEvent)			/* Autokey event record				*/
{
}										/* Null Method						*/


/******************************************************************************
 DoKeyUp {OVERRIDE}

		Respond to a key being released. Usually, these events will be
		masked out by the operating system. Ignore them by default.
 ******************************************************************************/

void	CApplication::DoKeyUp(
	char			theChar,			/* The associated character			*/
	Byte			keyCode,			/* Code for the associated key		*/
	EventRecord		*macEvent)			/* Key up event record				*/
{
}										/* Null Method						*/


/******************************************************************************
 DoCommand

		Handle a command. Application has no supervisor which can handle the
		command. If we don't handle the command here, it will never be
		executed (and that's an error).
 ******************************************************************************/

void	CApplication::DoCommand(
	register long	theCommand)			/* Command number					*/
{
	Str255		theDA;					/* Name of Desk Accessory to open	*/
	SFReply		macSFReply;				/* Standard File reply record		*/

	if (theCommand < 0) {
		if (HiShort(-theCommand) == MENUapple) {
										/* Open a DA or launch another		*/
										/*   application under MultiFinder	*/
			GetItem(GetMHandle(MENUapple), LoShort(-theCommand), theDA);
			OpenDeskAcc(theDA);
		}
	} else {
			
		switch (theCommand) {
		
			case cmdNew:
				SetCursor(*gWatchCursor);
				CreateDocument();
				break;
				
			case cmdOpen:
				ChooseFile(&macSFReply);
				if (macSFReply.good) {
					SetCursor(*gWatchCursor);
					OpenDocument(&macSFReply);
				}
				break;
				
			case cmdClose:
				if (IsSystemWindow((WindowPeek)FrontWindow()))
					CloseDeskAcc(((WindowPeek)FrontWindow())->windowKind);
				break;
								
			case cmdQuit:
				Quit();
				break;
				
			case cmdUndo:
			case cmdCut:
			case cmdCopy:
			case cmdPaste:
			case cmdClear:
				SystemEdit((short)(theCommand - cmdUndo));
				break;
				
			case cmdToggleClip:
				gClipboard->Toggle();
				break;
		}
	}
}
	

/******************************************************************************
 UpdateMenus {OVERRIDE}

		Perform menu management tasks
 ******************************************************************************/

void	CApplication::UpdateMenus()
{
	Str63	undoStr;
	CWindow *topWindow;
	Boolean isModalWindow;
	
		// Is the frontmost window modal?

	topWindow = gDesktop->GetTopWindow();
	isModalWindow = topWindow? topWindow->IsModal() : FALSE;
	
	gBartender->EnableCmd(cmdQuit);

	if (gInBackground) 
	{
		gBartender->EnableCmd(cmdClose);
		gBartender->EnableCmd(cmdUndo);
		gBartender->EnableCmd(cmdCut);
		gBartender->EnableCmd(cmdCopy);
		gBartender->EnableCmd(cmdPaste);
		gBartender->EnableCmd(cmdClear);
	} 
	else if (!isModalWindow)
	{
		gBartender->EnableCmd(cmdToggleClip);
	}

	if (!rainyDayUsed && !isModalWindow) 
	{
		gBartender->EnableCmd(cmdNew);
		gBartender->EnableCmd(cmdOpen);
	}
	
	GetIndString(undoStr, STRcommon, strUNDO);
	gBartender->SetCmdText(cmdUndo, undoStr);
}


/**** M E M O R Y   M A N A G E M E N T ****/
		

/******************************************************************************
 RequestMemory
 
 	Indicate whether subsequent memory requests can fail if not enough
 	memory is available.
 ******************************************************************************/

void	CApplication::RequestMemory(
	Boolean		aCanFail)
{
	canFail = aCanFail;
}

/******************************************************************************
 SetCriticalOperation
 
 	Indicate whether subsequent memory requests should be regarded as
 	part of a critical operation. Critical operations can use more of the memory
 	reserve to ensure that they succeed. 
 	
 ******************************************************************************/

void CApplication::SetCriticalOperation( Boolean inCriticalOperation)
{
	this->inCriticalOperation = inCriticalOperation;
}

/******************************************************************************
 GrowMemory

		A memory request can't be filled by the Mac Memory Manager. Free up
		memory if possible or gracefully shutdown.
		
		A rainy day fund of reserve memory is maintained to monitor low
		memory conditions. If the application can't free up enough memory
		on its own, we use the reserve fund. The user is notified when
		we tap into the reserve fund so that he/she can take steps to free
		up memory (perhaps by closing windows/files).
		
		If there is not enough memory in reserve to cover a request, we
		notify the user of impending doom. The Toolbox Memory Manager will
		return an error code of memFullErr. This will probably cause some kind
		of System Bomb.
		
 ******************************************************************************/

long	CApplication::GrowMemory( Size	bytesNeeded)
{
	Size	bytesToFree;
	Size	rainyDayBalance;
	Size	grow;
	Boolean	success;
	
	MemoryShortage(bytesNeeded);		/* Try to release some memory		*/
										/* See how much we still need		*/
	bytesToFree = bytesNeeded - MaxMem(&grow);
		
	if (bytesToFree < 1)				/* Hooray! Enough memory was freed!	*/
		return GROW_SUCCESS;			/*   Let's get out of here			*/
		
		
	success = FALSE;
	
	if (rainyDay != NULL) 
	{
		rainyDayBalance = GetHandleSize(rainyDay);
		if (inCriticalOperation)
		{
			if (rainyDayBalance - bytesToFree + tempAllocation >= toolboxBalance)
			{
				SetHandleSize( rainyDay, rainyDayBalance - bytesToFree);
				success = TRUE;
			}
		}
		else
		{
			if (rainyDayBalance - bytesToFree + tempAllocation >= criticalBalance)
			{
				SetHandleSize( rainyDay, rainyDayBalance - bytesToFree);
				success = TRUE;
			}			
		}
		
		if (!success && !canFail)
		{
			if (rainyDayBalance - bytesToFree >= MINIMUM_BALANCE)
			{
										/* We have enough savings to cover	*/
										/*   the request and maintain our	*/
										/*   minimum balance				*/
				SetHandleSize(rainyDay, MINIMUM_BALANCE);
				success = TRUE;
			}
			else
			{
				DisposHandle( rainyDay);
				rainyDay = NULL;
				success = TRUE;
			}
		}
		
		if (success)
		{
			gSleepTime = 0L;			/* Idle as soon as possible			*/
			rainyDayUsed = TRUE;		/* We've tapped the rainy day fund	*/
		}
	}
		
	if (success)
		return GROW_SUCCESS;
 	else
 	{
		if (canFail) 		// Requestor is prepared for failure
			return GROW_FAILURE;	
		else 				// Try the last resort
			return OutOfMemory(bytesNeeded);
	}
}	

/******************************************************************************
 MemoryShortage
 
		There is not enough memory to fill a request. Try to free up some
		memory. Subclasses should override this method.

		The application should free all non-essential blocks of
		memory (such as those used to cache data for speed or resources
		marked as unpurgeable to allow for quick access) and disable
		commands which could consume a lot of memory (such as opening files
		or creating new documents). The rainy day fund created by InitMemory
		should be large enough to allow (or even force) the user to take
		actions to reduce the memory strain without losing data (such as
		closing windows/files).

		IMPORTANT: This method should NOT allocate any blocks of memory.
 ******************************************************************************/

void	CApplication::MemoryShortage(
	Size	bytesNeeded)				/* Amount of memory requested		*/
{
}										/* Null Method						*/
		

/******************************************************************************
 MemoryReplenished
 
 		The rainy day fund of reserve memory has been replenished.
 		Subclasses must override this method if certain actions taken
 		by the MemoryShortage method need to be undone. For example, if
 		the "Open..." command was disabled by MemoryShortage, it should
 		be enabled here.
 ******************************************************************************/

void	CApplication::MemoryReplenished()
{
}										/* Null Method						*/
		

/******************************************************************************
 OutOfMemory
 
 		A memory request cannot be satisfied. At this point, the rainy day
 		fund has been liquidated. This method posts an Alert if there's
 		enough memory to do so, and then jumps back to the event loop.
 		Classes which override this method should pass back the appropriate
 		return value (same as that for a GrowZone function) if necessary.
 		This method performs a long jump, so no return value is needed.
 ******************************************************************************/

long	CApplication::OutOfMemory(
	Size	bytesNeeded)				/* Amount of memory requested		*/
{
	
	Failure(memFullErr, 0);
}


/**** E X E C U T I O N   M E T H O D S ****/


/******************************************************************************
 Run
 
		Run the application by processing events.
 ******************************************************************************/

		/*
		 * Perform a Chore
		 */
		static void	Chore_Perform(
			CChore	*theChore,
			long	*minSleep)
		{
			long	maxSleep = MAXLONG;
			
			theChore->Perform(&maxSleep);
			*minSleep = Min(*minSleep, maxSleep);
		}

void	CApplication::Run()
{	
	Preload();							/* Open/Print files selected from	*/
										/*   the Finder before launching	*/
										/*   application					*/
										
	TRY
	{	
		phase = appRunning;		/* now that an exception handler is posted,		*/
								/* mark the application as running				*/
		do {								/* Continuously poll for events		*/
											/*   and respond to them			*/
				
			Process1Event();
							
		} while (running);					/*   Until flag is turned off		*/
		
		phase = appQuitting;	/* exception handler is about to be removed,	*/
								/* mark the app as quitting.					*/
	}
	CATCH
	{
		/*
		 *	All exceptions (other than application initialization and cleanup)
		 *	are ultimately caught here. We display an alert to tell the
		 *	user what happened, and then do a RETRY to keep the application
		 *	going.
		 *	Special case if the error code is kSilentFailure. This tells
		 *	that either an alert has already been displayed or none is
		 *	desired.
		 */
		 
		 if (gLastError != kSilentErr)
		 	ErrorAlert( gLastError, gLastMessage);
		 HiliteMenu( 0);
		 
		 RETRY;
	}
	ENDTRY;
	
	RemovePatches();
}

/******************************************************************************
 Process1Event
 
 	Process and dispatch one event.
******************************************************************************/

void CApplication::Process1Event( void)
{
	Boolean				aDAIsActive;	/* Is a DA the front window?		*/
	long				minSleep = 0;

	aDAIsActive = IsSystemWindow((WindowPeek)FrontWindow());

	itsSwitchboard->ProcessEvent();
		
	SetCriticalOperation( FALSE); // A user command may have initiated a
								  // critical operation, e.g. Save, but
								  // it is complete once the event is handled
	tempAllocation = 0;			  // any temporary allocations should have been
								  // released by now.
								  
	gLastError = 0;
	gLastMessage = 0;
								  
	if (urgentsToDo) {				/* Carry out urgent chores			*/
		itsUrgentChores->DoForEach1((EachFunc1) Chore_Perform, (long)&minSleep);
		itsUrgentChores->DisposeItems();
		urgentsToDo = FALSE;
	}
	

									/* Check for context-switch with	*/
									/*   Desk Accessories				*/
	if (IsSystemWindow((WindowPeek)FrontWindow())) {
		gSleepTime = 0;
		if (!aDAIsActive) {
			aDAIsActive = TRUE;
			SwitchToDA();
		}
	} else {
		if (aDAIsActive) {
			aDAIsActive = FALSE;
			SwitchFromDA();
		}
	}
	CView::ForceNextPrepare();
	
	ForgetObject( unhandledTask);

}
/******************************************************************************
 Preload
 
		Open/Print files selected by the user from the Finder before
		launching the application
 ******************************************************************************/

void	CApplication::Preload()
{
	short			openOrPrint;		/* Type of action requested			*/
	short			numPreloads;		/* No. of files to process			*/
	register short	i;					/* Index for files					*/
	AppFile			macAppFile;			/* Info about a file				*/
	SFReply			macSFReply;			/* Standard File info record		*/
	
										/* Check if files were selected		*/
	CountAppFiles(&openOrPrint, &numPreloads);
	
	if (numPreloads > 0)
	{
		SetCursor( *gWatchCursor);
	
		for (i = 1; i <= numPreloads; i++) {/* Process files one by one			*/
			GetAppFiles(i, &macAppFile);	/* Get info about the file			*/
											/* Copy info into a Standard File 	*/
											/*   record							*/
			macSFReply.fType = macAppFile.fType;
			macSFReply.vRefNum = macAppFile.vRefNum;
			macSFReply.version = macAppFile.versNum;
			CopyPString(macAppFile.fName, macSFReply.fName);
			
			OpenDocument(&macSFReply);		/* Automatically open selected file	*/
			
			if (openOrPrint == appPrint) {	/* User wants to print Document		*/
				gDesktop->UpdateWindows();
				gGopher->DoCommand(cmdPrint);
			}
				
			ClrAppFiles(i);					/* We've processed this file		*/
		}
	}
	StartUpAction(numPreloads);			/* Perform start up action			*/
}


/******************************************************************************
 StartUpAction
 
		Perform any desired start up action. This method is invoked after
		files selected from the Finder have been opened/printed. The number
		of preloaded files is passed as a parameter. If no files were
		preloaded, this default method acts as if the user selected "New"
		and sends the Gopher a DoCommand(cmdNew) message.
 ******************************************************************************/

void	CApplication::StartUpAction(
	short	numPreloads)
{
	FlushEvents(everyEvent, 0);
	
		// TCL 1.1 - only do cmdNew if there are no AppleEvents.
		// If there are, we anticipate getting the open app event
		
	if (!gSystem.hasAppleEvents && (numPreloads == 0))
	{
		gGopher->DoCommand(cmdNew);
	}
}


/******************************************************************************
 Suspend
 
		Program is about to be swapped out under MultiFinder. Send the
		Suspend message to each of the Application's Directors.
 ******************************************************************************/
		
void	CApplication::Suspend()
{
	inherited::Suspend();

	gInBackground = TRUE;				/* We are going into the background	*/
}	
	

/******************************************************************************
 Resume
 
		Program is being swapped in under MultiFinder
 ******************************************************************************/
	
void	CApplication::Resume()
{
	CView::ForceNextPrepare();
										/* Do nothing if a DA is in front	*/
	if (!IsSystemWindow((WindowPeek)FrontWindow())) {
		gInBackground = FALSE;	
		
		inherited::Resume();
	}
}	


/******************************************************************************
 SwitchToDA
 
		A DA is becoming active
 ******************************************************************************/

void	CApplication::SwitchToDA()
{
	CurDeactive = NULL;
	Suspend();
}


/******************************************************************************
 SwitchFromDA
 
		Our application is becoming active after a DA had been running
 ******************************************************************************/

void	CApplication::SwitchFromDA()
{
	CurActivate = NULL;
	Resume();
}


/******************************************************************************
 Quit
 
		User requested quit. Need to perform a graceful shutdown,
		giving the user the opportunity to save changed files or to abort
		the quit process.
 ******************************************************************************/

Boolean	CApplication::Quit()
{
	register CWindow	*theWindow;
	CDirector			*theDirector;
	
	running = FALSE;					/* Assume we will stop running		*/
	
										/*   (Does not include floating		*/
										/*    windows or DAs)				*/
	theWindow = gDesktop->GetTopWindow();
	while (theWindow != NULL) {
		theDirector = (CDirector*) theWindow->GetSupervisor();
										/* Tell window's Director to quit	*/
										/*   If Quit method returns FALSE,	*/
										/*   abort the quit sequence		*/
		if (!theDirector->Quit()) {
			running = TRUE;				/* User still wants to run program	*/
			break;
		}
		theWindow = gDesktop->GetTopWindow();
	}
	return (running == FALSE);	/* TCL 1.1.1 DLP 9/18/91 */
}


/******************************************************************************
 Exit
 
		Program is about to terminate. Last chance to clean up.
 ******************************************************************************/

void	CApplication::Exit()
{
}										/* NULL Method						*/


/******************************************************************************
 JumpToEventLoop
 
		Jump to the beginning of the main event loop. Useful when trying
		to recover from an error.
		
		An error may occur BEFORE program ever gets to the event loop. In
		this case the jump buffer entry for A1 (the address to jump to) will
		be NULL and we exit the program.
 ******************************************************************************/

void	CApplication::JumpToEventLoop()
{
	/*
	 *	This call to Failure produces nearly the same result as the previous code.
	 *	If we are called within the activation of CApplication::Run,
	 *	then we jump to the event loop. Otherwise we jump to the
	 *	default handler posted in main, wich calls CApplication::Exit
	 *	followed by ExitToShell.
	 *
	 *	The only difference is that Failure will also cause any other
	 *	pending exception handlers to be called, which is really the
	 *	correct thing to do.
	 */
	 
	Failure( kSilentErr, 0);	/* kSilentErr inhibits the error alert */
	
}


/**** D O C U M E N T   H A N D L I N G   M E T H O D S ****/


/******************************************************************************
 CreateDocument
 
		Make a document in response to the "New" menu selection. This method
		must be overridden if the standard "New" command is used.
 ******************************************************************************/

void	CApplication::CreateDocument()
{
}										/* NULL Method						*/


/******************************************************************************
 OpenDocument
 
		Open an existing file and create a document object for displaying
		information. This method must be overridden if the standard "Open"
		command is used.
 ******************************************************************************/

void	CApplication::OpenDocument(
	SFReply		*macReply)
{
}										/* NULL Method						*/


/******************************************************************************
 ChooseFile
 
		Let the user select a file to open. Default procedure is to use
		the Standard GetFile dialog box.
 ******************************************************************************/

void	CApplication::ChooseFile(
	SFReply		*macSFReply)
{
	Point		corner;					/* Top left corner of dialog box	*/
	Boolean		wasLocked;
	
										/* Center dialog box on the screen	*/
	FindDlogPosition('DLOG', sfGetDLOGid, &corner);
	
	wasLocked = Lock( TRUE);
	
	SFPGetFile(corner, "\p", sfFileFilter, sfNumTypes, sfFileTypes,
				sfGetDLOGHook, macSFReply, sfGetDLOGid, sfGetDLOGFilter);
				
	Lock( wasLocked);
}

/**** C H O R E   H A N D L I N G   M E T H O D S ****/


/******************************************************************************
 Idle

		No event directed to our application is pending. Here's our chance
		to perform periodic tasks. The classic example is flashing a
		text insertion cursor.
 ******************************************************************************/

void	CApplication::Idle(
	EventRecord		*macEvent)			/* Usually a null or system event	*/
{
	Size					grow;		/* Amount heap may be grown			*/
	register CBureaucrat	*theBureaucrat;
	long					minSleep;
	long					maxSleep;


	if (rainyDayUsed) {					/* Memory reserve needs to be		*/
										/*   replenished					*/
		if (MaxMem(&grow) > rainyDayFund + MINIMUM_BALANCE) {
										/* Enough memory has been freed		*/
			if (rainyDay != NULL) {
				DisposHandle(rainyDay);
			}
			rainyDay = NewHandleCanFail(rainyDayFund);
			FailNIL( rainyDay);
			
			rainyDayUsed = FALSE;
			memWarningIssued = FALSE;
			MemoryReplenished();
		} else if (!memWarningIssued) {
										/* Inform user of memory shortage	*/
			gError->PostAlert(STRmemWarn, iMEM_LOW);
			memWarningIssued = TRUE;
		}
	}
			
										/* Idle time for objects in the		*/
										/*   active chain of command		*/
	theBureaucrat = gGopher;			/* Start at the bottom				*/
	minSleep = cMaxSleepTime;
	do {								/* Run thru the chain of command	*/
										/*   Let bureaucrat kill some time	*/
		maxSleep = MAXLONG;
		theBureaucrat->Dawdle(&maxSleep);
		minSleep = Min(minSleep, maxSleep);
										/*   Move up to his/her boss		*/
		theBureaucrat = theBureaucrat->itsSupervisor;
	} while (theBureaucrat != NULL);	/* Until end of command chain		*/
	
										/* Carry out assigned chores		*/
	itsIdleChores->DoForEach1((EachFunc1) Chore_Perform, (long)&minSleep);
	gSleepTime = minSleep;
}


/******************************************************************************
 AssignIdleChore
 
		Give Application a chore to perform at idle time. Idle chores are
		performed whenever there are no events pending for the Application.
 ******************************************************************************/

void	CApplication::AssignIdleChore(
	CChore	*theChore)					/* Chore to perform					*/
{
	itsIdleChores->Add(theChore);
	gSleepTime = 0;						/* Force immediate idle				*/
}	


/******************************************************************************
 CancelIdleChore
 
		Relieve Application of an idle time chore. A chore must NOT cancel
		itself or any other chore.
 ******************************************************************************/

void	CApplication::CancelIdleChore(
	CChore	*theChore)					/* Chore to cancel					*/
{
	itsIdleChores->Remove(theChore);
}	


/******************************************************************************
 AssignUrgentChore
 
		Give Application a chore to perform as soon as possible. Urgent
		chores are performed only once, then automatically disposed.
		An urgent chore must NOT post another urgent chore.
 ******************************************************************************/

void	CApplication::AssignUrgentChore(
	CChore	*theChore)					/* Chore to perform					*/
{
	itsUrgentChores->Add(theChore);
	urgentsToDo = TRUE;
}	

/******************************************************************************
 GetPhase
 
 	Return the instance variable representing the current application phase.
 	The valid values are appInitializing, appRunning, or appQuiting.
******************************************************************************/

short CApplication::GetPhase( void)
{
	return phase;
	
}

/******************************************************************************
 DoOpenOrPrintDocEvent

	Handle the kAEOpenDocuments, kAEPrintDocuments AppleEvents
 ******************************************************************************/

void CApplication::DoOpenOrPrintDocEvent( CAppleEvent *theEvent)
{
	CArray		*docList = NULL;
	long		i, numDocs;
	FSSpec		currDoc;
	DescType	eventID;
	
	TRY
	{
		SetCursor( *gWatchCursor);

		docList = theEvent->ExtractFromDescList( keyDirectObject, typeFSS, 
								sizeof( FSSpec));
		if (theEvent->GotRequiredParams())
		{
			eventID = theEvent->GetEventID();
			
			numDocs = docList->GetNumItems();
			for (i = 1; i <= numDocs; i++)
			{
				docList->GetItem( &currDoc, i);
				
				{
					short 	wdRefNum;
					SFReply	reply;
					FInfo	fileInfo;
				
					FailOSErr( OpenWD( currDoc.vRefNum, currDoc.parID,
									gSignature, &wdRefNum));
									
					FailOSErr( FSpGetFInfo( &currDoc, &fileInfo));
					
					reply.good = TRUE;
					reply.vRefNum = wdRefNum;
					reply.fType = fileInfo.fdType;
					reply.version = reply.copy = 0;
					CopyPString( currDoc.name, reply.fName);
					
					OpenDocument( &reply);
					
					if (eventID == kAEPrintDocuments)
					{
						gDesktop->UpdateWindows();
						gGopher->DoCommand( cmdPrint);
						gGopher->DoCommand( cmdClose);
					}
				}
			}
			// event was handled successfully
		
			theEvent->SetErrorResult( noErr);
		}	
		ForgetObject( docList);
		
	}
	CATCH
	{
		ForgetObject( docList);
	}
	ENDTRY;
}

/******************************************************************************
 PackageAppleEvent

	Package an incoming AppleEvent and its default reply into a CAppleEvent
	object. This method always returns an object of class CAppleEvent, but
	you may wish to override it and return your own subclass of CAppleEvent.
	
	The eventClass and eventID parameters are really of type DescType,
	but dragging AppleEvents.h into the CApplication header really hits
	compile times.
 ******************************************************************************/

CAppleEvent *CApplication::PackageAppleEvent( struct AppleEvent *theEvent, 
						struct AppleEvent *theReply, long theRefCon, 
						long eventClass, long eventID)
{
	CAppleEvent *event = new CAppleEvent;
	
	event->IAppleEvent( (AppleEvent*) theEvent, (AppleEvent*) theReply, 
					theRefCon, eventClass, eventID);
	return event;
}

/******************************************************************************
 DoAppleEvent {OVERRIDE}

	Respond to an AppleEvent. This method handles the four required
	AppleEvents: kAEOpenApplication, kAEOpenDocuments, kAEPrintDocuments,
	and kAEQuitApplication.
	
 ******************************************************************************/

void CApplication::DoAppleEvent( CAppleEvent *anAppleEvent)
{
	DescType	eventClass;
	DescType	eventID;
	OSErr		err;
	
	eventClass = anAppleEvent->GetEventClass();
	eventID = anAppleEvent->GetEventID();
	
	if (eventClass == kCoreEventClass)
	{
		switch (eventID)
		{
			case kAEOpenApplication:
					if (anAppleEvent->GotRequiredParams())
					{
						gGopher->DoCommand( cmdNew);
						anAppleEvent->SetErrorResult( noErr);
					}
					break;
					
			case kAEOpenDocuments:
					DoOpenOrPrintDocEvent( anAppleEvent);
					break;
					
			case kAEPrintDocuments:
					err = anAppleEvent->RequestInteraction( MAXLONG);
					if (err == noErr)
					{
						DoOpenOrPrintDocEvent( anAppleEvent);
					}
					else
						anAppleEvent->SetErrorResult( err);
					break;
				
			case kAEQuitApplication:
			
					/*
					 * This quit handler always requests interaction. We
					 * could check all open documents to see if any are dirty
					 * an will require a "Save changes?" alert, but that is
					 * not general enough. There may be other director subclasses
					 * or other reasons why the app may require interaction, and
					 * we can't test for them here.
					 */
					 
					if (anAppleEvent->GotRequiredParams())
					{
						err = anAppleEvent->RequestInteraction( MAXLONG);
						if (err == noErr)
						{
							Quit();
						
							// if we're still running then Quit didn't succeed.
							// but we don't know what error to report!
						}
						anAppleEvent->SetErrorResult( err);
					}
					break;
		}		
	}
}
