/******************************************************************************
 CControl.c

							The Control Class
		
	Abstract class for objects based on a Mac control (buttons, check boxes,
	radio buttons, scroll bars, and custom CDEF's). This class handles
	most of the interaction with the Mac Control Manager.
	
	SUPERCLASS = CPane
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- Modified DrawAll to support drawing pane borders
		- SetValue now sends a BroadcastChange message to notify dependants
		- implemented overrride of RefreshLongRect
	]

 ******************************************************************************/

#include "Global.h"
#include "CControl.h"
#include "CEnvironment.h"

/*** Global Variables ***/

extern RgnHandle	gUtilRgn;			/* Utility region					*/


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 Dispose {OVERRIDE}

		Initialize a Control object
 ******************************************************************************/

void	CControl::Dispose()
{
	DisposeControl(macControl);			/* Get rid of control record		*/
	macControl = NULL;
	inherited::Dispose();				/* Pass message on to superclass	*/
}


/**** A C C E S S I N G   M E T H O D S ****/


/******************************************************************************
 SetValue

		Specify the value of a Control
 ******************************************************************************/

void	CControl::SetValue(
	short		aValue)
{
	Prepare();
	SetCtlValue(macControl, aValue);	/* Set value with Toolbox call		*/
	
	BroadcastChange( controlValueChanged, &aValue); /* notify dependents		*/
}
	

/******************************************************************************
 GetValue

		Return the value of a Control
 ******************************************************************************/

short	CControl::GetValue()
{
	return(GetCtlValue(macControl));	/* Return value from Toolbox call	*/
}
	

/******************************************************************************
 SetMaxValue

		Specify the maximum value of a Control
 ******************************************************************************/

void	CControl::SetMaxValue(
	short		aValue)
{
	Prepare();
	SetCtlMax(macControl, aValue);		/* Set value with Toolbox call		*/
}
	

/******************************************************************************
 GetMaxValue

		Return the maximum value of a Control
 ******************************************************************************/

short	CControl::GetMaxValue()
{
	return(GetCtlMax(macControl));		/* Return value from Toolbox call	*/
}
	

/******************************************************************************
 SetMinValue

		Specify the minimum value of a Control
 ******************************************************************************/

void	CControl::SetMinValue(
	short		aValue)
{
	Prepare();
	SetCtlMin(macControl, aValue);		/* Set value with Toolbox call		*/
}
	

/******************************************************************************
 GetMinValue

		Return the minimum value of a Control
 ******************************************************************************/

short	CControl::GetMinValue()
{
	return(GetCtlMin(macControl));		/* Return value from Toolbox call	*/
}
	

/******************************************************************************
 SetTitle

		Specify the title of a Control
 ******************************************************************************/

void	CControl::SetTitle(
	Str255		aTitle)					/* Title for the Control			*/
{
	Prepare();
	SetCTitle(macControl, aTitle);
}
	

/******************************************************************************
 GetTitle

		Return the title of a Control
 ******************************************************************************/

void	CControl::GetTitle(
	Str255		aTitle)					/* Title of the Control				*/
{
	GetCTitle(macControl, aTitle);
}
	

/******************************************************************************
 SetActionProc

		Specify the action procedure which is called continuously while
		the mouse is held down within a part of the Control. Note that
		the action procedure must be declared as type "pascal".
		
		For example, for a part other than a moving indicator:
		
			pascal void MyAction(macControl, whichPart)
			ControlHandle	macControl;
			short			whichPart;
			{
				...
			}
			
		The action procedure has no parameters if it is to be called for
		a part which is a moving indicator (thumb).
		
		The DoClick method for this class does NOT use the action procedure
		for moving indicators (is uses it only for stationary parts). If
		you wish to use an action procedure for a moving indicator, you
		must override the DoClick method.
 ******************************************************************************/

void	CControl::SetActionProc(
	VoidFunc		anActionProc)		/* Pointer to action procedure		*/
{
	SetCtlAction( macControl, (ProcPtr) anActionProc);
}


/**** M A N I P U L A T I N G   M E T H O D S ****/
	
/******************************************************************************
 Show {OVERRIDE}

		Show a Control
 ******************************************************************************/

void CControl::Show(void)
{
	Rect	r;
	
	if (!visible)
	{
		Prepare();
		HidePen();						/* Tell Mac Control Manager to hide	*/
		ShowControl(macControl);		/*   control but suppress the		*/
		ShowPen();						/*   automatic drawing				*/
		
		r = (**macControl).contrlRect;
		InvalRect( &r);
		
		visible = TRUE;
		ForceNextPrepare();
	}
}

/******************************************************************************
 Hide {OVERRIDE}

		Hide a Control
 ******************************************************************************/

void	CControl::Hide(void)
{
	Rect r;
	
	if (visible) {
		Prepare();
		HidePen();						/* Tell Mac Control Manager to hide	*/
		HideControl(macControl);		/*   control but suppress the		*/
		ShowPen();						/*   automatic drawing				*/
		
		r = (**macControl).contrlRect;
		InvalRect( &r);
		
		visible = FALSE;
		ForceNextPrepare();
	}
}
	

/******************************************************************************
 Activate {OVERRIDE}

		Activate a Control
 ******************************************************************************/

void	CControl::Activate()
{
	Rect			tempRect;

	if (!active) {
		active = TRUE;
		ForceNextPrepare();
		
		Prepare();						/* Tell Mac Control Manager to		*/
		HidePen();						/*   activate the control but		*/
		HiliteControl(macControl, 0);	/*   suppress the automatic drawing	*/
		ShowPen();
	
		FrameToWindR( &aperture, &tempRect);	/* Draw active control				*/
		DrawAll(&tempRect);
		
										/* Don't redraw if there's a		*/
		ValidRect(&tempRect);			/*   pending update event			*/
	}
}
	

/******************************************************************************
 Deactivate {OVERRIDE}

		Deactivate a Control
 ******************************************************************************/

void	CControl::Deactivate()
{
	Rect			tempRect;

	if (active) {
		active = FALSE;
		ForceNextPrepare();

		Prepare();						/* Tell Mac Control Manager to		*/
		HidePen();						/*   deactivate the control but		*/
		HiliteControl(macControl, 255);	/*   suppress automatic drawing		*/
		ShowPen();
		
		FrameToWindR( &aperture, &tempRect);	/* Draw inactive control			*/
		DrawAll(&tempRect);
	}
}


/******************************************************************************
 Offset {OVERRIDE}

		Move the location of a Control. Superclass does the work for
		our data structures (frame, aperture, etc.). Here we have to
		deal with the Mac Control Manager to get the Toolbox to offset
		the location of the control.
 ******************************************************************************/

void	CControl::Offset(
	long		hOffset,				/* Pixels to offset horizontally	*/
	long		vOffset,				/* Pixels to offset vertically		*/
	Boolean		redraw)					/* Do we need to redraw control?	*/
{
	Point				wOffset;		/* Translation from global to local	*/
	LongRect			interior;		/* interior of the pane */

										/* Superclass does some of the work	*/
	inherited::Offset(hOffset, vOffset, redraw);

										/* Now we have to deal with the Mac	*/
										/*   Control Manager stuff			*/
										
	SetPort(macPort);					/* Set up port to Window coords		*/
	SetOrigin(0,0);						/*   'cause that's what the Control	*/
										/*   Manager uses					*/
	
	if (!redraw) {						/* If we don't want to redraw the	*/
										/*   Control, we must save the		*/
										/*   current update region because	*/
										/*   the Control Mgr will change it	*/		
		CopyRgn(((WindowPeek)macPort)->updateRgn, gUtilRgn);
	}

	HidePen();							/* Suppress the automatic drawing	*/
										/*   performed by the Toolbox		*/

	if (redraw) {						/* Hide control for now. Draw		*/
		HideControl(macControl);		/*   method will show it again		*/
	}
										/* Offset Control to new location	*/
	GetInterior( &interior);
	MoveControl(macControl, -(short) hOrigin + interior.left,
				-(short) vOrigin + interior.top);
	
	ShowPen();							/* Turn drawing back on				*/
	
	if (!redraw) {						/* MoveControl automatically adds	*/
										/*   the old location to the update	*/
										/*   region. If we don't want to	*/
										/*   redraw, we must restore the	*/
										/*   original update region.		*/
		
		ValidRect(&macPort->portRect);	/* Clear update region				*/
		
		wOffset.h = wOffset.v = 0;		/* Update region is in global		*/
		GlobalToLocal(&wOffset);		/*   coords. Convert to Window		*/
										/*   coords by finding the proper	*/
										/*   offset from global to local	*/
		OffsetRgn(gUtilRgn, wOffset.h, wOffset.v);
		InvalRgn(gUtilRgn);				/* Restore saved updated region		*/
		
										/* ??? We could just copy directly	*/
										/*   into the updateRgn field of	*/
										/*   the window record, but this	*/
										/*   could be bad Toolbox Karma		*/
										/*   (see TechNote 227). It's safer	*/
										/*   to treat fields in Toolbox		*/
										/*   records as read-only.			*/
	}
	ForceNextPrepare();
}


/******************************************************************************
 ChangeSize {OVERRIDE}

		Change the size of a Control. Superclass handles the work for
		our data structures. Here we have to deal with the Mac Control
		Manager to get it to change the size, and possible position,
		of the control.
 ******************************************************************************/

void	CControl::ChangeSize(
	Rect		*delta,					/* Change in position of each side	*/
	Boolean		redraw)					/* Redraw to reflect change?		*/
{
	Point				wOffset;		/* Translation from global to local	*/
	LongRect			interior;		/* interior of the pane */

										/* Superclass does some of the work	*/
	inherited::ChangeSize(delta, redraw);

										/* Now we have to deal with the Mac	*/
										/*   Control Manager stuff			*/
										
	SetPort(macPort);					/* Set up port to Window coords		*/
	SetOrigin(0,0);						/*   'cause that's what the Control	*/
										/*   Manager uses					*/
	
	if (!redraw) {						/* If we don't want to redraw the	*/
										/*   Control, we must save the		*/
										/*   current update region because	*/
										/*   the Control Mgr will change it	*/		
		CopyRgn(((WindowPeek)macPort)->updateRgn, gUtilRgn);
	}
	
	HidePen();							/* Suppress the automatic drawing	*/
										/*   performed by the Toolbox		*/

	if (redraw) {						/* Hide control for now. Draw		*/
		HideControl(macControl);		/*   method will show it again		*/
	}
										/* Changing size means setting new	*/
										/*   top left location and sizing	*/
										/*   to the new width and height	*/
	GetInterior( &interior);
	
	MoveControl(macControl, -(short) hOrigin + interior.left,
				-(short) vOrigin + interior.top);
	SizeControl(macControl, interior.right - interior.left,
							interior.bottom - interior.top);
	
	ShowPen();							/* Turn drawing back on				*/
	
	if (!redraw) {						/* MoveControl automatically adds	*/
										/*   the old location to the update	*/
										/*   region. If we don't want to	*/
										/*   redraw, we must restore the	*/
										/*   original update region.		*/
		
		ValidRect(&macPort->portRect);	/* Clear update region				*/
		
		wOffset.h = wOffset.v = 0;		/* Update region is in global		*/
		GlobalToLocal(&wOffset);		/*   coords. Convert to Window		*/
										/*   coords by finding the proper	*/
										/*   offset from global to local	*/
		OffsetRgn(gUtilRgn, wOffset.h, wOffset.v);
		InvalRgn(gUtilRgn);				/* Restore saved updated region		*/
		
										/* ??? We could just copy directly	*/
										/*   into the updateRgn field of	*/
										/*   the window record, but this	*/
										/*   could be bad Toolbox Karma		*/
										/*   (see TechNote 227). It's safer	*/
										/*   to treat fields in Toolbox		*/
										/*   records as read-only.			*/
	}
	ForceNextPrepare();
}


/**** D R A W I N G   M E T H O D S ****/



/******************************************************************************
 DrawAll {OVERRIDE} TCL 1.1.1 DLP 9/18/91

		Draw a Control
 ******************************************************************************/

void	CControl::DrawAll(
	Rect		*area)					/* Portion which must be drawn		*/
{
	Rect	wAperture;
		
	Prepare();
	
	if (!printing) {					/* Control Manager works in Window	*/
										/*   coordinates					*/
			
		FrameToWindR( &aperture, &wAperture);
		SectRect( area, &wAperture, &wAperture);
		ClipRect(&wAperture);
		
		Draw( area);
	}
	else
	{
			// temporarily change the control's port so it draws to 
			// the printer rather than onscreen
			
		(**macControl).contrlOwner = thePort;
		Draw( area);
		(**macControl).contrlOwner = macPort;
	}
}


/******************************************************************************
 Prepare {OVERRIDE} TCL 1.1.1 DLP 9/18/91

		Prepare for drawing a Control. Port must be set to Window coords
		(origin at 0,0) in order for the Mac Control Manager to draw
		properly. Thus we must convert the aperture to Window coordinates
		in order to set the proper clipping region.
 ******************************************************************************/

void	CControl::Prepare()
{
	Rect			wAperture;			/* Aperture in Window coords		*/

	CView::cPreparedView = this;
	
	if (printing)
		PrepareToPrint();
	else
	{
	
		SetPort(macPort);					 /* Use the Port of this Control	*/
		SetOrigin(0, 0);					 /* Must use origin of (0, 0)		*/
	
											 /* Set clipping region to the		*/
		FrameToWindR( &aperture, &wAperture);/*   Control's aperture expressed	*/
		ClipRect(&wAperture);				 /*   in Window coordinates			*/
	}
	
	if (itsEnvironment != NULL)
		itsEnvironment->Restore();		/* Restore its drawing environment	*/
}

/******************************************************************************
 PrepareToPrint {OVERRIDE}	TCL 1.1.1 DLP 9/18/91

		Set up coordinate system and clipping region for printing.
 ******************************************************************************/

void	CControl::PrepareToPrint()
{
	Rect		clipR;

	SetOrigin( 0, 0);

		// get the clipping rectangle in Window coordinates
				
	switch (printClip) {
	
		case clipAPERTURE:
			FrameToWindR( &aperture, &clipR);
			break;
			
		case clipFRAME:
			FrameToWindR( &frame, &clipR);
			break;
			
		case clipPAGE:
			clipR = cPageArea;
			break;
	}
	
	ClipRect(&clipR);
}

/******************************************************************************
 Draw {OVERRIDE}

		Draw a Control
 ******************************************************************************/

void	CControl::Draw(
	Rect		*area)					/* Portion which must be drawn		*/
{	
	LongRect	interior;
	
	if ((**macControl).contrlVis != 255) {
		
			/* TRICKY - A control may actually be invisible on screen	*/
			/* even when the visible instance variable is TRUE. We do	*/
			/* this to circumvent the automatic redraw which occurs		*/
			/* when the Toolbox calls to move or resize a control are	*/
			/* executed. We hide the control before the move or resize.	*/
			/* This Draw method will be called during the resulting		*/
			/* update event, so we can just show the control. With this	*/
			/* ruse, we ensure that all drawing is perform by this		*/
			/* method only. Drawing during other methods may cause ugly	*/
			/* flashes (the control is drawn, erased, then drawn again)	*/
			/* or mess up the layering of overlapping panes (the panes	*/
			/* may be drawn out of order).								*/
			/*															*/
			/* We also move the control to the proper location first.	*/
			/* This saves us the trouble of moving a control when it's	*/
			/* not visible. In fact, if the control is within a Pane	*/
			/* which is larger than the QuickDraw coordinate system, it	*/
			/* will not	always be possible to move the control to the	*/
			/* proper hidden location. The methods for handling window	*/
			/* and pane updates ensure that this Draw method is invoked	*/
			/* only when the Control is actually visible on screen.		*/
		
		GetInterior( &interior);
		MoveControl(macControl, -(short) hOrigin + interior.left,
					 -(short) vOrigin + interior.top);
		ShowControl(macControl);
	} else {
		Draw1Control(macControl);
	}
}

		
/**** C L I C K   R E S P O N S E   M E T H O D S ****/


/******************************************************************************
 DoClick {OVERRIDE}

		Handle a mouse click inside a Control. Note that this method does
		NOT use the action procedure when the mouse is held down within a
		moving indicator.
 ******************************************************************************/

void	CControl::DoClick(
	Point		hitPt,					/* Mouse location in Frame coords	*/
	short		modifierKeys,			/* State of modifier keys			*/
	long		when)					/* Tick time of mouse click			*/
{
	register short	whichPart;			/* Part of control clicked in		*/
	short			origValue;			/* Original control setting			*/
	short			delta;				/* Change in control setting		*/
	LongPt			tmp;
	
	QDToLongPt( hitPt, &tmp);
	FrameToWind( &tmp, &hitPt);

										/* Find part of control clicked in	*/
										/*   Note that a part code of zero	*/
										/*   means that the control is		*/
										/*   inactive, so we do nothing		*/
	whichPart = TestControl( macControl, hitPt);
	
	PenNormal();
	
	if (whichPart >= inThumb) {			/* Click inside moving indicator	*/
										/* Save the original setting		*/
		origValue = GetCtlValue(macControl);
										/* Track the mouse but do NOT call	*/
										/*   a continuous action procedure	*/
		TrackControl(macControl, hitPt, NULL);
		
										/* Respond if the setting changed	*/
		if ( (delta = GetCtlValue(macControl) - origValue) != 0 )
			DoThumbDragged(delta);
		
	} else if (whichPart > 0) {			/* Click inside stationary part		*/
	
		/* The parameter of -1L	means that the actionProc in the	*/
		/* Toolbox control record will be called continuously while	*/
		/* the mouse is down.  If actionProc is NULL, no additional	*/
		/* action is performed.										*/

		if (TrackControl(macControl, hitPt, (ProcPtr)-1L))
		
				/* TrackControl returns the part code if user	*/
				/* executed a good click, i.e., the mouse was	*/
				/* pressed and released in the same part. If so	*/
				/* respond to the click; otherwise, do nothing.	*/

			DoGoodClick(whichPart);
	}
}
		

/******************************************************************************
 DoThumbDragged

		Make appropriate adjustments when the thumb of the control
		has been dragged. This method should be overridden for
		controls which have a thumb.
 ******************************************************************************/

void	CControl::DoThumbDragged(
	short		delta)					/* Change in control setting		*/
{
}										/* Null Method						*/
		

/******************************************************************************
 DoGoodClick

		Make appropriate adjustments after a good click (mouse
		pressed and released in the same part) in a part of a Control.
		This method should be overridden.
 ******************************************************************************/

void	CControl::DoGoodClick(
	short		whichPart)				/* Part code of clicked part		*/
{
}										/* Null Method						*/

/******************************************************************************
 RefreshLongRect

		Force a redraw of an area by adding it to the update region
		of its window. Drawing will not actually take place until the next
		update event is processed. Only the portion of the area that
		intersects the pane's aperture is invalidated.
 ******************************************************************************/

void CControl::RefreshLongRect( LongRect *area)
{
	Rect	visArea;
	
	if (ReallyVisible() && SectAperture( area, &visArea))
	{
		OffsetRect( &visArea, -hOrigin, -vOrigin);
		InvalRect( &visArea);
	}
}	

