/******************************************************************************
 CDesktop.c

							The Desktop Class
	
	View which encompasses the entire screen. There is typically (always?)
	only one instance of CDesktop (or a subclass) and it is stored as a
	global variable (gDesktop). The Desktop at the top of the visual
	hierarchy -- it encloses all other views and has no enclosure. The
	Desktop manages windows and has its own QuickDraw grafPort which may
	be used to draw directly on the screen (for example, to implement "zoom"
	rectangles to animate the opening and closing of windows, as done by
	the Finder).
		
	SUPERCLASS = CView
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- added support for modal windows.
	]

 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "Commands.h"
#include "CDesktop.h"
#include "CDirector.h"
#include "CApplication.h"
#include "CBartender.h"
#include "TBUtilities.h"
#include "CList.h"
#include "CWindow.h"

#include <Script.h>

/*** Global Variables ***/

extern CBartender	*gBartender;		/* Manages all menus				*/
extern CBureaucrat	*gGopher;			/* First in line to get commands	*/
extern Boolean		gInBackground;		/* In background under MultiFinder	*/


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/

extern Boolean CrsrVis : 0x8CC;	// TRUE if cursor is visible

/******************************************************************************
 IDesktop

		Initialize a Desktop object
 ******************************************************************************/

void	CDesktop::IDesktop(
	CBureaucrat		*aSupervisor)		/* Supervisor in the command chain	*/
{
	RgnHandle grayRgn;
	
	CView::IView(NULL, aSupervisor);	/* Initialize superclass			*/
	
										/* Allocate space for a QD port		*/
	macPort = (GrafPtr) NewPtr(sizeof(GrafPort));
	OpenPort(macPort);					/* Initialize QuickDraw port		*/
	
		/* Make the port rectangle coincide with the bounding box of	*/
		/* the gray region (entire screen excluding the menu bar). Also	*/
		/* adjust the origin so that the local coords of this port are	*/
		/* the same as global coords.									*/
	grayRgn = GetGrayRgn();
	bounds = (**grayRgn).rgnBBox;
	PortSize(bounds.right - bounds.left, bounds.bottom - bounds.top);
	MovePortTo(bounds.left, bounds.top);
	SetOrigin(bounds.left, bounds.top);
	
	SetClip(grayRgn);					/* Make the clipping and visible	*/
										/*   regions same as the grayRgn	*/
	CopyRgn(grayRgn, macPort->visRgn);

	visible = TRUE;
	active = TRUE;
	wantsClicks = TRUE;
	
	itsWindows = new(CList);			/* Initialize our window list		*/
	itsWindows->IList();
	topWindow = NULL;
}
	

/******************************************************************************
 Dispose

		Throw out objects created by a Desktop: GrafPort and window list
 ******************************************************************************/

void	CDesktop::Dispose()
{
	itsWindows->DisposeAll();
	itsWindows = NULL;
	
	ClosePort(macPort);
	DisposPtr( (Ptr) macPort);
	macPort = NULL;
	
	inherited::Dispose();
}
	

/******************************************************************************
 DispatchClick {OVERRIDE}

		Respond to a mouse down. Determine where the mouse down occurred
		and send the appropriate messages to the affected objects. The
		where field of macEvent is in Global coordinates.
 ******************************************************************************/

void	CDesktop::DispatchClick(
	register EventRecord	*macEvent)	/* Mouse down event record			*/
{
	short					thePart;	/* Location of mouse click			*/
	WindowPeek				macWindow;	/* Window where click occurred		*/
	register CWindow		*theWindow;	/* Corresponding window object		*/
	register long			menuChoice;	/* Selection from a menu			*/
		
	thePart = FindWindow(macEvent->where, (WindowPtr*)&macWindow);
	
	if (macWindow != NULL) 
	{
		if (macWindow->windowKind != OBJ_WINDOW_KIND)	/* TCL 1.1.2 DLP 12/13/91 */
			return;
		theWindow = (CWindow*) GetWRefCon((WindowPtr) macWindow);
	}
	else theWindow = NULL;
	
	/*
	 *	Check for any click the could potentially deactivate a window
	 *	that wants to be modal. Menu bar clicks are OK because
	 *	menus should be accessible from modal dialogs.
	 */
	 
	if (topWindow && (theWindow != topWindow) && topWindow->IsModal())
	{
		if (thePart != inMenuBar)
		{
			SysBeep( 3);
			return;
		}
	}

	switch (thePart) {					/* Take appropriate action			*/
	
		case inDesk:					/* Click on insignificant area of	*/
										/*   the Desktop					*/
			CountClicks(this, macEvent);
			DoClick(macEvent->where, macEvent->modifiers, macEvent->when);
			break;
		
		case inMenuBar:					/* Pull-down menus					*/
			gBartender->UpdateAllMenus();
			menuChoice = MenuSelect(macEvent->where);
			if (HiShort(menuChoice)) {	/* A selection was made. Execute	*/
										/*   the corresponding command.		*/
				gGopher->DoCommand(gBartender->FindCmdNumber
									(HiShort(menuChoice), LoShort(menuChoice)));
				HiliteMenu(0);			/* Unhighlight the menu title		*/
			}
			break;
			
		case inSysWindow:				/* Click inside a DA window			*/
			SystemClick(macEvent, (WindowPtr) macWindow);
			break;
			
		case inContent:
			if (!theWindow->active || ( (theWindow->floating &&
								((WindowPtr)macWindow != FrontWindow())))) {
										/* This is an inactive window or	*/
										/*   an underlying floater			*/
				theWindow->Select();
				
				if (!theWindow->actClick) {
					break;				/* Exit here unless window want to	*/
				}						/*   process activating click		*/
                
                
                else
                    theWindow->Activate();
			}
			
			if (theWindow->wantsClicks) {
				UpdateWindows();		/* Window handles clicks. Make sure	*/
										/*   all windows are updated before	*/
										/*   responding to the click.		*/
				theWindow->DispatchClick(macEvent);
				
			} else {					/* Window ignores clicks. Handle	*/
										/*   same as a click on the desktop	*/
				CountClicks(this, macEvent);
				DoClick(macEvent->where, macEvent->modifiers, macEvent->when);
			}
			break;
			
		case inDrag:
			theWindow->Drag(macEvent);
			break;
			
		case inGrow:
			theWindow->Resize(macEvent);
			break;
			
		case inGoAway:
			if (TrackGoAway((WindowPtr) macWindow, macEvent->where))
				theWindow->Close();
			break;
			
		case inZoomIn:
		case inZoomOut:
			if (TrackBox( (WindowPtr)macWindow, macEvent->where, thePart))
				theWindow->Zoom(thePart);
			break;
	}
}
	

/******************************************************************************
 DoMouseUp {OVERRIDE}

		Respond to the mouse button being release. By default, ignore it.
 ******************************************************************************/

void	CDesktop::DoMouseUp(
	EventRecord		*macEvent)			/* Mouse up event record			*/
{
}										/* Null Method						*/


/******************************************************************************
 DispatchCursor

		Adjust the shape of the cursor depending on the location of the mouse.
 ******************************************************************************/

void	CDesktop::DispatchCursor(
	Point		where,					/* Mouse location in Global coords	*/
	RgnHandle	mouseRgn)				/* Region containing the mouse		*/
{
	WindowPeek		macWindow;			/* Window containing the mouse		*/
	CWindow			*theWindow;			/* Corresponding window object		*/
	short			thePart;			/* Part of window containing mouse	*/
	Rect			mbarRect;
	
	if (gInBackground || !CrsrVis) {	/* Do nothing if we are not the		*/
		return;							/*   foreground process				*/
										/* or the cursor is hidden			*/
	}
	
	CView::cCurrHelpView = NULL;
										/* Determine which window contains	*/
										/*   the mouse						*/
	thePart = FindWindow(where, (WindowPtr*) &macWindow);
	
	if ( (thePart == inContent) || (thePart == inGrow) ) 
	{
										/* Mouse is within the content		*/
										/*   region of a window. Check if	*/
										/*   that window is active.			*/
			/* TCL 1.1.2 DLP 12/13/91 */
			if (macWindow->windowKind == OBJ_WINDOW_KIND)
				theWindow = (CWindow*) GetWRefCon((WindowPtr)macWindow);
			else
				theWindow = NULL;
				
			if (theWindow && theWindow->IsActive()) 
			{							/* Window containing the mouse is	*/
										/*   active. Tell it to adjust the	*/
										/*   cursor shape.					*/
				theWindow->DispatchCursor(where, mouseRgn);
			}
			else						/* cursor over inactive window 		*/
				AdjustCursor( where, mouseRgn); 
	}
	else if (thePart == inMenuBar) 
	{
		SetCursor(&arrow);				/* Use arrow cursor in menu bar		*/
		mbarRect = screenBits.bounds;
		mbarRect.bottom = mbarRect.top + GetMBarHeight();
		RectRgn(mouseRgn, &mbarRect);
	}
	else
	{
		AdjustCursor(where, mouseRgn);	/* All checks have failed. Mouse is	*/
										/* in some insignificant area of	*/
										/* the Desktop						*/
	}
	CView::cLastHelpView = CView::cCurrHelpView;		
}	


/******************************************************************************
 AdjustCursor {OVERRIDE}

		Mouse is located where no other object wants to set the cursor shape.
		Adjust the shape to the standard arrow cursor
 ******************************************************************************/

void	CDesktop::AdjustCursor(
	Point		where,					/* Mouse location in global coords	*/
	RgnHandle	mouseRgn)				/* Region containing the mouse		*/
{
	WindowPeek	macWindow;
	
	SetCursor(&arrow);
	
	if (topWindow != NULL) {			/* A Window is active, so subtract	*/
										/*   its content region from the	*/
										/*   desktop						*/
		macWindow = (WindowPeek) topWindow->GetMacPort();
		DiffRgn(GetGrayRgn(), ((WindowPeek)topWindow->GetMacPort())->contRgn,
					mouseRgn);
		
	} else {							/* There's no active window, so		*/
										/*   cursor has the same shape over	*/
										/*   the entire desktop				*/
		CopyRgn(GetGrayRgn(), mouseRgn);
	}
}
	

/******************************************************************************
 Show {OVERRIDE}

		Make a Desktop visible by showing all its windows
 ******************************************************************************/

		/*
		 *	Show a window
		 */
		void	Window_Show(
			CWindow		*theWindow)
		{
			theWindow->Show();
		}

void	CDesktop::Show()
{
	if (!visible) {
		visible = TRUE;
		itsWindows->DoForEach((EachFunc) Window_Show);
	}
}
	

/******************************************************************************
 Hide {OVERRIDE}

		Make a Desktop invisible by hiding all its windows
 ******************************************************************************/

		/*
		 *	Hide a window
		 */
		void	Window_Hide(
			CWindow		*theWindow)
		{
			theWindow->Hide();
		}

void	CDesktop::Hide()
{
	if (visible) {
		itsWindows->DoForEach((EachFunc) Window_Hide);
		visible = FALSE;
		if (gGopher == this)
			itsSupervisor->BecomeGopher( TRUE);
	}
}
	

/******************************************************************************
 Activate {OVERRIDE}

		Activate a Desktop by activating its top window
 ******************************************************************************/

void	CDesktop::Activate()
{
	if (!active) {
		active = TRUE;
		if (!BecomeGopher( TRUE))
			return;
			
		if (topWindow != NULL)
			topWindow->Activate();
	}
}
	

/******************************************************************************
 Deactivate {OVERRIDE}

		Deactivate a Desktop by deactivating its top window
 ******************************************************************************/

void	CDesktop::Deactivate()
{
	if (active) {
		active = FALSE;
		if (topWindow != NULL)
			topWindow->Deactivate();
		if (gGopher == this)
			itsSupervisor->BecomeGopher( TRUE);
	}
}
	

/******************************************************************************
 ReallyVisible {OVERRIDE}

		Check whether a Desktop is visible. Desktop has no enclosure.
 ******************************************************************************/

Boolean	CDesktop::ReallyVisible()
{
	return(visible);
}
	

/******************************************************************************
 AddWind

		Add a window to the desktop
 ******************************************************************************/

void	CDesktop::AddWind(
	CWindow		*theWindow)				/* Window object to add				*/
{
	itsWindows->Add(theWindow);
}
	

/******************************************************************************
 RemoveWind

		Remove a window from the desktop
 ******************************************************************************/

void	CDesktop::RemoveWind(
	CWindow		*theWindow)				/* Window object to remove			*/
{
	if (theWindow->IsVisible()) {
		HideWind(theWindow);			/* Hide window before removing it	*/
	}
	itsWindows->Remove(theWindow);
}
	

/******************************************************************************
 SelectWind

		Select a window by bringing it to the front and making it active
 ******************************************************************************/

void	CDesktop::SelectWind(
	register CWindow	*theWindow)		/* Window to select					*/
{
	if (theWindow == topWindow) {		/* This is already the top window	*/
		if (IsSystemWindow((WindowPeek)FrontWindow())) {
										/* A DA is on top					*/
										/*   Make this the active window	*/
			SelectWindow(theWindow->macPort);
		}
	
	} else {							/* This is not the current top one	*/
										/* Move this one to the front		*/
		if (!theWindow->IsVisible()) {
			BringToFront(theWindow->macPort);
			ShowWindow(theWindow->macPort);
		} else {
			SelectWindow(theWindow->macPort);
		}

		if (topWindow != NULL) {		/* Deactivate the former top		*/
			topWindow->Deactivate();	/*   regular window					*/
			((CDirector*)topWindow->itsSupervisor)->active = FALSE;
		}
		itsWindows->BringFront(theWindow);
		topWindow = theWindow;			/* Mark it as the top regular		*/
		if (gInBackground) {
			((CDirector*)topWindow->itsSupervisor)->Activate();
		}
	}
}
	

/******************************************************************************
 ShowWind

		Make a window visible on the desktop
 ******************************************************************************/

		/*
		 *	Check if a window is visible
		 */
		Boolean		Window_IsVisible(
			CWindow		*theWindow)
		{
			return theWindow->IsVisible();
		}

void	CDesktop::ShowWind(
	CWindow		*theWindow)				/* Window to show					*/
{
	CWindow		*oldTop;

	ShowHide(theWindow->macPort, TRUE);	/* Make window visible				*/
	
	oldTop = topWindow;					/* There may be a new top window	*/
	topWindow = (CWindow*) itsWindows->FirstSuccess((TestFunc) Window_IsVisible);
	
	if (theWindow == topWindow) {		/* This is the new top window		*/
		if (oldTop != NULL) {			
			oldTop->Deactivate();		/*   Deactivate former top window	*/
		}
		topWindow->Activate();			/*   Activate this window			*/
	}
}
	

/******************************************************************************
 HideWind

		Make a window invisible on the desktop
 ******************************************************************************/

void	CDesktop::HideWind(
	CWindow		*theWindow)				/* Window to hide					*/
{
	WindowPtr	frontWind;
	
	ShowHide(theWindow->macPort, FALSE);
	theWindow->Deactivate();
	
	if (theWindow == topWindow) {		/* This was the top window			*/
										/* Find the new top top window		*/
		topWindow = (CWindow*) itsWindows->FirstSuccess((TestFunc) Window_IsVisible);

		frontWind = FrontWindow();
		if (IsSystemWindow((WindowPeek)frontWind)) {
										/* A DA is now on top				*/
			SelectWindow(frontWind);
			CurActivate = frontWind;
			
			if (topWindow != NULL) {	/* Set internal flag so our new top	*/
										/*   window will become active if	*/
										/*   the DA quits					*/
				((CDirector*)topWindow->itsSupervisor)->active = TRUE;
			}
			
		} else if (topWindow != NULL) {	/* Our new top window needs to be	*/
			topWindow->Activate();		/*   activated						*/
		}
	}
}


/******************************************************************************
 DragWind

		Drag a window on the desktop
 ******************************************************************************/

void	CDesktop::DragWind(
	CWindow		*theWindow,				/* Window to drag					*/
	EventRecord	*macEvent)				/* Associated mouse down event		*/
{
	Rect		dragBox;				/* Where window may be dragged		*/
	
										/* For an underlying window, bring	*/
										/*   it to the front (unless the	*/
										/*   command key is down) before	*/
										/*   dragging it					*/
	if ((theWindow != topWindow) && !(macEvent->modifiers & cmdKey)) {
		SelectWind(theWindow);
	}
	
	dragBox = bounds;					/* Let user drag window, but leave	*/
										/*   a margin so window doesn't get	*/
										/*   dragged too far off the screen	*/
	InsetRect(&dragBox, DRAG_MARGIN, DRAG_MARGIN);
	DragWindow(theWindow->macPort, macEvent->where, &dragBox);
}


/******************************************************************************
 UpdateWindows

		Update contents of all windows
 ******************************************************************************/

		/*
		 * Update contents of a window
		 */
		 
		static void		Window_Update(
			CWindow			*theWindow)
		{
			WindowPeek		macWindow;
			
			macWindow = (WindowPeek) theWindow->GetMacPort();
			if ( !EmptyRgn(macWindow->updateRgn) ) {
				theWindow->Update();
			}
		}

void	CDesktop::UpdateWindows()
{
	itsWindows->DoForEach((EachFunc) Window_Update);
}
	

/******************************************************************************
 GetTopWindow

		Return the top most window belonging to a Desktop (does not
		include DA windows)
 ******************************************************************************/

CWindow*	CDesktop::GetTopWindow()
{
	return(topWindow);					/* Return instance variable			*/
}
	

/******************************************************************************
 GetBounds

		Return the bounds of a Desktop
 ******************************************************************************/

void	CDesktop::GetBounds(
	Rect		*theBounds)
{
	*theBounds = bounds;				/* Return copy of instance variable	*/
}

	
/******************************************************************************
 GetAperture {OVERRIDE}

		Return the portion of the Desktop where drawing may occur.
 ******************************************************************************/

void	CDesktop::GetAperture(
	LongRect		*theAperture)
{
	QDToLongRect( &bounds, theAperture);
}

	
/******************************************************************************
 Contains {OVERRIDE}

		Check whether a Desktop contains a specified point. Point is
		specified in global coordinates.
 ******************************************************************************/

Boolean		CDesktop::Contains(
	Point			thePoint)
{
	return(PtInRect(thePoint, &bounds));
}

	
/******************************************************************************
 HitSamePart

		Check whether two hit points are close enough together to be
		part of a multiple click inside the Desktop.
 ******************************************************************************/

Boolean		CDesktop::HitSamePart(
	Point			pointA,
	Point			pointB)
{
	return( (Abs(pointA.h - pointB.h) < REASONABLY_CLOSE)  &&
			(Abs(pointA.v - pointB.v) < REASONABLY_CLOSE) );
}

	
/******************************************************************************
 Prepare {OVERRIDE}

		Set current grafPort to the Desktop's port
 ******************************************************************************/

void	CDesktop::Prepare()
{
	inherited::Prepare();
	SetPort(macPort);
}
