/******************************************************************************
 CDirector.c
	
							The Director Class
	
	Abstract class for an object which directs communication between the
	Application and a Window. Subclasses will probably have additional
	instance variables with references to other objects which need to
	communicate with each other as well as with the Application and
	a Window. Directors are the "department heads" in the chain of
	command. Typically, subclasses of CDirector will override the
	DoCommand method and dispatch messages to carry out the commands to
	the objects it manages.
	
	SUPERCLASS = CBureaucrat
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- implemented GetWindow, FindViewByID methods
	]

 ******************************************************************************/

#include "Global.h"
#include "Commands.h"
#include "CBartender.h"
#include "CDirector.h"
#include "CWindow.h"
#include "CList.h"
#include "CPane.h"

/**** Global Variables ****/

extern CBureaucrat	*gGopher;			/* First in line to get commands	*/
extern CBartender	*gBartender;		/* Manages all menus				*/
extern long			gSleepTime;			/* Max time between events			*/

/******************************************************************************
 IDirector

		Initialize a Director object. Subclasses must implement their
		own initialization method which should call this method as
		its first statement. Note that the supervisor of a Director
		MUST be the Application.
 ******************************************************************************/

void	CDirector::IDirector( CDirectorOwner   *aSupervisor)
{
										/* Initialize Superclass			*/
	CDirectorOwner::IDirectorOwner(aSupervisor);
	
	aSupervisor->AddDirector( this);	
	itsWindow = NULL;					/* No window yet					*/
	active = FALSE;						/* Can't be active with no window	*/
	itsGopher = this;					/* Make self Gopher upon activate	*/
	activateWindOnResume = FALSE;
	alreadyClosing = FALSE;				/* TCL 1.1.1 DLP 9/18/91			*/

}
	

/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a Director
 ******************************************************************************/

void	CDirector::Dispose()
{
	if (OwnsWindow( itsWindow))
		ForgetObject( itsWindow);

	((CDirectorOwner*)itsSupervisor)->RemoveDirector(this);
	inherited::Dispose();
}
	

/******************************************************************************
 DoCommand {OVERRIDE}

		Execute a command
 ******************************************************************************/

void	CDirector::DoCommand(
	long		theCommand)
{
	switch (theCommand) {
	
		case cmdClose:
			Close(FALSE);
			break;
			
		default:
			inherited::DoCommand(theCommand);
			break;
	}
}
	

/******************************************************************************
 UpdateMenus {OVERRIDE}

		Perform menu management tasks
 ******************************************************************************/

void	CDirector::UpdateMenus()
{
	inherited::UpdateMenus();

	if (itsWindow || (itsDirectors && (itsDirectors->GetNumItems() > 0)))
	{
		gBartender->EnableCmd(cmdClose);
	}
}


/******************************************************************************
 Activate

		Make a Director active
 ******************************************************************************/

void	CDirector::Activate()
{
	if (!active) {
		active = TRUE;					/* Turn on active flag				*/
	
		if (!itsWindow->IsFloating()) {	/* Set the gopher unless its window	*/
			if (!itsGopher->BecomeGopher( TRUE))	/*   floats				*/
				return;
				
			gSleepTime = 0L;
		}
	}
	((CDirectorOwner*)itsSupervisor)->ActivateDirector( this);
}
	

/******************************************************************************
 Deactivate

		Make a Director inactive
 ******************************************************************************/

void	CDirector::Deactivate()
{
	if (active) {
		active = FALSE;					/* Turn off active flag				*/
		if (gGopher == this) 
		{								/* Inactive Director can't be the	*/
			if (!itsSupervisor->BecomeGopher( TRUE))	/*   gopher			*/
				return;
		}
	}
	((CDirectorOwner*)itsSupervisor)->DeactivateDirector( this);
}
	

/******************************************************************************
 Suspend

		Suspend Director before being swapped out under MultiFinder
 ******************************************************************************/

void	CDirector::Suspend()
{
	inherited::Suspend();
	
	if (active)
	{		
							/* If director is currently active	*/
							/* Visually deactivate its window	*/
							/*   This has the side effect of	*/
							/*   deactivating the Director also	*/
							/* Restore active flag so Director	*/
							/*    will reactivate when the		*/
							/*    application resumes			*/
							
		if (OwnsWindow( itsWindow) && itsWindow->IsActive())
		{
			itsWindow->Deactivate();				
			active = TRUE;
			activateWindOnResume = TRUE;
		}
	}
}
	

/******************************************************************************
 Resume

		Resume Director after being swapped in under MultiFinder
 ******************************************************************************/

void	CDirector::Resume()
{
	inherited::Resume();
	
	if (active)
	{		
		if (OwnsWindow( itsWindow) && activateWindOnResume)
		{		
										/* If we were active before suspend	*/
			active = FALSE;				/*   We're not really active yet	*/
			itsWindow->Activate();		/*   Visually activate window		*/
			activateWindOnResume = FALSE;
		}
	}
}


/******************************************************************************
 Close				TCL 1.1.1 DLP 9/18/91

		Close a Director. Return FALSE if Director doesn't want to close.
 ******************************************************************************/

Boolean		CDirector::Close(
	Boolean			quitting)			/* Is this part of a quit sequence?	*/
{
	Boolean closeSucceeded;
	
		// if we're already in the middle of processing a Close
		// message then we just return failure for the new one.
	
	if (alreadyClosing)
		return FALSE;
		
		// use an exception handler to reset the alreadyClosing
		// flag if an exception occurs.		
	TRY
	{
		alreadyClosing = TRUE;
		if (!inherited::Close( quitting))
		{
			closeSucceeded = FALSE;
			alreadyClosing = FALSE;
		}
		else
		{
			closeSucceeded = TRUE;
			this->Dispose();			/* Closing means Disposing			*/
		}
	}
	CATCH
	{
		alreadyClosing = FALSE;	
	}
	ENDTRY;
	
	return closeSucceeded;
}

/******************************************************************************
 CloseWind

	Close a window owned by this Director. The director is closed if:
		- theWindow was itsWindow and all subdirectors are willing to close, or
		- there are no remaining subdirectors.
		
		
 ******************************************************************************/

void	CDirector::CloseWind(
	CWindow		*theWindow)
{

	if (theWindow == itsWindow)
		Close( FALSE);
	else
	{
		theWindow->Dispose();
		if (!itsWindow && !itsDirectors)
			Close( FALSE);
	}
}

/******************************************************************************
 ActivateWind

		Window owned by this Director has been activated
 ******************************************************************************/

void	CDirector::ActivateWind(
	CWindow		*theWindow)
{
	Activate();
}


/******************************************************************************
 DeactivateWind

		Window owned by this Director has been deactivated
 ******************************************************************************/

void	CDirector::DeactivateWind(
	CWindow		*theWindow)
{
	Deactivate();
}

/******************************************************************************
 ActivateDirector

		Director owned by this Director has been activated
 ******************************************************************************/

void CDirector::ActivateDirector( CDirector *aDirector)
{
	inherited::ActivateDirector( aDirector);
	((CDirectorOwner*)itsSupervisor)->ActivateDirector( this);

}	/* CDirector::ActivateDirector */

/******************************************************************************
 DeactivateDirector

		Director owned by this Director has been deactivated
 ******************************************************************************/

void CDirector::DeactivateDirector( CDirector *aDirector)
{
	inherited::DeactivateDirector( aDirector);
	((CDirectorOwner*)itsSupervisor)->DeactivateDirector( this);

}	/* CDirector::DeactivateDirector */

/******************************************************************************
 IsActive

		Check whether a Director is active
 ******************************************************************************/

Boolean		CDirector::IsActive()
{
	return(active);						/* Return instance variable			*/
}

/******************************************************************************
 GetWindow
 
 	Return the Director's window
******************************************************************************/

CWindow* CDirector::GetWindow( void)
{
	return itsWindow;

}	/* CDirector::GetWindow */

/******************************************************************************
 FindViewByID
 
 	Locate a view within the Director's window
******************************************************************************/

CView* CDirector::FindViewByID( long aViewID)
{	
	return itsWindow? itsWindow->FindViewByID( aViewID) : NULL;

}	/* CDirector::FindViewByID */

/******************************************************************************
 ProviderChanged			TCL 1.1.1 DLP 9/18/91
 
 	Receive notification that a provider changed. Directors want to know
 	if any subordinate has become the gopher in order to keep the instance
 	variable itsGopher up to date.
******************************************************************************/

void CDirector::ProviderChanged( CCollaborator *aProvider, long reason, void* info)
{
	if (active && member( aProvider, CBureaucrat) && (reason == bureaucratIsGopher))
	{	
	
		/*
		 * When a bureaucrat becomes the gopher, it automatically
		 * sends the bureaucratIsGopher update to all its dependents
		 * and up the chain of command. When a director receives this
		 * update, the bureacrat could be a pane in the director's
		 * window, a subdirector, a subdirector's window, a pane in
		 * a subdirector's window, or potentially any other bureaucrat
		 * the application may have placed in the chain of command.
		 *
		 * The purpose of itsGopher is to "remember" which bureaucrat
		 * is supposed to become the gopher when the director's window
		 * is made active. Since subdirectors have their own itsGopher,
		 * and we don't want to set the gopher to panes in windows that
		 * aren't active, we must filter out bureaucrats that we don't
		 * want itsGopher to point to.
		 */
		 
		if (member( aProvider, CDirector))
			; // itsGopher shouldn't be a subdirector
			
		else if (member( aProvider, CPane))
		{
			  // itsGopher should only reference panes in the
			  // directors window(s).
			  
			if (OwnsWindow( ((CPane*)aProvider)->GetWindow()))
				itsGopher = (CBureaucrat*) aProvider;
		}
		else  // this is some other bureaucrat
			itsGopher = (CBureaucrat*) aProvider;
	}
	inherited::ProviderChanged( aProvider, reason, info);

}	/* CDirector::ProviderChanged */

/******************************************************************************
 OwnsWindow
 
 	A Director owns a window if it directly supervises the window. It does
 	not own the window if a subordinate director owns the window
******************************************************************************/

Boolean	CDirector::OwnsWindow( CWindow *aWindow)
{
	return (aWindow && (aWindow->GetSupervisor() == this));

}	/* CDirector::OwnsWindow */

/******************************************************************************
 RemoveDirector {OVERRIDE}
 
 ******************************************************************************/

void CDirector::RemoveDirector( CDirector *aDirector)
{
	inherited::RemoveDirector( aDirector);
	if (!itsWindow && !itsDirectors)
		this->Close( kNotQuitting);
}
