/******************************************************************************
 CDocument.c

							The Document Class
	
	Abstract class which associates a Window with a File. This is the
	typical model used by most Mac applications, where the relevent
	information for an application is displayed to the user in a window
	and stored on disk as a file. Programmers must create their own
	subclass of CDocument if their program supports the standard Mac
	concept of creating a new document ("New" in the File menu) or
	opening an existing document ("Open..." in the File menu).
		
	SUPERCLASS = CDirector
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- added exception handling
		- implemented MakePrinter, DoAppleEvent method.
	]

 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "Commands.h"
#include "CDocument.h"
#include "CBartender.h"
#include "CPrinter.h"
#include "TBUtilities.h"
#include "CPanorama.h"
#include "CApplication.h"
#include "CPane.h"
#include "CFile.h"
#include "CTask.h"
#include "CWindow.h"

/*** Global Variables ***/

extern CBartender	*gBartender;		/* Manages all menus				*/
extern CBureaucrat	*gGopher;			/* First in line to get commands	*/
extern CursHandle	gWatchCursor;		/* Watch cursor for waiting			*/


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IDocument

		Initialize a Document object
		
	TCL 1.1:
		IDocument now checks for NULL return value from CPrinter::GetPrintRecord
		
 ******************************************************************************/

void	CDocument::IDocument(
    CApplication        *aSupervisor, 
	Boolean			printable)			/* Is document printable?			*/
{
	THPrint			macTPrint;
	
	CDirector::IDirector(aSupervisor);	/* Initialize superclass			*/
	
	itsMainPane = NULL;
	itsFile = NULL;
	lastTask = NULL;
	undone = FALSE;
	dirty = FALSE;
	
	if (printable) {
	
		MakePrinter();
		macTPrint = itsPrinter->GetPrintRecord();
		if (macTPrint)
		{
			pageWidth = (**macTPrint).prInfo.rPage.right;
			pageHeight = (**macTPrint).prInfo.rPage.bottom;
		}
		else
		{
			pageWidth = STD_PAGE_WIDTH;
			pageHeight = STD_PAGE_HEIGHT;		
		}
		
	} else {
		itsPrinter = NULL;
		pageWidth = STD_PAGE_WIDTH;
		pageHeight = STD_PAGE_HEIGHT;
	}
}
	

/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a Document
 ******************************************************************************/

void	CDocument::Dispose()
{
	ForgetObject( itsFile);
	ForgetObject( lastTask);
	ForgetObject( itsPrinter);
		
	inherited::Dispose();
}
	

/******************************************************************************
 Notify

		Notification that a subordinate has completed a task
 ******************************************************************************/

void	CDocument::Notify(
	CTask		*theTask)
{
	if (lastTask != NULL) {
		lastTask->Dispose();
	}

	lastTask = theTask;
	undone = FALSE;
	dirty = TRUE;
}
	

/******************************************************************************
 DoCommand {OVERRIDE}

		Execute a command
 ******************************************************************************/

void	CDocument::DoCommand(
	long		theCommand)
{
	switch (theCommand) {
	
		case cmdClose:
			this->Close(FALSE);
			break;
			
		case cmdSave:
			SetCursor(*gWatchCursor);
			DoSave();
			break;
			
		case cmdSaveAs:
			DoSaveFileAs();
			break;
			
		case cmdRevert:
			PositionDialog('ALRT', ALRTrevert);
			if (CautionAlert(ALRTrevert, NULL) == OK) {
				SetCursor(*gWatchCursor);
				DoRevert();
			}
			break;
			
		case cmdPageSetup:
			if (itsPrinter != NULL) {
				itsPrinter->DoPageSetup();
			}
			break;
			
		case cmdPrint:
			if (itsPrinter != NULL) {
				itsPrinter->DoPrint();
			}
			break;
			
		case cmdUndo:
			if (lastTask != NULL) {
				if (undone) {
					lastTask->Redo();
				} else {
					lastTask->Undo();
				}
				undone = !undone;
				UpdateUndo();
			}
			break;
			
		default:
			inherited::DoCommand(theCommand);
			break;
	}
}
	

/******************************************************************************
 UpdateMenus {OVERRIDE}

		Perform menu management tasks
 ******************************************************************************/

void	CDocument::UpdateMenus()
{
	inherited::UpdateMenus();
	
	gBartender->EnableCmd(cmdSaveAs);

	if (dirty) {
		gBartender->EnableCmd(cmdSave);
		if (itsFile != NULL) {
			gBartender->EnableCmd(cmdRevert);
		}
	}
	
	if (itsPrinter != NULL) {
		gBartender->EnableCmd(cmdPageSetup);
		gBartender->EnableCmd(cmdPrint);
	}
	
	if (lastTask != NULL) {
		gBartender->EnableCmd(cmdUndo);
		UpdateUndo();
	}
}


/******************************************************************************
 Close {OVERRIDE}

		Close a Document
 ******************************************************************************/

Boolean	CDocument::Close(
	Boolean		quitting)				/* Is this part of a quit sequence?	*/
{
	if (ConfirmClose(quitting)) {		/* Is it OK to close document?		*/
		if (itsFile != NULL) {			/* Yes. Close associated file.		*/
			itsFile->Close();
		}
	
		inherited::Close(quitting);
		return(TRUE);					/* Document has been closed			*/
		
	} else {							/* User does not want to close		*/
		return(FALSE);					/*   the Document					*/
	}
}
	

/******************************************************************************
 NewFile

		Set up a document with a new file, usually in response the "New"
		command in the File menu.
 ******************************************************************************/

void	CDocument::NewFile()
{
}
	

/******************************************************************************
 OpenFile

		Open a file for a document, usually in response the "Open"
		command in the File menu.
 ******************************************************************************/

void	CDocument::OpenFile(
	SFReply		*macSFReply)
{
}
	

/******************************************************************************
 CloseWind {OVERRIDE}

		A window belonging this document needs to be closed, usually in
		response to a click in its go away box or selecting the "Close"
		command in the File menu.
 ******************************************************************************/

void	CDocument::CloseWind(
	CWindow		*theWindow)
{
	inherited::CloseWind( theWindow); // TCL 1.1.2 DLP 12/9/91
}


/**** P R I N T I N G   M E T H O D S ****/


/******************************************************************************
 PageCount

		Return the number of pages in a Document
 ******************************************************************************/

short	CDocument::PageCount()
{
	short	pageCount;
	short	hStrips, vStrips;
	
	if (itsPrinter)
	{
		if (!itsPrinter->HavePagination())	// TCL 1.1.2 DLP 12/9/91
			Paginate();
			
		itsPrinter->GetStripCount( &hStrips, &vStrips);
		
		if (((long)hStrips) * vStrips > 999L)
			Failure( iPrAbort, excPrintJobTooBig);
			
		pageCount = hStrips * vStrips;
	}
	else
		pageCount = 0;
        
	return pageCount;
}
	
/******************************************************************************
 Paginate

	Determine how many vertical and horizontal strips it will take
	to print the document. The default method passes this message to
	itsMainPane.
 ******************************************************************************/

void CDocument::Paginate( void)
{
	THPrint		macTPrint;

	if (itsPrinter)
	{
		macTPrint = itsPrinter->GetPrintRecord();
		pageWidth = (**macTPrint).prInfo.rPage.right;
		pageHeight = (**macTPrint).prInfo.rPage.bottom;

		if (itsMainPane)
		{
			itsMainPane->Paginate( itsPrinter, pageWidth, pageHeight);
		}
		else
			itsPrinter->ResetPagination();
	}

}	/* CDocument::Paginate */

/******************************************************************************
 AboutToPrint

		The specified range of pages is about to be printed. Make sure
		that the range is valid and make changes if it's not.
		
		This method also gives a Document the chance to perform any necessary
		bookkeeping, such as repaginating, before a print cycle is started.
		This is also the time for the Document to request information about
		the page size from its Printer.
 ******************************************************************************/

void	CDocument::AboutToPrint(
	short		*firstPage,
	short		*lastPage)
{
	THPrint		macTPrint;
	short		swap;

	macTPrint = itsPrinter->GetPrintRecord();
	pageWidth = (**macTPrint).prInfo.rPage.right;
	pageHeight = (**macTPrint).prInfo.rPage.bottom;

	if (*firstPage > *lastPage) {		/* firstPage must be less than or	*/
		swap = *lastPage;				/*   equal to lastPage. Swap values	*/
		*lastPage = *firstPage;			/*   otherwise.						*/
		*firstPage = swap;
	}
										/* Can't print more pages than are	*/
										/*   contained in the document		*/
	*lastPage = Min(*lastPage, PageCount());
	
	if (itsMainPane != NULL) {
		itsMainPane->AboutToPrint(firstPage, lastPage);
	}
}
	

/******************************************************************************
 PrintPageOfDoc

		Print a single page of the document. This method must NOT send
		messages to the Printer object. At this time, the print driver
		is open and waiting for information to be drawn. Requests for
		status information, such as page specifications, will foul up
		the state of the print driver.
 ******************************************************************************/

void	CDocument::PrintPageOfDoc(
	short		pageNum)
{
	if (itsMainPane != NULL) {
		itsMainPane->PrintPage(pageNum, pageWidth, pageHeight, itsPrinter);
	}
}
	

/******************************************************************************
 DonePrinting

		Print loop has been completed
 ******************************************************************************/

void	CDocument::DonePrinting()
{
	if (itsMainPane != NULL) {
		itsMainPane->DonePrinting();
	}
}


/**** F I L I N G   M E T H O D S ****/
	

/******************************************************************************
 DoSave

		Save a file under the same name. Return TRUE if save was successful.
		Subclasses must override this method.
 ******************************************************************************/

Boolean	CDocument::DoSave()
{
}										/* Null Method						*/
	

/******************************************************************************
 DoSaveAs

		Save a file under another name. Return TRUE if save was successful.
		Subclasses must override this method.
 ******************************************************************************/

Boolean	CDocument::DoSaveAs(
	SFReply		*macSFReply)
{
}										/* Null Method						*/
	

/******************************************************************************
 DoRevert

		Throw out the current contents of a document and revert to the
		last saved version. Subclasses must override this method.
 ******************************************************************************/

void	CDocument::DoRevert()
{
}
	

/******************************************************************************
 DoSaveFileAs

		Respond to SaveAs menu command by posting a put file dialog
		box and sending the DoSaveAs message if confirmed.
		Return TRUE if save was successful.
 ******************************************************************************/

Boolean	CDocument::DoSaveFileAs()
{
	SFReply		macSFReply;				/* Standard File reply record		*/
	
	PickFileName(&macSFReply);			/* Let user enter a file name		*/

	if (macSFReply.good) {
		SetCursor(*gWatchCursor);
		return( DoSaveAs(&macSFReply) );

	} else {							/* User cancelled SaveAs dialog box	*/
		return(FALSE);
	}
}
	

/******************************************************************************
 PickFileName

		Put up a standard put file dialog box to allow the user to
		specify a new name under which to save a copy of the document.
 ******************************************************************************/

void	CDocument::PickFileName(
	SFReply		*macSFReply)
{
	Point			corner;				/* Top left corner of dialog box	*/
	Str255			origName;			/* Default name for file			*/
	StringHandle	prompt;				/* Prompt string					*/
		
										/* Center dialog box on the screen	*/
	FindDlogPosition('DLOG', putDlgID, &corner);
	
	GetName(origName);
	
	prompt = GetString(STRprompt);
	FailNILRes( prompt);
	
	MoveHHi( (Handle) prompt);
	HLock((Handle) prompt);
	SFPPutFile(corner, *prompt, origName, NULL, macSFReply, putDlgID, NULL);
	HUnlock((Handle) prompt);
}


/******************************************************************************
 ConfirmClose

		Don't close a dirty document without giving user a chance to save
		the changes
 ******************************************************************************/

Boolean	CDocument::ConfirmClose(
	Boolean		quitting)				/* Is this part of a quit sequence?	*/
{
	Boolean		okToClose = TRUE;
	Str255		docName;
	Str255		status;
	short		response;

	if (dirty) {						/* If Document has changed ...		*/
	
        if (itsWindow != NULL)      
            itsWindow->Update();
            
		GetName(docName);
		GetIndString(status, STRcommon,
							(quitting) ? strQUITTING : strCLOSING);
		ParamText(docName, status, NULL, NULL);
										/* Ask user if s/he want to save	*/
										/*   the changes					*/
		PositionDialog('ALRT', ALRTsaveChanges);
		InitCursor();
		response = Alert(ALRTsaveChanges, NULL);
		
		if (response == answerYES) {
			SetCursor(*gWatchCursor);
			okToClose = DoSave();		/* Save Document before closing it	*/
			
		} else if (response == answerCANCEL) {
			okToClose = FALSE;			/* Abort the close					*/
			
		}								/* If answer is NO, we don't need	*/
										/*   to do anything here			*/
	}
	
	return(okToClose);
}
	

/******************************************************************************
 GetName

		Return the name of a Document.
 ******************************************************************************/

void	CDocument::GetName(
	Str255		theName)
{
	if (itsFile != NULL) {				/* 1st choice: name of its file		*/
		itsFile->GetName(theName);
	} else if (itsWindow != NULL) {		/* 2nd choice: name of its window	*/
		itsWindow->GetTitle(theName);
	} else {							/* 3rd choice: NULL string			*/
		theName[0] = 0;
	}
}
	

/******************************************************************************
 UpdateUndo

		Update the text of the "Undo" command to reflect the current
		status of undo/redo and the task name.
 ******************************************************************************/

void	CDocument::UpdateUndo()
{
	Str255		status;
	Str255		task;
	
	if (active) {
		GetIndString(status, STRcommon, (undone) ? strREDO : strUNDO);
		GetIndString(task, STRtaskNames, lastTask->GetNameIndex());
		ConcatPStrings(status, task);
		
	} else {
		GetIndString(status, STRcommon, strUNDO);
	}
	
	gBartender->SetCmdText(cmdUndo, status);
}

/******************************************************************************
 MakePrinter

	Make the printer object for a printable document. Subclasses can override
	to use a subclass of CPrinter
 ******************************************************************************/

void CDocument::MakePrinter( void)
{
	itsPrinter = new(CPrinter);
	itsPrinter->IPrinter(this, NULL);
}

