/******************************************************************************
 CPane.c

							The Pane Class
	
	Abstract class for objects which may be displayed within windows. Panes
	can also respond to the mouse, keyboard, and commands. Subclasses must
	be created to display different kinds of informations and implement
	different ways of responding to user actions.
		
	SUPERCLASS = CView
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	TCL 1.1 CHANGES
	[
		- implemented GetWindow, GetBorder, SetResBorder, SetBorder methods
		- modified DrawAll to include borders
		- Moved TrackMouse from CView to CPane.
		- for 32-bit support, changed interface for SetFrameOrigin, 
		  GetFrame, GetAperture, Place, Offset, EnclosureScrolled, 
		  WindToFrame, WindToFrameR, FrameToWind, FrameToWindR, EnclToFrame, 
		  EnclToFrameR, FrameToEncl, FrameToEnclR, and FrameToGlobalR methods 
		  to use long coordinates.
		- implemented new methods RefreshLongRect, QDToFrame, QDToFrameR, 
		  FrameToQD, FrameToQDR, and SectAperture.
		  
		- see comments for Prepare, Draw, and DrawAll for some important
		  changes.
	]
	
 ******************************************************************************/

#include "Global.h"
#include "CPane.h"
#include "CWindow.h"
#include "CPaneBorder.h"
#include "CPrinter.h"
#include "TBUtilities.h"
#include "CList.h"
#include "CEnvironment.h"
#include "CMouseTask.h"

#define		kOriginConstraint		8192

/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


Rect	CPane::cPageArea;

/******************************************************************************
 IPane

		Initialize a Pane object
 ******************************************************************************/

void	CPane::IPane(
	CView			*anEnclosure,
	CBureaucrat		*aSupervisor,
	short			aWidth,
	short			aHeight,
	short			aHEncl,
	short			aVEncl,
	SizingOption	aHSizing,
	SizingOption	aVSizing)
{
										/* Initialize superclass			*/
	CView::IView(anEnclosure, aSupervisor);
	
	visible = TRUE;
	active = TRUE;

	width = aWidth;						/* Set instance variables			*/
	height = aHeight;
	hEncl = aHEncl;
	vEncl = aVEncl;
	hSizing = aHSizing;
	vSizing = aVSizing;
	
	autoRefresh = TRUE;
	printClip = clipFRAME;
	
	IPaneX();							/* Extra initialization				*/
}
	

/******************************************************************************
 IViewTemp {OVERRIDE}

		Initialize a Pane using a template
 ******************************************************************************/

void	CPane::IViewTemp(
	CView			*anEnclosure,		/* Enclosing view					*/
	CBureaucrat		*aSupervisor,		/* Boss in chain of command			*/
	Ptr				viewData)			/* Template with View data			*/
{
	register PaneTempP	p;
	
	
	p = (PaneTempP) viewData;			/* Initialize superclass			*/
	inherited::IViewTemp(anEnclosure, aSupervisor, (Ptr) &p->sViewTemp);
	
	width = p->width;					/* Set instance vars from template	*/
	height = p->height;
	hEncl = p->hEncl;
	vEncl = p->vEncl;
	hSizing = p->hSizing;
	vSizing = p->vSizing;
    autoRefresh = p->autoRefresh != 0;
	printClip = p->printClip;
	
	IPaneX();							/* Extra initialization				*/
}
	

/******************************************************************************
 IPaneX {OVERRIDE}

		Extra initialization for a Pane
 ******************************************************************************/

void	CPane::IPaneX()
{
	itsBorder = NULL;
	itsLastTask = NULL;

	itsEnclosure->AddSubview(this);		/* This is a subview of enclosure	*/

	printing = FALSE;
	macPort = itsEnclosure->GetMacPort();
	CalcFrame();						/* Set up Frame coordinates			*/
	CalcAperture();						/* Determine visible portion		*/	
	
	itsEnvironment = NULL;
}
	

/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a Pane
 ******************************************************************************/

void	CPane::Dispose()
{
	ForgetObject( itsEnvironment);	/* Throw out drawing environment	*/
	ForgetObject( itsBorder);		/* Throw out pane border			*/
	
		/* makes sure that the last undo task is disposed. Since the 	*/
		/* undo tasks usually have references back to this pane, 		*/
		/* it would be bad if this pane was disposed and the task was 	*/
		/* not informed.												*/

	if (itsLastTask)
	{
		Notify( NULL);			/* the document disposes of the task 	*/
								/* removed ASSERT TCL 1.1.2 DLP 10/28/91 */
	}
		
	inherited::Dispose();			/* Pass message on to superclass	*/
}


/**** A C C E S S I N G   M E T H O D S ****/
	

/******************************************************************************
 SetFrameOrigin

		Set the coordinates of the top left corner of a Pane's frame. The
		bottom right corner is adjusted to according to the width and
		height. The frame determines the local QuickDraw coordinates system
		(called Frame coordinates) used for drawing the pane.
 ******************************************************************************/

void	CPane::SetFrameOrigin(
	long		fLeft,
	long		fTop)
{
	frame.left = fLeft;					/* Set top left coordinates			*/
	frame.top = fTop;
	frame.right = frame.left + width;	/* Set right and bottom using the	*/
	frame.bottom = frame.top + height;	/*   current width and height		*/
	
	ForceNextPrepare();
}

	
/******************************************************************************
 GetFrame

		Return the frame of the pane
 ******************************************************************************/

void	CPane::GetFrame(
	LongRect		*theFrame)
{
	*theFrame = frame;					/* Return copy of instance variable	*/
}

	
/******************************************************************************
 GetLengths

		Return the width and height of a Pane
 ******************************************************************************/

void	CPane::GetLengths(
	short		*theWidth,
	short		*theHeight)
{
	*theWidth = width;					/* Return instance variables		*/
	*theHeight = height;
}


/******************************************************************************
 GetOrigin

		Return the origin of a Pane. Origin is the location of the
		top left corner of the Pane's window, expressed in the Pane's
		Frame coordinates.
 ******************************************************************************/

void	CPane::GetOrigin(
	long		*theHOrigin,
	long		*theVOrigin)
{
	*theHOrigin = hOrigin;				/* Return instance variables		*/
	*theVOrigin = vOrigin;
}

	
/******************************************************************************
 GetAperture {OVERRIDE}

		Return the portion of the pane where drawing may occur. Aperture
		is in the pane's Frame coordinates.
 ******************************************************************************/

void	CPane::GetAperture(
	LongRect		*theAperture)
{
	*theAperture = aperture;			/* Return copy of instance variable	*/
}

/******************************************************************************
 GetWindow
 
 	Return the CWindow that the pane is enclosed within.
******************************************************************************/
CWindow *CPane::GetWindow( void)
{
	return (CWindow*) GetWRefCon( macPort);
}

/******************************************************************************
 Contains {OVERRIDE}

		Check whether a point, specified in Window coordinates, is contained
		within the aperture of a Pane.
 ******************************************************************************/

Boolean		CPane::Contains(
	Point			windPt)
{
	LongPt	framePt;
	
	WindToFrame( windPt, &framePt);		/* Convert to Frame coordinates		*/
										/* Is point inside aperture?		*/
	return(PtInLongRect(&framePt, &aperture));
}
	

/******************************************************************************
 ReallyVisible {OVERRIDE}

		Check whether a Pane and all its enclosures are visible. Also
		make sure that the origin does not place the pane outside
		valid QuickDraw coordinates.
 ******************************************************************************/

Boolean	CPane::ReallyVisible()
{
	if ( visible
			&&  ((frame.left - hOrigin) >= MININT)
			&&  ((frame.right - hOrigin) <= MAXINT)
			&&  ((frame.top - vOrigin) >= MININT)
			&&  ((frame.bottom - vOrigin) <= MAXINT) ) {
			
		return(itsEnclosure->ReallyVisible());
	} else {
		return(FALSE);
	}
}


/******************************************************************************
 GetPixelExtent

		Retrieve the dimensions of a Pane in pixels, which is the same as
		the width and height of its frame
 ******************************************************************************/

void	CPane::GetPixelExtent(
	long		*hExtent,
	long		*vExtent)
{
	*hExtent = width;
	*vExtent = height;
}


/******************************************************************************
 SetPrintClip

		Specify the clipping option used when printing
 ******************************************************************************/

void	CPane::SetPrintClip(
	ClipOption		aPrintClip)
{
	printClip = aPrintClip;
}

/******************************************************************************
 SetBorder
 
  Set the border for this pane.
******************************************************************************/

void CPane::SetBorder( CPaneBorder *aBorder)
{
	Rect		paneFrame, borderRect, paneAperture;
	Boolean hadPendingUpdate;
	
	FrameToGlobalR( &aperture, &paneAperture);
	hadPendingUpdate = RectInRgn( &paneAperture, ((WindowPeek)macPort)->updateRgn);
		
	/* Can't assume that the pane's enclosure is a CPane	*/
	/* so RefreshRect() is not safe	here.					*/
		
	FrameToWindR( &frame, &paneFrame);
	GetWindow()->Prepare();
	
	if (itsBorder)
	{
		borderRect = paneFrame;
		itsBorder->CalcBorderRect( &borderRect);
		
		InvalRect( &borderRect);	/* clear old border			*/
		itsBorder->Dispose();
	}
		
	itsBorder = aBorder;
	
	if (itsBorder)
	{
		borderRect = paneFrame;
		itsBorder->CalcBorderRect( &borderRect);
		
		InvalRect( &borderRect); /* draw new border			*/
	}
	if (!hadPendingUpdate)
	{
		FrameToWindR( &aperture, &paneAperture);
		ValidRect( &paneAperture);		/* the pane itself is not invalid */
	}
}

/******************************************************************************
 SetResBorder
 
 	Create a border for this pane from a 'PBrd' resource
******************************************************************************/

void CPane::SetResBorder( short resID)
{
	ForgetObject( itsBorder);
	itsBorder = new CPaneBorder;
	itsBorder->IResPaneBorder( resID);
}

/******************************************************************************
 GetBorder
******************************************************************************/

CPaneBorder *CPane::GetBorder( void)
{
	return itsBorder;
}

/**** D I S P L A Y   M E T H O D S ****/
	

/******************************************************************************
 Show {OVERRIDE}

		Make a Pane visible
 ******************************************************************************/

void	CPane::Show()
{
	if (!visible) {
		inherited::Show();
		Refresh();
		RefreshBorder();
	}
}
	

/******************************************************************************
 Hide {OVERRIDE}

		Make a Pane invisible
 ******************************************************************************/

void	CPane::Hide()
{
	if (visible) {
		Refresh();
		RefreshBorder();
		inherited::Hide();
	}
}


/**** S I Z E   A N D   L O C A T I O N   M E T H O D S ****/


/******************************************************************************
 Place

		Place the top left corner of a Pane at a location specified in
		the coordinates of its enclosure
 ******************************************************************************/

void	CPane::Place(
	long		aHEncl,
	long		aVEncl,
	Boolean		redraw)
{
	long		hOffset;
	long		vOffset;

	hEncl = aHEncl;
	vEncl = aVEncl;
										/* Determine location of pane in	*/
										/*   Window coords					*/
	itsEnclosure->SubpaneLocation(hEncl, vEncl, &aHEncl, &aVEncl);

	hOffset = hOrigin - frame.left + aHEncl;
	vOffset = vOrigin - frame.top + aVEncl;
	
	Offset(hOffset, vOffset, redraw);
}


/******************************************************************************
 Offset

		Translate the location of the Pane within its enclosure
 ******************************************************************************/

		/*
		 * Tell a Pane that its enclosure has moved. Send Offset message
		 * with redraw set to FALSE, since enclosure will have already
		 * taken care of adding the appropriate areas to the update region.
		 */
		 static void	Pane_EnclosureMoved(
		 	CPane		*thePane,
		 	LongPt		*offset)
		 {
		 	thePane->Offset(offset->h, offset->v, FALSE);
		 }
		 
void	CPane::Offset(
	long		hOffset,				/* Pixels to offset horizontally	*/
	long		vOffset,				/* Pixels to offset vertically		*/
	Boolean		redraw)					/* Redraw Pane or not?				*/
{
	LongPt		offset;					/* Offsets combined into a point	*/
	long		enclhOrig;
	long		enclvOrig;
					
	if (redraw) {
		Refresh();						/* Force redraw of area covered by	*/
										/*   current location of pane		*/
		RefreshBorder();				/* and border, if any				*/
	}

	hOrigin -= hOffset;					/* Change location of the pane		*/
	vOrigin -= vOffset;
	
		/* Keep hEncl, and vEncl up to date when pane is moved.		*/
		/* Thanks to Eric Olson for the fix.						*/
		
	itsEnclosure->GetOrigin(&enclhOrig, &enclvOrig);
	hEncl = enclhOrig - hOrigin;
	vEncl = enclvOrig - vOrigin;
	
	CalcAperture();						/* Visible portion of pane could	*/
										/*   change as a result of a shift	*/
										/*   in its location				*/
	
	if (redraw) {									
		Refresh();						/* Redraw pane at its new location	*/
		RefreshBorder();				/* and border, if any				*/
	}
	
	if (itsSubviews != NULL) {			/* Inform subviews of the move		*/
		offset.h = hOffset;				/*   so they can adjust their		*/
		offset.v = vOffset;				/*   locations						*/
		itsSubviews->DoForEach1((EachFunc1) Pane_EnclosureMoved, (long) &offset);
	}
}
	

/******************************************************************************
 ChangeSize

		Change the size of a Pane. The delta rectangle parameter specifies
		how each side of the pane should change. Positive values mean a
		move to the right or down, negative values to the left or up.
 ******************************************************************************/

		/*
		 *	Adjust subviews when the size of enclosure changes
		 */
		void	Pane_AdjustToEnclosure(
			CPane		*thePane,
			Rect		*delta)
		{
			thePane->AdjustToEnclosure(delta);
		}

void	CPane::ChangeSize(
	register Rect		*delta,			/* Movement for each side			*/
	Boolean				redraw)			/* Redraw Pane or not?				*/
{
    
    LongRect    oldAperture = aperture;     /* Save old aperture                */
    LongRect	oldFrame = frame;
	Rect 		tmp;

	if (redraw) {
		Refresh();						/* Force redraw of area currently	*/
										/*   occupied by the pane			*/
		RefreshBorder();				/*  and redraw border if there is one */
	}
	
	ResizeFrame(delta);					/* Change dimensions of pane		*/
						
    CalcAperture();                     /* Determine new aperture           */
	
	if (redraw) {
		Refresh();						/* Force draw of pane at new size	*/
		RefreshBorder();	/* TCL 1.1.1 DLP 10/1/91 */

										/* If autoRefresh is FALSE (and		*/
										/*   other safety checks), we do	*/
										/*   not want to redraw the area	*/
										/*   which is common to the old and	*/
										/*   new pane size. Note that the	*/
										/*   Refresh method will have set	*/
										/*   the port for us already.		*/
		if (!autoRefresh && ReallyVisible()) {
			if (SectLongRect(&aperture, &oldAperture, &oldAperture))
			{
				
				FrameToQDR( &oldAperture, &tmp);
				ValidRect( &tmp);
			}
			if (itsBorder)
			{
				Rect oldBorder, newBorder;
				
				FrameToWindR( &oldFrame, &oldBorder);
				itsBorder->CalcBorderRect( &oldBorder);
				FrameToWindR( &frame, &newBorder);
				itsBorder->CalcBorderRect( &newBorder);
				
				if (SectRect( &oldBorder, &newBorder, &tmp))
				{
					GetWindow()->Prepare();
					ValidRect( &tmp);
					Prepare();
				}
				
			}
		}
	}
	
	if (itsSubviews != NULL)
		itsSubviews->DoForEach1((EachFunc1) Pane_AdjustToEnclosure, (long)delta);
}
	

/******************************************************************************
 AdjustToEnclosure

		Adjust size and/or location of a Pane after its enclosure's size
		has changed. The deltaEncl rectangle parameter specifies how each
		side of the enclosure has changed. Positive values mean a move to
		the right or down, negative values to the left or up.
 ******************************************************************************/
  
void	CPane::AdjustToEnclosure(
	Rect		*deltaEncl)				/* Change in enclosure's frame		*/
{
	Rect		delta;					/* Change in this pane's frame		*/
	short		hOffset = 0;			/* Horizontal offset in location	*/
	short		vOffset = 0;			/* Vertical offset in location		*/
	Boolean		moved = FALSE;			/* Do we need to move the pane?		*/
	Boolean		sized = FALSE;			/* Do we need to size the pane?		*/
	
	SetRect(&delta, 0, 0, 0, 0);
	AdjustHoriz(deltaEncl, &delta, &hOffset, &moved, &sized);
	AdjustVert(deltaEncl, &delta, &vOffset, &moved, &sized);
	
	if (moved) {
		if (sized) {					/* Pane must be moved and sized		*/
            Offset(hOffset, vOffset, TRUE);
			ChangeSize(&delta, TRUE);
		} else {
										/* Move pane by offsetting location	*/
			Offset(hOffset, vOffset, TRUE);
		}
	} else if (sized) {					/* Change size of Pane				*/
		ChangeSize(&delta, TRUE);
	} else {							/* Pane stays the same, but change	*/
		CalcAperture();					/*   in enclosure may change its	*/
										/*   aperture						*/
	}
}
		

/******************************************************************************
 AdjustHoriz

		Adjust the horizontal size and/or location of a Pane
 ******************************************************************************/

void	CPane::AdjustHoriz(
	register Rect	*deltaEncl,
	register Rect	*delta,
	short			*offset,
	Boolean			*moved,
	Boolean			*sized)
{
	switch (hSizing) {
	
		case sizFIXEDLEFT:
			if (*offset = deltaEncl->left) {
				*moved = TRUE;
			}
			break;
			
		case sizFIXEDRIGHT:
			if (*offset = deltaEncl->right) {
				*moved = TRUE;
			}
			break;
			
		case sizELASTIC:
			delta->left = deltaEncl->left;
			delta->right = deltaEncl->right;
			if ( (delta->left != 0)  ||  (delta->right != 0) )
				*sized = TRUE;
			break;
	}
}
		

/******************************************************************************
 AdjustVert

		Adjust the vertical size and/or location of a Pane
 ******************************************************************************/

void	CPane::AdjustVert(
	register Rect	*deltaEncl,
	register Rect	*delta,
	short			*offset,
	Boolean			*moved,
	Boolean			*sized)
{
	switch (vSizing) {
	
		case sizFIXEDTOP:
			if (*offset = deltaEncl->top) {
				*moved = TRUE;
			}
			break;
			
		case sizFIXEDBOTTOM:
			if (*offset = deltaEncl->bottom) {
				*moved = TRUE;
			}
			break;
			
		case sizELASTIC:
			delta->top = deltaEncl->top;
			delta->bottom = deltaEncl->bottom;
			if ( (delta->top != 0)  ||  (delta->bottom != 0) )
				*sized = TRUE;
			break;
	}
}
	

/******************************************************************************
 EnclosureScrolled

		Adjust location of a Pane after its enclosure has scrolled. This
		only matters if the Pane is sticky in the direction of movement.
 ******************************************************************************/

void	CPane::EnclosureScrolled(
	long		hOffset,
	long		vOffset)
{
	if (hSizing != sizFIXEDSTICKY) {
		hOffset = 0;
	}
		
	if (vSizing != sizFIXEDSTICKY) {
		vOffset = 0;
	}
	
	if ( (hOffset != 0)  ||  (vOffset != 0) ) {
		Offset(-hOffset, -vOffset, FALSE);
	}
}

/**** A D A P T I N G   M E T H O D S ****/
	

/******************************************************************************
 FitToEnclosure

		Make the frame of a Pane "fit" to the interior of its enclosure in
		the horizontal and/or vertical directions. For a horizontal fit,
		the Pane's width and left edge are changed to coincide with
		those of its enclosure's interior. Similarly for a vertical fit.
		Pane is NOT automatically redrawn.
 ******************************************************************************/

void	CPane::FitToEnclosure(
	Boolean		horizFit,
	Boolean		vertFit)
{
	LongRect	interior;
	short	hOffset = 0;
	short	vOffset = 0;
	Rect	delta;
	
	itsEnclosure->GetInterior(&interior);
	EnclToFrameR(&interior);
	
	delta.left = delta.top = delta.right = delta.bottom = 0;

	if (horizFit) {
		hOffset = interior.left - frame.left;
		delta.right = (interior.right - interior.left) - width;
	}
	
	if (vertFit) {
		vOffset = interior.top - frame.top;
		delta.bottom = (interior.bottom - interior.top) - height;
	}
	
	Offset(hOffset, vOffset, FALSE);
	ChangeSize(&delta, FALSE);
}
	

/******************************************************************************
 FitToEnclFrame

		Make the frame of a Pane "fit" to the frame of its enclosure in
		the horizontal and/or vertical directions. For a horizontal fit,
		the Pane's width and left edge are changed to coincide with
		those of its enclosure's frame. Similarly for a vertical fit.
		Pane is NOT automatically redrawn.
 ******************************************************************************/

void	CPane::FitToEnclFrame(
	Boolean		horizFit,
	Boolean		vertFit)
{
	LongRect	enclFrame;
	short	hOffset = 0;
	short	vOffset = 0;
	Rect	delta;
	
	itsEnclosure->GetFrame(&enclFrame);
	EnclToFrameR(&enclFrame);
	
	delta.left = delta.top = delta.right = delta.bottom = 0;

	if (horizFit) {
		hOffset = enclFrame.left - frame.left;
		delta.right = (enclFrame.right - enclFrame.left) - width;
	}
	
	if (vertFit) {
		vOffset = enclFrame.top - frame.top;
		delta.bottom = (enclFrame.bottom - enclFrame.top) - height;
	}
	
	Offset(hOffset, vOffset, FALSE);
	ChangeSize(&delta, FALSE);
}
	

/******************************************************************************
 CenterWithinEnclosure

		Center a Pane's frame within the interior of its enclosure's frame
		either horizontally and/or vertically. Only the location of the
		Pane is changed; its size is not altered. Pane is NOT automatically
		redrawn.
 ******************************************************************************/

void	CPane::CenterWithinEnclosure(
	Boolean		horizCenter,
	Boolean		vertCenter)
{
	LongRect	interior;
	short	hOffset = 0;
	short	vOffset = 0;
	
	itsEnclosure->GetInterior(&interior);
	EnclToFrameR(&interior);
	
	if (horizCenter) {
		hOffset = interior.left - frame.left +
					(interior.right - interior.left - width) / 2;
	}
	
	if (vertCenter) {
		vOffset = interior.top - frame.top +
					(interior.bottom - interior.top - height) / 2;
	}
	
	Offset(hOffset, vOffset, FALSE);
}


/**** D R A W I N G   M E T H O D S ****/
	

/******************************************************************************
 Draw

		Draw the contents of the Pane. The area parameter, specified in
		the pane's Frame coordinates, indicates the portion of the Pane
		which needs to be drawn. Subclasses of CPane which display
		data (rather than just manage subviews) must override this method.
		
		TCL 1.1
			- When usingLongCoord is FALSE, area is in Frame coordinates,
			  and panes may draw directly in Frame coordinates.
			  
			- When usingLongCoord is TRUE, area is in QuickDraw coordinates,
			  and drawing must be done in QuickDraw coordinates.
			  Use QDToFrame/QDToFrameR to convert from QuickDraw to
			  Frame coordinates, and use FrameToQD/FrameToQDR to convert
			  from Frame to QuickDraw coordinates. These utilities may
			  also be used when usingLongCoord is FALSE, and they still
			  do the right thing.
		
 ******************************************************************************/

void	CPane::Draw(
	Rect		*area)
{
}										/* Null Method						*/
	

/******************************************************************************
 DrawAll

		Draw a Pane and all of its subviews. The port must be prepared
		ahead of time.
		
		TCL 1.1
			- DrawAll now draws pane borders (within Pane_Draw)
			- DrawAll always expects area to be in Window coordinates.
			- When calling Draw, DrawAll passes a rectangle in Frame
			  coordinates when usingLongCoord is FALSE, and QD coordinates
			  when it is TRUE.
 ******************************************************************************/

		/*
		 * Prepare and Draw a Pane
		 */
			void	Pane_Draw(
				register CPane	*thePane,
				Rect			*area)	/* Area in Window coords	*/
			{
				Rect		paneArea;	/* Area in Pane's Frame coords		*/
				Rect		paneAperture;
				CPaneBorder *border;
				
				if (thePane->ReallyVisible()) {
										/* Only draw if pane is visible		*/
	

					border = thePane->GetBorder();
					if (border)
					{
						Rect paneFrame, borderRect;
						
						/* Get the pane's frame in window coordinates		*/
						/* then ask the border for the border rect.			*/
						/* If that rect intersects the area to draw, then	*/
						/* go ahead and draw the border						*/
						
						thePane->FrameToWindR( &thePane->frame, &paneFrame);
						borderRect = paneFrame;
						border->CalcBorderRect( &borderRect);
						if (SectRect( &borderRect, area, &borderRect))
						{
							if (thePane->printing)
								SetOrigin( 0,0);
							else
								thePane->GetWindow()->Prepare();
							ClipRect( &borderRect);
							CView::ForceNextPrepare();
							border->DrawBorder( &paneFrame);
						}
					}
					/* DrawAll always receives the area in Window 			*/
					/* coordinates. If we're not printing, clip the			*/
					/* area to the aperture.								*/
					
					if (!thePane->printing)
					{					
							/* get its aperture in Window coordinates			*/
									
						thePane->FrameToWindR( &thePane->aperture, &paneAperture);
						
											/* See if area intersects the		*/
											/*   aperture of the Pane			*/
						if (SectRect( &paneAperture, area, &paneArea)) {
											/* Some portion of this Pane must	*/
											/*   be drawn.						*/
							thePane->DrawAll(&paneArea);
						}
					}
					else
					{	/* not printing, don't clip the area		*/
						paneArea = *area;
						thePane->DrawAll( &paneArea);
					}
				}
			}
			
void	CPane::DrawAll(
	Rect		*area)					/* Portion of Pane to draw			*/
{
	Rect	tmpArea = *area;
	
	Prepare();
	
	/* area is in Window coordinates, put it into Frame or QD coordinates	*/

	if (usingLongCoord)
		OffsetRect( &tmpArea, thePort->portRect.left, thePort->portRect.top);
	else
		OffsetRect( &tmpArea, hOrigin, vOrigin);

	if (!printing) {	/* no need to SectRect here, already done in Pane_Draw */

		ClipRect( &tmpArea);
	}

	Draw( &tmpArea);
				
	if (itsSubviews != NULL)
	{
		// clip area to aperture of this view before drawing subviews
		
		if (!printing)
		{
			FrameToWindR( &aperture, &tmpArea);
			SectRect( &tmpArea, area, &tmpArea);
		}
		else			/* TCL 1.1.2 DLP 10/28/91 */
			tmpArea = *area;
			
		itsSubviews->DoForEach1((EachFunc1) Pane_Draw, (long) &tmpArea);
	}

}

/******************************************************************************
 Refresh

		Force a redraw of an entire Pane
 ******************************************************************************/

void	CPane::Refresh()
{
	LongRect	pFrame;
	
	pFrame = frame;						/* Refresh frame of Pane			*/
	RefreshLongRect( &pFrame);
}


/******************************************************************************
 RefreshLongRect

		Force a redraw of an area by adding it to the update region
		of its window. Drawing will not actually take place until the next
		update event is processed. Only the portion of the area that
		intersects the pane's aperture is invalidated.
 ******************************************************************************/

void CPane::RefreshLongRect( LongRect *area)
{
	Rect	visArea;
	
	if (ReallyVisible() && SectAperture( area, &visArea))
		InvalRect( &visArea);

}	
/******************************************************************************
 RefreshRect

		Force a redraw of an area by adding it to the update region
		of its window. Drawing will not actually take place until the next
		update event is processed.
 ******************************************************************************/

void	CPane::RefreshRect(
	Rect		*area)					/* Area in Frame coordinates		*/
{
	LongRect	tmp;
	
	QDToLongRect( area, &tmp);
	RefreshLongRect( &tmp);
	
}

/******************************************************************************
 RefreshBorder

	Force a redraw of a Pane's border
 ******************************************************************************/

void	CPane::RefreshBorder( void)
{
	Rect	borderRect;
	
	if (itsBorder)
	{
		FrameToWindR( &frame, &borderRect);
		itsBorder->CalcBorderRect( &borderRect);
		GetWindow()->Prepare();
		
		InvalRect( &borderRect);
	}
}


/**** P R I N T I N G   M E T H O D S ****/
	

/******************************************************************************
 Paginate

		Specify the pagination of a Pane. By default, Panes only print one
		page.
 ******************************************************************************/

void CPane::Paginate( CPrinter *aPrinter, short pageWidth, short pageHeight)			
{
		/* Tell the printer we have only one page, and let it use the	*/
		/* printable page area for the strip sizes						*/
		
	aPrinter->SetStrips( 1, 1);

}	/* CPane::Paginate */

/******************************************************************************
 AboutToPrint

		The specified range of pages is about to be printed.
 ******************************************************************************/
 	typedef struct
 	{
 		short firstPage;
 		short lastPage;
 	
 	} tPageRange;
 	
 	static void Pane_AboutToPrint( CObject *aPane, long aPageRange)
 	{
 		tPageRange *pageRange = (tPageRange*) aPageRange;
 		
 		((CPane*) aPane)->AboutToPrint( &pageRange->firstPage,
 							&pageRange->lastPage);
 	}

void	CPane::AboutToPrint(
	short		*firstPage,
	short		*lastPage)
{

	tPageRange	pageRange;
	
	printing = TRUE;
	ForceNextPrepare();
	
		/* subviews should not be able to modify the page range	*/
		/* so they will be passed a local copy of the range		*/
		
	if (itsSubviews)
	{
		pageRange.firstPage = *firstPage;
		pageRange.lastPage = *lastPage;
		itsSubviews->DoForEach1( (EachFunc1) Pane_AboutToPrint, (long) &pageRange);
	}
	
}


/******************************************************************************
 PrintPage

		Print the specified page. This method ignores the page number and
		simply draws the entire pane. Subclasses which support multi-page
		panes must override this method.
 ******************************************************************************/

void	CPane::PrintPage(
	short		pageNum,				/* Number of the page to print		*/
	short		pageWidth,				/* Width of printed page in pixels	*/
	short		pageHeight,				/* Height of page in pixels			*/
	CPrinter	*aPrinter)
{
	LongRect	area;
	Rect		qdArea;
	
	/* setup class variable cPageArea with portion of page to be printed */
	aPrinter->GetPageArea( pageNum, &area);
	FrameToWindR( &area, &cPageArea);
	ForceNextPrepare();
	
	/* pass a rectangle to DrawAll that is based on the frame top left	*/
	/* but is the same size as the page.								*/
	
	area.left = frame.left;
	area.top = frame.top;
	area.right = area.left + pageWidth;
	area.bottom = area.top + pageHeight;
	FrameToWindR( &area, &qdArea);
	DrawAll( &qdArea);
	
	SetRect( &cPageArea, 0, 0, 0, 0);
}
	

/******************************************************************************
 DonePrinting

		Print loop has been completed. Perform any necessary clean-up.
 ******************************************************************************/

	static void Pane_DonePrinting( CObject *aPane)
	{
		((CPane*)aPane)->DonePrinting();
	}
	
void	CPane::DonePrinting()
{
	printing = FALSE;
	ForceNextPrepare();
	
	if (itsSubviews)
		itsSubviews->DoForEach( (EachFunc) Pane_DonePrinting);
}


/******************************************************************************
 PrepareToPrint

		Set up coordinate system and clipping region for printing. This
		method prepares the port so that only the portion of the pane
		which is visible on the screen is printed.
 ******************************************************************************/

void	CPane::PrepareToPrint()
{
	Rect		clipR;
	LongRect 	lr;

	if (usingLongCoord)
		SetOrigin( hOrigin % kOriginConstraint, vOrigin % kOriginConstraint);
	else
		SetOrigin((short) hOrigin, (short) vOrigin);
		
	switch (printClip) {
	
		case clipAPERTURE:
			FrameToQDR( &aperture, &clipR);
			break;
			
		case clipFRAME:
			FrameToQDR( &frame, &clipR);
			break;
			
		case clipPAGE:
			clipR = cPageArea;
			if (usingLongCoord)
				OffsetRect( &clipR, thePort->portRect.left, thePort->portRect.top);
			else
			{
				WindToFrameR( &clipR, &lr);
				LongToQDRect( &lr, &clipR);
			}		 
			break;
	}
	
	ClipRect(&clipR);
}


/**** C A L I B R A T I N G   M E T H O D S ****/


/******************************************************************************
 Prepare

		Prepare for drawing the Pane by setting the port, establishing
		the correct coordinate system, clipping the the proper area, and
		restoring its drawing environment.
		
		TCL 1.1
			- Now has an optimization to reduce the cost of making 
			  redundant calls to Prepare. The class variable CView::cPreparedView
			  is kept in synch with the currently prepared view. Prepare
			  can now check if the pane has already been prepared. For
			  this to work correctly, it is important that CView::ForceNextPrepare
			  is called whenever the port, origin, or clipping region are
			  altered outside the Prepare mechanism. ForceNextPrepare
			  is called automatically after each event has been handled.
			  
			- Prepare sets up the port differently depending on the state
			  of the usingLongCoord flag. When it is FALSE, the origin is
			  set up as before, taking it directly from hOrigin and vOrigin.
			  This sets up the port so the pane can draw directly in
			  Frame coordinates.
			  
			  When usingLongCoord is TRUE, Frame coordinates, hOrigin, and vOrigin
			  are potentially beyond the 16 bit QuickDraw limits. Panes
			  must call FrameToQD and FrameToQDR to convert Frame coordinates
			  to QuickDraw coordinates before drawing. It would be possible
			  to simply set the origin to 0,0 in this case, except that
			  patterns and pixpats would not draw correctly. QuickDraw uses
			  the origin to align patterns. To accomodate this, we move the
			  origin within an area constrained by kOriginConstraint. 
			  
 ******************************************************************************/

void	CPane::Prepare()
{
	Rect		tempRect;		

	if (CView::cPreparedView == this)
		return;

	inherited::Prepare();
	if (printing)
	{
		PrepareToPrint();
	} 
	else
	{

		SetPort(macPort);				/* Use the Port of this Pane		*/
										/* Set to Frame coordinates			*/
										
				/* Setup the origin for the Pane and clip drawing			*/
				/* to the visible portion.									*/
										
		if (usingLongCoord)
		{
			SetOrigin( hOrigin % kOriginConstraint, vOrigin % kOriginConstraint);
			FrameToQDR( &aperture, &tempRect);
		}
		else
		{
			SetOrigin((short) hOrigin, (short) vOrigin);
			LongToQDRect( &aperture, &tempRect);
		}
		ClipRect(&tempRect);
		
	}
	RestoreEnvironment();

}

	
/******************************************************************************
 RestoreEnvironment

		Restore the drawing environment of a Pane
 ******************************************************************************/

void	CPane::RestoreEnvironment()
{
	if (itsEnvironment != NULL)
		itsEnvironment->Restore();
}

	
/******************************************************************************
 CalcFrame

		Determine the coordinates of the Pane's frame and origin based on
		the width, height, hEncl, vEncl. These instance variables must be
		set before invoking this method.
		
		The frame is set to have its top left corner at (0, 0). The defines
		the Frame coordinates of the Pane. Subclasses must override this
		method if they use a different Frame coordinate system.
 ******************************************************************************/

void	CPane::CalcFrame()
{
	long		hLocation;
	long		vLocation;

	frame.left = frame.top = 0;			/* Frame defines pane coords		*/
	frame.right = width;				/* Top left of frame is at (0, 0)	*/
	frame.bottom = height;
	
										/* Determine location of pane in	*/
										/*   Window coords					*/
	itsEnclosure->SubpaneLocation(hEncl, vEncl, &hLocation, &vLocation);

	hOrigin = frame.left - hLocation;	/* Origin maps between Frame and	*/
	vOrigin = frame.top - vLocation;	/*   Window coords. Origin is the	*/
										/*   top left corner of the window	*/
										/*   expressed in Frame coords		*/
	ForceNextPrepare();
}

	
/******************************************************************************
 ResizeFrame

		Adjust a Pane's frame when the size of the pane changes. The delta
		rectangle parameter specifies the change in location of each side
		of the Pane. Positive values mean a move to the right or down;
		negative values mean a move to the left or up.

		This method always adjusts the width and height to the new size
		and puts the top left of the frame at (0, 0). This will cause a
		shift in the origin if the left or top sides of the pane move.
 ******************************************************************************/

void	CPane::ResizeFrame(
	register Rect	*delta)
{
										/* Adjust dimensions of the pane	*/
	width += delta->right - delta->left;
	height += delta->bottom - delta->top;

	frame.top = frame.left = 0;			/* Put top left of frame at (0, 0)	*/
	frame.right = width;
	frame.bottom = height;
	
	hOrigin -= delta->left;				/* Origin moves with the top left	*/
	vOrigin -= delta->top;				/*   corner of the pane				*/	
	
	ForceNextPrepare();		
}

	
/******************************************************************************
 CalcAperture {OVERRIDE}

		Determine the area of the pane where drawing may occur. This is
		the area within the frame that is not obscured by the bounds
		of enclosing views. The aperature is in Frame coordinates.
 ******************************************************************************/

        /*
         *  Calculate aperture of subviews when aperture of enclosure changes.
         */
         
         static void    Pane_CalcAperture(
            CPane           *thePane)
         {
            thePane->CalcAperture();
         }

void	CPane::CalcAperture()
{
	LongRect		enclAperture;			/* Aperture of pane's enclosure		*/
	LongRect		prevAperture;
	
		// TCL 1.1 Save the previous aperture before recalculating.
		// Since a pane's aperture is the intersection of its frame
		// and its enclosure's aperture, we know that if this
		// pane's aperture didn't change, then none if its
		// subpane's apertures will change either. 
		// So we recalc the aperture, and if it doesn't change,
		// there's no need to pass the message to the subpanes/
		// Thanks to Doug Wyatt for the tip.
		
	prevAperture = aperture;
	
										/* Get enclosure's aperture			*/
	itsEnclosure->GetAperture(&enclAperture);
	EnclToFrameR(&enclAperture);		/* Convert it to Frame coords		*/
	
										/* Aperture is the frame restricted	*/
										/*   by the enclosure's aperture	*/
	SectLongRect(&frame, &enclAperture, &aperture);
	
	ForceNextPrepare();
    
    if (!EqualLongRect( &prevAperture, &aperture) && itsSubviews != NULL)
        itsSubviews->DoForEach((EachFunc) Pane_CalcAperture);  /* Pass it on to subviews   */
}


/**** C O O R D I N A T E  T R A N S F O R M   M E T H O D S ****/

	
/******************************************************************************
 WindToFrame

		Convert a point from Window coords to the Frame coords of a Pane
 ******************************************************************************/

void	CPane::WindToFrame(
	Point		windPt,
	LongPt		*framePt)
{
	framePt->h = (long) windPt.h + hOrigin;
	framePt->v = (long) windPt.v + vOrigin;

}

	
/******************************************************************************
 WindToFrameR

		Convert a rectangle from Window coords to the Frame coords of a Pane
 ******************************************************************************/

void	CPane::WindToFrameR(
	Rect		*windRect,
	LongRect	*frameRect)
{
	QDToLongRect( windRect, frameRect);
	OffsetLongRect( frameRect, hOrigin, vOrigin);

}

	
/******************************************************************************
 FrameToWind

		Convert a point from the Frame coords of a Pane to Window coords
 ******************************************************************************/

void	CPane::FrameToWind(
	LongPt		*framePt, 
	Point		*windPt)
{
	windPt->h = (long)framePt->h - hOrigin;
	windPt->v = (long)framePt->v - vOrigin;
	
}

	
/******************************************************************************
 FrameToWindR

		Convert a rectangle from the Frame coords of a Pane to Window coords
 ******************************************************************************/

void	CPane::FrameToWindR(
	LongRect	*frameRect,
	Rect		*windRect)
{
	LongRect tmp = *frameRect;
	
	OffsetLongRect( &tmp, -hOrigin, -vOrigin);
	LongToQDRect( &tmp, windRect);
}

	
/******************************************************************************
 EnclToFrame

		Convert a point from the Frame coords of the enclosure to the
		Frame coords of the Pane
 ******************************************************************************/

void	CPane::EnclToFrame(
	LongPt		*thePoint)
{
	long		enclHOrigin;
	long		enclVOrigin;
	
	itsEnclosure->GetOrigin(&enclHOrigin, &enclVOrigin);
	thePoint->h += hOrigin - enclHOrigin;
	thePoint->v += vOrigin - enclVOrigin;
}

	
/******************************************************************************
 EnclToFrameR

		Convert a rectangle from the Frame coords of the enclosure to the
		Frame coords of the Pane
 ******************************************************************************/

void	CPane::EnclToFrameR(
	LongRect		*theRect)
{
	long		enclHOrigin;
	long		enclVOrigin;
	
	itsEnclosure->GetOrigin(&enclHOrigin, &enclVOrigin);
	OffsetLongRect(theRect, hOrigin - enclHOrigin,
						vOrigin - enclVOrigin);
}

	
/******************************************************************************
 FrameToEncl

		Convert a point from the Frame coords of a Pane to the
		Frame coords of its enclosure
 ******************************************************************************/

void	CPane::FrameToEncl(
	LongPt		*thePoint)
{
	long		enclHOrigin;
	long		enclVOrigin;
	
	itsEnclosure->GetOrigin(&enclHOrigin, &enclVOrigin);
	thePoint->h += enclHOrigin - hOrigin;
	thePoint->v += enclVOrigin - vOrigin;
}

	
/******************************************************************************
 FrameToEnclR

		Convert a rectangle from the Frame coords of a Pane to the
		Frame coords of its enclosure
 ******************************************************************************/

void	CPane::FrameToEnclR(
	LongRect		*theRect)
{
	long		enclHOrigin;
	long		enclVOrigin;
	
	itsEnclosure->GetOrigin(&enclHOrigin, &enclVOrigin);
	OffsetLongRect( theRect, enclHOrigin - hOrigin, enclVOrigin - vOrigin);
}

	
/******************************************************************************
 FrameToGlobalR {OVERRIDE}

		Convert a rectangle from Frame to Global coordinates
 ******************************************************************************/

void	CPane::FrameToGlobalR(
	LongRect *frameRect,
	Rect	*globalRect)
{
	Point		offset;
	
	LongToQDRect( frameRect, globalRect);
	
	offset = topLeft((**((WindowPeek)macPort)->contRgn).rgnBBox);
	
	OffsetRect( globalRect, offset.h - hOrigin, offset.v - vOrigin);
}

/******************************************************************************
 TrackMouse 

		Follow the mouse while it is being held down and repeatedly inform
		a specified Task of the mouse location
 ******************************************************************************/

void	CPane::TrackMouse(
	CMouseTask	*theTask,
	LongPt		*startPt,
	LongRect	*pinRect)
{
	LongPt			currPt;
	LongPt			prevPt;
	EventRecord		macEvent;
	Point			qdPt;
	
	Prepare();
	
	theTask->BeginTracking( startPt);
	currPt = prevPt = *startPt;

	while (StillDown()) {				/* Track while button is pressed	*/
	
										/* ??? We could call WaitNextEvent	*/
										/*   here to give background apps	*/
										/*   some CPU time. However, this	*/
										/*   makes mouse tracking a little	*/
										/*   bit sluggish.					*/
		Prepare();
		
		GetMouse(&qdPt);
		QDToFrame( qdPt, &currPt);
			
		PinInRect(pinRect, &currPt);
		theTask->KeepTracking( &currPt, &prevPt, startPt);
		prevPt = currPt;
	}
	
	Prepare();
	if (OSEventAvail(mUpMask, &macEvent)) {
		qdPt = macEvent.where;
		GlobalToLocal(&qdPt);
		QDToFrame( qdPt, &currPt);
	}
	
	PinInRect( pinRect, &currPt);
	theTask->EndTracking( &currPt, &prevPt, startPt);
}

/******************************************************************************
 QDToFrame

		Convert a point from QuickDraw coords to the Frame coords of a Pane.
		The qdPt parameter is assumed to be in whatever QuickDraw's idea
		of the Pane's local coordinate system is. Without long coordinates,
		this matches Frame coordinates. With long coordinates, we have to
		account for both the frame's origin and the partial offsetting
		we apply to QuickDraw's origin.
 ******************************************************************************/

void	CPane::QDToFrame(
	Point		qdPt,
	LongPt		*framePt)
{
	Prepare();
	
	if (usingLongCoord)
	{
		WindToFrame( qdPt, framePt);
		framePt->h -= thePort->portRect.left;
		framePt->v -= thePort->portRect.top;
	}
	else
		QDToLongPt( qdPt, framePt);
}

	
/******************************************************************************
 QDToFrameR

		Convert a rectangle from QuickDraw coords to the Frame coords of a Pane
 ******************************************************************************/

void	CPane::QDToFrameR(
	Rect		*qdRect,
	LongRect	*frameRect)
{
	Prepare();
	
	if (usingLongCoord)
	{
		WindToFrameR( qdRect, frameRect);
		OffsetLongRect( frameRect, -thePort->portRect.left, -thePort->portRect.top);
	}
	else
		QDToLongRect( qdRect, frameRect);
}

	
/******************************************************************************
 FrameToQD

		Convert a point from the Frame coords of a Pane to QuickDraw coords
 ******************************************************************************/

void	CPane::FrameToQD(
	LongPt		*framePt, 
	Point		*qdPt)
{
	Prepare();
	
	if (usingLongCoord)
	{
		FrameToWind( framePt, qdPt);
		qdPt->h += thePort->portRect.left;
		qdPt->v += thePort->portRect.top;
	}
	else
		LongToQDPt( framePt, qdPt);
}

	
/******************************************************************************
 FrameToQDR

		Convert a rectangle from the Frame coords of a Pane to QuickDraw coords
 ******************************************************************************/

void	CPane::FrameToQDR(
	LongRect	*frameRect,
	Rect		*qdRect)
{
	Prepare();

	if (usingLongCoord)
	{
		FrameToWindR( frameRect, qdRect);	
		OffsetRect( qdRect, thePort->portRect.left, thePort->portRect.top);
	}
	else
		LongToQDRect( frameRect, qdRect);

}


/******************************************************************************
 SectAperture

	Clip the LongRect to the pixels visible thru the aperture, and return
	the result in a QD rect. Returns TRUE if resulting rect is non-empty.
 ******************************************************************************/

Boolean CPane::SectAperture( LongRect *srcRect, Rect *destRect)
{
	LongRect	tmp;
	Rect		tmpAperture;
	
	Prepare();
	if (SectLongRect( &aperture, srcRect, &tmp))
	{
		FrameToQDR( &tmp, destRect);
		return !EmptyRect( destRect);
	}
	else
	{
		SetRect( destRect, 0, 0, 0, 0);
		return FALSE;
	}
}	

/******************************************************************************
 GetHelpResID

	Return the resource ID of the 'hrct' resource that has the 
	help balloon info for this pane. This method passes the message to the
	pane's enclosing window, which return the ID to be used for all panes
	in the window.
	
 ******************************************************************************/

short CPane::GetHelpResID( void)
{
	return GetWindow()->GetHelpResID();
}
