/******************************************************************************
 CPaneBorder.c

		CPaneBorder is a helper class for CPane. It draws a border
		outside the frame of a client frame when requested. It also
		can calculate the rectangle that encompasses the pane and its
		border. CPaneBorder makes no assumptions about which coordinate
		system to use: it always calculates and draws relative to 
		a rectangle that is passed to it.
		CPaneBorder can draw rectangular, shadowed rectangular,
		rounded rectangular, or oval borders. It may be subclassed
		to provide additional borders.
		
	SUPERCLASS = CObject
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CPaneBorder.h"
#include "Global.h"
#include "TCLUtilities.h"

/******************************************************************************
 IPaneBorder
 
 	Initialize a PaneBorder. The borderFlags are taken from the
 	argument, and the other border attributes are initialized
 	with defaults. All the defaults may be changed via methods.
******************************************************************************/

void CPaneBorder::IPaneBorder( short borderFlags)
{
	borderPen.h = borderPen.v = 1;
	shadowOffset.h = shadowOffset.v = 2;
	shadowPen.h = shadowPen.v = 2;
	roundDiameter.h = roundDiameter.v = 16;
	doShadow = FALSE;
	
	SetRect( &margin, 0, 0, 0, 0);
	SetBorderFlags( borderFlags);
	SetPattern( black);
	
}	/* CPaneBorder::IPaneBorder */

/******************************************************************************
 IResPaneBorder
 
 	Initialize a pane border from a 'PBrd' resource.
 	
******************************************************************************/

	static PatHandle GetPat( short patID)
	{
		return GetPattern( patID);
	}

void CPaneBorder::IResPaneBorder( short resID)
{
	Boolean			savedAlloc;
	PaneBorderTemp	**template = NULL, *t;
	short			patID = 0;
	PatHandle		pat = NULL;
	
	TRY
	{
		doShadow = FALSE;
		savedAlloc = SetAllocation( kAllocCanFail);
		template = (PaneBorderTemp**) GetResource( kPaneBorderRes, resID);
		if (template && ((**template).patID))
		{
			patID = (**template).patID;
			pat = GetPat( patID);
		}
		SetAllocation( savedAlloc);
		
		FailNILRes( template);
		if (patID > 0) FailNILRes( pat);
		
		if (pat) 
		{
			SetPattern( **pat);
			HPurge( (Handle) pat);
		}
		else
			SetPattern( black);
			
		t = *template;
		SetBorderFlags( t->borderFlags);
		SetPenSize( t->borderPen.h, t->borderPen.v);
		if ((t->shadowPen.h > 0) && (t->shadowPen.v > 0))
		{
			SetShadow( t->shadowOffset.h, t->shadowOffset.v,
					t->shadowPen.h, t->shadowPen.v);
		}
		if (borderFlags & kBorderRoundRect)
		{
			SetRounding( t->roundDiameter.h, t->roundDiameter.v);
		}
		SetMargin( &t->margin);
	}
	CATCH
	{
		ForgetResource( template);
		ForgetResource( pat);
	}
	ENDTRY;
	
}

/******************************************************************************
 SetPattern
 
 	Set the pattern used for drawing all border parts.
******************************************************************************/

void CPaneBorder::SetPattern( Pattern aPattern)
{
	BlockMove( aPattern, penPat, sizeof(Pattern));

}	/* CPaneBorder::SetPattern */

/******************************************************************************
 GetPattern
 
 	Return the current pattern.
******************************************************************************/

void CPaneBorder::GetPattern( Pattern aPattern)
{
	BlockMove( penPat, aPattern, sizeof(Pattern));


}	/* CPaneBorder::GetPattern */

/******************************************************************************
 SetBorderFlags
 
 	Set the border flags. The currently defined flags are:
 		
		kBorderNone			0x00		no border
		kBorderLeft			0x01		line on left side
		kBorderTop			0x02		line on top side
		kBorderRight		0x04		line on right side
		kBorderBottom		0x08		line on bottom
		kBorderOval			0x10		oval border
		kBorderRoundRect	0x20		rounded rectangle border
		kBorderRsrv1		0x40		reserved
		kBorderRsrv2		0x80		reserved
		
	For convenience the constant kBorderFrame is defined to provide a
	rectangular border. It is equivalent to specifying
	kBorderLeft+kBorderTop+kBorderRight+kBorderBottom.

******************************************************************************/

void CPaneBorder::SetBorderFlags( long aBorderFlags)
{	
	borderFlags = aBorderFlags;
	
	/* Shadowing only applies to the kBorderFrame setting */
	if (borderFlags != kBorderFrame)
		doShadow = FALSE;
		
}	/* CPaneBorder::SetBorderFlags */

/******************************************************************************
 GetBorderFlags
 
 	Return the current border flags
******************************************************************************/

long CPaneBorder::GetBorderFlags( void)
{
	return borderFlags;

}	/* CPaneBorder::GetBorderFlags */

/******************************************************************************
 SetPenSize
 
 	Sert the pen size used to draw the border.
******************************************************************************/

void CPaneBorder::SetPenSize( short penWidth, short penHeight)
{
	SetPt( &borderPen, penWidth, penHeight);

}	/* CPaneBorder::SetPenSize */

/******************************************************************************
 GetPenSize
 
 	Return the pen size used to draw the border.
******************************************************************************/

void CPaneBorder::GetPenSize( short *penWidth, short *penHeight)
{
	*penWidth = borderPen.h;
	*penHeight = borderPen.v;

}	/* CPaneBorder::GetPenSize */

/******************************************************************************
 SetShadow
 
 	Set the shadow attributes. hOffset and vOffset specify the offset of the
 	shadow, i.e. how many pixels the shadow rectangle appears to be
 	offset from the pane frame. Width and height specify the pen size
 	used to draw the shadow.
******************************************************************************/

void CPaneBorder::SetShadow( short hOffset, short vOffset, short width, short height)
{
	SetBorderFlags( kBorderFrame);
	doShadow = TRUE;
	SetPt( &shadowOffset, hOffset, vOffset);
	SetPt( &shadowPen, width, height);
	
}	/* CPaneBorder::SetShadow */

/******************************************************************************
 GetShadow
 
 	Return the current shadow attributes.
******************************************************************************/

void CPaneBorder::GetShadow( short *hOffset, short *vOffset, short *width, short *height)
{
	*hOffset = shadowOffset.h;
	*vOffset = shadowOffset.v;
	*width = shadowPen.h;
	*height = shadowPen.v;

}	/* CPaneBorder::GetShadow */

/******************************************************************************
 SetRounding
 
 	Set the rounding diameters. These are ultimately passed to FrameRoundRect.
 	Also sets the border style to kBorderRoundRect.
******************************************************************************/

void CPaneBorder::SetRounding( short hDiameter, short vDiameter)
{
	SetBorderFlags( kBorderRoundRect);
	SetPt( &roundDiameter, hDiameter, vDiameter);
	
}	/* CPaneBorder::SetRounding */

/******************************************************************************
 GetRounding
 
 	Returns the current rounding diameters.
******************************************************************************/

void CPaneBorder::GetRounding( short *hDiameter, short *vDiameter)
{
	*hDiameter = roundDiameter.h;
	*vDiameter = roundDiameter.v;

}	/* CPaneBorder::GetRounding */

/******************************************************************************
 SetMargin
 
 	Set the margin rectangle. Each side of the margin rectangle specifies
 	the amount whitespace to add around the pane. All the sides should be
 	non-negative so that the border is outside the pane's frame.
 	For example, to have a one pixel thick black border and another
 	pixel of whitespace, specify a border with a width and height of 1, and
 	specify a 1 pixel margin. 
******************************************************************************/

void CPaneBorder::SetMargin( Rect *aMargin)
{
	margin = *aMargin;
	
}	/* CPaneBorder::SetMargin */

/******************************************************************************
 GetMargin
 
 	Return the current border margin
******************************************************************************/

void CPaneBorder::GetMargin( Rect *aMargin)
{
	*aMargin = margin;

}	/* CPaneBorder::GetMargin */

/******************************************************************************
 CalcBorderRect
 
 	Modify the given rectangle by expanding the sides as needed to accomodate
 	the border. The size of the border depends upon the currently specified 
 	border attributes.
******************************************************************************/

void CPaneBorder::CalcBorderRect( Rect *paneFrame)
{
	Rect	exterior;
	
	exterior = *paneFrame;
	
	exterior.left += margin.left;
	exterior.top += margin.top;
	exterior.right += margin.right;
	exterior.bottom += margin.bottom;
	
	if ((borderFlags == kBorderFrame) || (borderFlags == kBorderRoundRect) ||
		(borderFlags == kBorderOval))
	{
		exterior.left -= borderPen.h;
		exterior.right += borderPen.h;
		exterior.top -= borderPen.v;
		exterior.bottom += borderPen.v;
		if (doShadow)
		{
			exterior.right += shadowPen.h;
			exterior.bottom += shadowPen.v;
		}
	}
	else
	{
		if (borderFlags & kBorderLeft)
			exterior.left -= borderPen.h;
		if (borderFlags & kBorderRight)
			exterior.right += borderPen.h;
		if (borderFlags & kBorderTop)
			exterior.top -= borderPen.v;
		if (borderFlags & kBorderBottom)
			exterior.bottom += borderPen.v;
	}
	
	*paneFrame = exterior;
	
}	/* CPaneBorder::CalcBorderRect */

/******************************************************************************
 DrawBorder
 
 	Draw the border with paneFrame. Assumes the pane is already Prepared().
******************************************************************************/

void CPaneBorder::DrawBorder( Rect* paneFrame)
{
	Rect	r = *paneFrame;
	
	CalcBorderRect( &r);
	
	PenPat( penPat);
	PenSize( borderPen.h, borderPen.v);
	
	if (borderFlags == kBorderFrame)
	{
		if (doShadow)
		{
			r.right -= shadowPen.h;
			r.bottom -= shadowPen.v;
			FrameRect( &r);
			
			PenSize( shadowPen.h, shadowPen.v);
			MoveTo( r.left + shadowOffset.h, r.bottom);
			//LineTo( r.right - shadowPen.h, r.bottom);
			LineTo( r.right, r.bottom);
			MoveTo( r.right, r.top + shadowOffset.v);
			//LineTo( r.right, r.bottom - shadowPen.v);
			LineTo( r.right, r.bottom);
		}
		else
			FrameRect( &r);
		
	}
	else if (borderFlags & kBorderRoundRect)
		FrameRoundRect( &r, roundDiameter.h, roundDiameter.v);
	else if (borderFlags & kBorderOval)
		FrameOval( &r);
	else
	{
		r.bottom -= borderPen.v;
		r.right -= borderPen.h;		
		if (borderFlags & kBorderTop)
		{
			MoveTo( r.left, r.top);
			LineTo( r.right, r.top);
		}
		if (borderFlags & kBorderLeft)
		{
			MoveTo( r.left, r.top);
			LineTo( r.left, r.bottom);
		}
		if (borderFlags & kBorderRight)
		{
			MoveTo( r.right, r.top);
			LineTo( r.right, r.bottom);
		}
		if (borderFlags & kBorderBottom)
		{
			MoveTo( r.left, r.bottom);
			LineTo( r.right, r.bottom);
		}		
	}
	PenNormal();	/* TCL 1.1.1 DLP 9/26/91 */

}	/* CPaneBorder::DrawBorder */

