/******************************************************************************
 CPanorama.c

							The Panorama Class
	
	Abstract class for object which may have an extent which is larger
	than what is visible on the screen. Typically, this means panes
	that can be scrolled.
	
		
	SUPERCLASS = CPane
	
	Copyright  1989 Symantec Corporation. All rights reserved.
	
	TCL 1.1 CHANGES
	[
		- Changed interface for SetBounds, GetBounds, SetPosition, GetPosition,
		  GetHomePosition, ScrollTo, SetBounds, and GetBounds 
		  to accomodate 32 bit coordinates.	
		- implemented ClipToAperture method	 
		- Because SetBounds takes a LongRect,  Panoramas can now have
		  32-bit dimensions. If your Panorama subclass can ever extend beyond
		  the normal QuickDraw 16-bit limit, then you should call
		  UseLongCoordinates( TRUE) when you initialize it. When long
		  coordinates are in use, you must convert your local Frame
		  coordinates to QuickDraw coordinates before drawing. The
		  methods FrameToQD and FrameToQDR can do this for you. 
	]

 ******************************************************************************/

#include "Global.h"
#include "CPanorama.h"
#include "CScrollPane.h"
#include "CPrinter.h"
#include "Constants.h"      
#include "CWindow.h"
#include "CList.h"
	
/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IPanorama

		Initialize a Panorama object
 ******************************************************************************/

void	CPanorama::IPanorama(
	CView			*anEnclosure,
	CBureaucrat		*aSupervisor,
	short			aWidth,
	short			aHeight,
	short			aHEncl,
	short			aVEncl,
	SizingOption	aHSizing,
	SizingOption	aVSizing)
{
	CPane::IPane(anEnclosure, aSupervisor,
					aWidth, aHeight, aHEncl, aVEncl, aHSizing, aVSizing);
	
	printClip = clipPAGE;
	bounds = frame;
	position.h = bounds.left;
	position.v = bounds.top;
	hScale = vScale = 1;
	itsScrollPane = NULL;	
}
	

/******************************************************************************
 IViewTemp {OVERRIDE}

		Initialize a Panorama using a template
 ******************************************************************************/

void	CPanorama::IViewTemp(
	CView			*anEnclosure,		/* Enclosing view					*/
	CBureaucrat		*aSupervisor,		/* Boss in chain of command			*/
	Ptr				viewData)			/* Template with View data			*/
{
	register PanoramaTempP	p;
	
	p = (PanoramaTempP) viewData;		/* Initialize superclass			*/
	inherited::IViewTemp(anEnclosure, aSupervisor, (Ptr) &p->sPaneTemp);
	
							/* Set instance vars from template	*/

	QDToLongRect( &p->bounds, &bounds);
	hScale = p->hScale;
	vScale = p->vScale;
	QDToLongPt( p->position, &position);
	itsScrollPane = NULL;
}


/**** A C C E S S I N G   M E T H O D S ****/


/******************************************************************************
 GetExtent

		Get the total horizontal and vertical size of the Panorama.
		Subclasses can use any unit of size they wish.
		For a graphics program, this could be the width and height of a
		page in pixels. For a text program, this could be the pixel width
		of a page and the number of lines in the document.
		
		This method uses the lengths of the sides of the bounds rectangle
		as the extents. Subclasses must override this method if they
		define their extents differently. An example is a very large
		Panorama which uses a 32-bit coordinate system (QuickDraw rectangles
		use only 16-bit values).
 ******************************************************************************/

void	CPanorama::GetExtent(
	long		*theHExtent,
	long		*theVExtent)
{
	*theHExtent = bounds.right - bounds.left;
	*theVExtent = bounds.bottom - bounds.top;
}


/******************************************************************************
 GetFramePosition

		Get the position of the top left corner of the frame with respect
		to the entire panorama. That is, what portion of the panorama is
		visible through its frame?
		
		This method uses the "position" instance variable. Subclasses should
		override this method if they use a different means of monitoring
		the frame postion. An example is a very large Panorama which uses
		a 32-bit coordinate system (QuickDraw points use only 16-bit values).
 ******************************************************************************/

void	CPanorama::GetFramePosition(
	long		*theHPos,
	long		*theVPos)
{
	*theHPos = position.h - bounds.left;
	*theVPos = position.v - bounds.top;
}


/******************************************************************************
 GetFrameSpan

		Return the number of units (the same units as used when defining
		the extent -- set GetExtent method) spanned by the frame of a
		Panorama. This method assumes that units in each direction are of
		constant size in pixels, as specified by the "hScale" and "vScale"
		instance variables. Subclasses must override this method if they
		use variable-sized units. An example is a spreadsheet which supports
		variable-sized cells and uses a cell as a unit.
 ******************************************************************************/

void	CPanorama::GetFrameSpan(
	short		*theHSpan,
	short		*theVSpan)
{										/* Integer division truncates		*/
	*theHSpan = width / hScale;			/*   We get the number of complete	*/
	*theVSpan = height / vScale;		/*   units spanned by the frame		*/
}


/******************************************************************************
 SetBounds

		Set the bounds of the Panorama. The bounds define the overall
		size of the data displayed by the Panorama and determine the
		Pane coordinates. 
		
 ******************************************************************************/

void	CPanorama::SetBounds(
	LongRect		*aBounds)
{
	bounds = *aBounds;					/* Set instance variable			*/
	
	ASSERT( position.v <= bounds.bottom);
	
	if (itsScrollPane != NULL) {
		itsScrollPane->AdjustScrollMax();
	}
}


/******************************************************************************
 GetBounds

		Retrieve the bounds of the Panorama.
 ******************************************************************************/

void	CPanorama::GetBounds(
	LongRect		*theBounds)
{
	*theBounds = bounds;				/* Pass back copy of instance var	*/
}


/******************************************************************************
 SetPosition

		Set the position of the frame in relation to the Panorama. Position
		specifies the location of the top left corner of the frame in
		Pane coordinates.
 ******************************************************************************/

void	CPanorama::SetPosition(
	LongPt		*aPosition)
{
	position = *aPosition;				/* Set instance variable			*/
	
	if (itsScrollPane != NULL) {
		itsScrollPane->Calibrate();
	}
}


/******************************************************************************
 GetPosition

		Retrieve the position of the frame in relation to the Panorama.
 ******************************************************************************/

void	CPanorama::GetPosition(
	LongPt		*thePosition)
{
	*thePosition = position;			/* Pass back copy of instance var	*/
}


/******************************************************************************
 SetScales

		Set the horizontal and vertical scale factors. The scale factors
		specify the number of pixels per Pane unit.
 ******************************************************************************/

void	CPanorama::SetScales(
	short		aHScale,
	short		aVScale)
{
    
    hScale = Max(aHScale, 1);               
    vScale = Max(aVScale, 1);
	
	if (itsScrollPane != NULL) {
		itsScrollPane->AdjustScrollMax();
	}
}


/******************************************************************************
 GetScales

		Retrieve the horizontal and vertical scale factors.
 ******************************************************************************/

void	CPanorama::GetScales(
	short		*theHScale,
	short		*theVScale)
{
	*theHScale = hScale;				/* Return copies of instance vars	*/
	*theVScale = vScale;
}


/******************************************************************************
 SetScrollPane

		Specify a Scroll Pane for controlling a Panorama
 ******************************************************************************/

void	CPanorama::SetScrollPane(
	CScrollPane		*aScrollPane)
{
	itsScrollPane = aScrollPane;		/* Set instance variable			*/
}


/******************************************************************************
 GetHomePosition

		Retrieve the location of the top left corner of the Panorama
		in Pane coordinates. This method assumes that the bounds instance
		variable defines the Pane coordinates.
 ******************************************************************************/

void	CPanorama::GetHomePosition(
	LongPt		*theHomePos)
{
	*theHomePos = topLeftL(bounds);
}


/******************************************************************************
 GetPixelExtent {OVERRIDE}

		Retrieve the dimensions of a Panorama in pixels. This method
		assumes that the bounds instance variable defines the size
		of the Panorama and that each unit has a constant size (hScale
		horizontally and vScale vertically).
 ******************************************************************************/

void	CPanorama::GetPixelExtent(
	long		*hExtent,
	long		*vExtent)
{
	*hExtent = (bounds.right - bounds.left) * hScale;
	*vExtent = (bounds.bottom - bounds.top) * vScale;
}


/**** C A L I B R A T I N G   M E T H O D S ****/

	
/******************************************************************************
 ResizeFrame {OVERRIDE}

		Adjust a Panorama's frame when its size changes. The delta
		rectangle parameter specifies the change in location of each side
		of the Pane. Positive values mean a move to the right or down;
		negative values mean a move to the left or up.
 ******************************************************************************/

void	CPanorama::ResizeFrame(
	register Rect		*delta)
{
										/* Adjust dimensions of the pane	*/
	width += delta->right - delta->left;
	height += delta->bottom - delta->top;
	
	frame.left += delta->left;			/* Shift edges by delta				*/
	frame.top += delta->top;
	frame.right += delta->right;
	frame.bottom += delta->bottom;
	
    
    position.h += delta->left / hScale; /* Scroll position is tied to the   */
    position.v += delta->top / vScale;  /* left of the frame.               */
    
    ForceNextPrepare();
}


/**** S C R O L L I N G   M E T H O D S ****/


/******************************************************************************
 Scroll

		Scroll a Panorama within its frame by the specified number of
		units in each direction. Units are Panorama-specific (see
		GetExtent method). This method assumes that units are fixed in
		size (hScale and vScale).
 ******************************************************************************/

		/*
		 * Inform subviews when enclosure scrolls
		 */
         
         
        void    Pane_EnclosureScrolled(
			CPane		*thePane,
			LongPt		*offset)
		{
			thePane->EnclosureScrolled(offset->h, offset->v);
		}
		
void	CPanorama::Scroll(
	long		hDelta,					/* Units to scroll horizontally		*/
	long		vDelta,					/* Units to scroll vertically		*/
	Boolean		redraw)					/* Redraw Panorama after scrolling?	*/
{
	long				hPixels;
	long				vPixels;
	Rect				tempRect;
	extern RgnHandle	gUtilRgn;
	LongPt				offset;
	
	hPixels = hDelta * hScale;
	vPixels = vDelta * vScale;
	
	if (redraw) {
		
		if ((Abs(hPixels) < width) && (Abs(vPixels) < height)) /* TCL 1.1.2 DLP 12/13/91 */
		{
			Prepare();	
			FrameToQDR( &aperture, &tempRect);

			ScrollRect(&tempRect, -hPixels, -vPixels, gUtilRgn);
			InvalRgn(gUtilRgn);
		}
		else
			Refresh();
	}

	OffsetLongRect(&frame, hPixels, vPixels);
	OffsetLongRect(&aperture, hPixels, vPixels);
	
	position.h += hDelta;
	position.v += vDelta;

	hOrigin += hPixels;
	vOrigin += vPixels;
	
	ForceNextPrepare();
	
	if (itsSubviews != NULL) {
		offset.h = hPixels;
		offset.v = vPixels;
		itsSubviews->DoForEach1((EachFunc1) Pane_EnclosureScrolled, (long) &offset);
	}
	    
	if (redraw)
        ((CWindow *) (((WindowPeek) macPort)->refCon))->Update();
}


/******************************************************************************
 ScrollTo

		Scroll a Panorama so that the top left of its frame is at the
		specified position
 ******************************************************************************/
		
void	CPanorama::ScrollTo(
	LongPt		*aPosition,
	Boolean		redraw)
{
	Scroll(aPosition->h - position.h, aPosition->v - position.v, redraw);

	if (itsScrollPane != NULL) {
		itsScrollPane->AdjustScrollMax();
		itsScrollPane->Calibrate();
	}
}


/******************************************************************************
 ScrollToSelection

		Scroll a Panorama so that the current selection is visible within
		the frame. Subclasses must override this method.
 ******************************************************************************/
		
void	CPanorama::ScrollToSelection()
{
}										/* Null Method						*/


/******************************************************************************
 AutoScroll

		Automatically scroll a Panorama during mouse-down selecting.
		Return value indicates whether or not scrolling occurred.
 ******************************************************************************/
		
Boolean		CPanorama::AutoScroll(
	LongPt		*mouseLoc)				/* Mouse location in Frame coords	*/
{
	short		hDelta = 0;
	short		vDelta = 0;
	short		hSpan;
	short		vSpan;
	short		hStep = 1;
	short		vStep = 1;
	
	GetFrameSpan(&hSpan, &vSpan);
	if (itsScrollPane != NULL) {
		itsScrollPane->GetSteps(&hStep, &vStep);
	}
	
	
	if (mouseLoc->h < frame.left) {
		hDelta = Max(-hStep, bounds.left - position.h);
		if (hDelta > 0) {
			hDelta = 0;
		}
	} else if (mouseLoc->h > frame.right) {
		hDelta = Min(hStep, bounds.right - position.h - hSpan);
		if (hDelta < 0) {
			hDelta = 0;
		}
	}
	
	if (mouseLoc->v < frame.top) {
		vDelta = Max(-vStep, bounds.top - position.v);
		if (vDelta > 0) {
			vDelta = 0;
		}
	} else if (mouseLoc->v > frame.bottom) {
		vDelta = Min(vStep, bounds.bottom - position.v - vSpan);
		if (vDelta < 0) {
			vDelta = 0;
		}
	}
	
	if ( (hDelta != 0) || (vDelta != 0) ) {
		Scroll(hDelta, vDelta, TRUE);
		if (itsScrollPane != NULL) {
			itsScrollPane->Calibrate();
        }
        Prepare(); 
		return(TRUE);
	} else {
		return(FALSE);
	}
}


/******************************************************************************
 DoKeyDown  (OVERRIDE)

        Implement Home, End, Page Up, and Page Down scrolling keys on
        extended keyboards.
 ******************************************************************************/
 
void CPanorama::DoKeyDown(
    char        theChar,
    Byte        keyCode,
    EventRecord *macEvent)
{
    LongPt   thePosition;
    long    theHExtent, theVExtent;
    
    switch (keyCode) {
    
        case KeyHome:
            if (itsScrollPane != NULL) {
                GetHomePosition(&thePosition);
                ScrollTo( &thePosition, TRUE);
            }
            break;
            
        case KeyEnd:
            if (itsScrollPane != NULL) {
                GetExtent(&theHExtent, &theVExtent);
                thePosition.h = Max(0, theHExtent - itsScrollPane->hSpan);
                thePosition.v = Max(0, theVExtent - itsScrollPane->vSpan);
                ScrollTo( &thePosition, TRUE);
            }
            break;
            
        case KeyPageUp:
            if (itsScrollPane != NULL) {
                itsScrollPane->DoVertScroll(inPageUp);
                itsScrollPane->AdjustScrollMax();
            }
            break;
            
        case KeyPageDown:
            if (itsScrollPane != NULL) {
                itsScrollPane->DoVertScroll(inPageDown);
                itsScrollPane->AdjustScrollMax();
            }
            break;
            
        default:
            inherited::DoKeyDown(theChar, keyCode, macEvent);
            break;
    }
}

        
/**** P R I N T I N G   M E T H O D S ****/
    
    
/******************************************************************************
 Paginate {OVERRIDE}

	Determine how many pages to print. For pagination purpose, the panorama
	is divided into horizontal and vertical strips of pages. The size of
	each strip may be set separately. The default is to divide the extent
	into equal size strips.
 ******************************************************************************/

void CPanorama::Paginate( CPrinter *aPrinter, short pageWidth, short pageHeight)
{
	long	pixWidth, pixHeight;
	short	hStrips, vStrips;
	    
	    				/* Get the full extent of the panorama		*/    
	GetPixelExtent( &pixWidth, &pixHeight);

						/* determine how many horizontal and 		*/
						/* vertical strips are required				*/
	hStrips = pixWidth / pageWidth + 1;
	vStrips = pixHeight / pageHeight + 1;

						/* tell the printer how many strips, strips	*/
						/* are initially set to page width and		*/
						/* height									*/
	aPrinter->SetStrips( hStrips, vStrips);
	aPrinter->SetAllStripWidths( pageWidth);
	aPrinter->SetAllStripHeights( pageHeight);

}

/******************************************************************************
 AboutToPrint {OVERRIDE}

		The specified range of pages is about to be printed.
 ******************************************************************************/

void	CPanorama::AboutToPrint(
	short		*firstPage,
	short		*lastPage)
{
	savePosition = position;			/* Save current position since it	*/
										/*   may have to be changed during	*/
										/*   the print loop					*/
										
 	inherited::AboutToPrint( firstPage, lastPage);
}


/******************************************************************************
 PrintPage {OVERRIDE}

		Print the specified page.
 ******************************************************************************/

void	CPanorama::PrintPage(
	short		pageNum,				/* Number of the page to print		*/
	short		pageWidth,				/* Width of printed page in pixels	*/
	short		pageHeight,				/* Height of page in pixels			*/
	CPrinter	*aPrinter)
{
	GrafPtr		savePort;
	LongRect	area;
	Rect		qdArea;
	LongPt		scrollPos;
	
	aPrinter->GetPageArea( pageNum, &area);
	
	/* scroll to the start of the page	*/
			
	GetPort(&savePort);
	scrollPos = topLeftL( area);
	scrollPos.h /= hScale;
	scrollPos.v /= vScale;
	ScrollTo( &scrollPos, FALSE);
	SetPort(savePort);

	/* setup class variable cPageArea with portion of page to be printed */

	FrameToWindR( &area, &CPane::cPageArea);

	ForceNextPrepare();
	Prepare();	

	FrameToWindR( &area, &qdArea);
	DrawAll( &qdArea);

	SetRect( &CPane::cPageArea, 0, 0, 0, 0);
}
	

/******************************************************************************
 DonePrinting {OVERRIDE}

		Print loop has been completed. Perform any necessary clean-up.
 ******************************************************************************/

void	CPanorama::DonePrinting()
{
	LongPt	savePos = savePosition;
	
	inherited::DonePrinting();

	if (!EqualLongPt(&savePos, &position)) {
		ScrollTo( &savePos, FALSE);
	}
}

