/******************************************************************************
 CPrinter.c

							The Printer Class
	
	SUPERCLASS = CObject
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	TCL 1.1 CHANGES
		- implemented ClosePrintMgr, SetPrintDir, HavePagination, ResetPagination,
		  SetStrips, SetHorizPageBreak, SetVertPageBreak, SetAllStripWidths,
		  SetStripHeight, SetAllStripHeights, GetStripCount, PageNumToStrips,
		  GetPageStart, GetPageArea methods
	] 

 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "CPrinter.h"
#include "TCLUtilities.h"
#include "TBUtilities.h"
#include "CRunArray.h"
#include "CDocument.h"
#include "LongCoordinates.h"

/*** Global Variables ***/

extern CursHandle	gWatchCursor;		/* Watch cursor for waiting			*/


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IPrinter

		Initialize a Printer object. Printer can be passed a handle to
		an existing Toolbox print recordl. If handle is NULL, a new
		print record with default values is created.
		
		Boolean result indicates whether the print record passed as a
		parameter was changed because it was incompatible with the
		currently installed printer. This allows documents which associate
		a printer record with a file to update itself if desired.
 ******************************************************************************/

Boolean		CPrinter::IPrinter(
	CDocument		*aDocument,
	THPrint			aMacTPrint)
{
	Boolean		changed = FALSE;		/* Has print record changed?		*/
	OSErr		err;
	
	savedResFile = -1;
	printDocOpen = printPageOpen = FALSE;
	macPrintPort = NULL;
	printMgrOpen = FALSE;
	
	OpenPrintMgr( FALSE);				/* Open printer driver				*/

	if ( !PrError() ) {					/* Don't bother to report an error	*/
										/*   Wait until the user actually	*/
										/*   tries to perform some print	*/
										/*   action.						*/
	
		if (aMacTPrint == NULL) {		/* Create new Toolbox print record	*/
			aMacTPrint = (THPrint) NewHandleCanFail( sizeof(TPrint));
			FailNIL( aMacTPrint);
			PrintDefault(aMacTPrint);	/* Set default print record values	*/

		} else {						/* Check validity of existing		*/
										/*   print record					*/
			changed = PrValidate(aMacTPrint);
		}

		/*
		 * If an error occured in PrintDefault or PrValidate we want
		 * to inform the user, but we don't want to signal failure. 
		 * Therefore we check the error and if needed, call ErrorAlert.
		 */
		 err = PrError();
		 if (err != noErr)
		 	ErrorAlert( err, 0);
	}
		
	ClosePrintMgr();						/* Close the printer driver			*/
	
	itsDocument = aDocument;			/* Set instance variables			*/
	macTPrint = aMacTPrint;
	
	printDirection = printVert;
	
	itsStripWidths = new(CRunArray);
	itsStripWidths->IRunArray();
	
	itsStripHeights = new(CRunArray);
	itsStripHeights->IRunArray();

	return(changed);
}


/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a Printer object. The Toolbox print record is thrown
		out also, even if it was passed a valid handle in the IPrinter
		method.
 ******************************************************************************/

void	CPrinter::Dispose()
{
	ForgetHandle( macTPrint);			/* Throw out Toolbox print record	*/
	
	ForgetObject( itsStripWidths);
	ForgetObject( itsStripHeights);

	inherited::Dispose();				/* Pass message on to superclass	*/
}


/******************************************************************************
 OpenPrintMgr

		Prepare the Mac Printing Manager by opening its driver and resource
		file. An error here usually means that the user has not selected
		a printer from the Chooser or does not have the proper drivers in
		the System folder. This method notifies the user with an Alert in
		such a case.
 ******************************************************************************/

Boolean		CPrinter::OpenPrintMgr( Boolean fCheckFailure)
{
	ASSERT( !printMgrOpen);
	
	savedResFile = CurResFile();
	printMgrOpen = TRUE;
	PrOpen();
	
	if (fCheckFailure && PrError()) 
	{
		PositionDialog('ALRT', ALRT_NOPRINTER);
		StopAlert(ALRT_NOPRINTER, NULL);
		return(FALSE);
	} 
	else 
	{
		return(TRUE);
	}
}
	
/******************************************************************************
 ClosePrintMgr

	Close the print mgr, balancing out any other PrintMgr open calls that
	were made.
 ******************************************************************************/

void CPrinter::ClosePrintMgr( void)
{
	if (printMgrOpen)
	{
		if (printDocOpen)
		{	
			if (printPageOpen)
			{
				printPageOpen = FALSE;
				PrClosePage( macPrintPort);
			}
			printDocOpen = FALSE;
			PrCloseDoc( macPrintPort);
		}
		printMgrOpen = FALSE;
		macPrintPort = NULL;
		PrClose();
		if (savedResFile != -1)
		{
			UseResFile( savedResFile);
			savedResFile = -1;
		}
	}
}

/******************************************************************************
 GetPrintRecord

		Returns a handle to the Toolbox print record associated with
		a Printer object. Other objects should treat this record as
		read-only. Note that this method calls PrValidate before
		returning the print record. This is important because the user
		could have picked another printer via the Chooser since the
		last time PrValidate was called.
 ******************************************************************************/

THPrint		CPrinter::GetPrintRecord()
{
	OpenPrintMgr( TRUE);			/* Open the printer driver			*/
	if ( !PrError() ) {
		
									/* Check validity of print record	*/
		PrValidate(macTPrint);
		FailOSErr( PrError());
	}
		
	ClosePrintMgr();					/* Close the printer driver			*/
	
	return(macTPrint);					/* Return instance variable			*/
}


/******************************************************************************
 GetPageInfo

		Returns information about the paper size and printable area of
		the paper. This info is required for setting accurate margins.
		The paper and page rectangles are specified in dots. The hRes and
		vRes values indicate the number of dots per inch. Thus dividing
		the width of the paperRect by hRes gives the width of the
		paper in inches.
		
		Note that this method calls PrValidate before returning the print
		record. This is important because the user could have picked another
		printer via the Chooser since the last time PrValidate was called.
		
		This method extracts the pertinent information from the print record.
		The programmer could also use the GetPrintRecord command and
		access the information directly.
 ******************************************************************************/

void	CPrinter::GetPageInfo(
	Rect		*paperRect,				/* The physical paper size			*/
	Rect		*pageRect,				/* Printable area of the page		*/
	short		*hRes,					/* Horizontal resolution			*/
	short		*vRes)					/* Vertical resolution				*/
{
	register TPPrint		recPtr;		/* Deference of print record handle	*/
	
	OpenPrintMgr( FALSE);				/* Open the printer driver			*/
	if ( !PrError() ) {
	
										/* Check validity of print record	*/
		PrValidate(macTPrint);
		FailOSErr( PrError());
	}
		
	ClosePrintMgr();							/* Close the printer driver			*/
	
	if (macTPrint != NULL) {			/* Extract info from print record	*/
		recPtr = *macTPrint;
		*paperRect = recPtr->rPaper;
		*pageRect = recPtr->prInfo.rPage;
		*hRes = recPtr->prInfo.iHRes;
		*vRes = recPtr->prInfo.iVRes;
	
	} else {							/* There is no print record			*/
		SetRect(paperRect, 0, 0, 0, 0);
		*pageRect = *paperRect;
		*hRes = *vRes = 0;
	}
}


/******************************************************************************
 DoPageSetup

		Respond to the Page Setup command by displaying the standard
		Toolbox printer style dialog, which allows the user to set
		the page dimensions and other goodies.
		
		Boolean result indicates whether the user pressed OK to confirm
		possible changes in the print record. This allows documents which
		associate a printer record with a file to update itself if desired.
 ******************************************************************************/

Boolean		CPrinter::DoPageSetup()
{
	Boolean		changed = FALSE;		/* Has print record changed?		*/

	if (OpenPrintMgr( TRUE)) {				/* Open printer driver				*/
										/* Toolbox trap posts the dialog	*/
		changed = PrStlDialog(macTPrint);
		FailOSErr( PrError());
	}
		
	ClosePrintMgr();							/* Close the printer driver			*/
	
	if (changed)
		ResetPagination();
	
	return(changed);
}


/******************************************************************************
 DoPrint

		Respond to the Print command by displaying the standard printer
		job dialog. Then print the associated Document if the user
		confirms the print request.
 ******************************************************************************/

void	CPrinter::DoPrint()
{
	Boolean		wantsToPrint = FALSE;
	
	if (OpenPrintMgr( TRUE)) {				/* Open printer driver				*/
		wantsToPrint = PrJobDialog(macTPrint);
		FailOSErr( PrError());
	}
	
	ClosePrintMgr();							/* Close the printer driver			*/
	
	if (wantsToPrint) {
		SetCursor(*gWatchCursor);
		PrintPageRange((**macTPrint).prJob.iFstPage,
					   (**macTPrint).prJob.iLstPage);
	}
}


/******************************************************************************
 PrintPageRange

		Print the specified range of pages. Normally, this message is
		sent by the DoPrint method after the user clicks "OK" for the Job
		Dialog. Programs may send this message on their own, but bypassing
		the standard dialogs is highly discouraged (see TechNote 122).
 ******************************************************************************/

void	CPrinter::PrintPageRange(		/* TCL 1.1.2 DLP 10/28/91 			*/
	short		firstPage,
	short		lastPage)
{
	TPrStatus			printStatus;	/* Status of spooled print job		*/
	register short		page;			/* Page number being printed		*/
	short				lastError;		/* Print error before PrClose		*/
	
	TRY
	{
		itsDocument->AboutToPrint(&firstPage, &lastPage);
		
		if (OpenPrintMgr( TRUE))
		{
			if (PrValidate( macTPrint))
				ResetPagination();
		}
		ClosePrintMgr();
		
		if (!HavePagination())		/* TCL 1.1.1 DLP 9/18/91 */
		{
			itsDocument->Paginate();
			if (!HavePagination())
				return;
		}
		
	
		if (OpenPrintMgr( TRUE)) {				/* Open printer driver				*/
			(**macTPrint).prJob.iFstPage = 1;
			(**macTPrint).prJob.iLstPage = lastPage - firstPage + 1;
			
			printDocOpen = TRUE;
			macPrintPort = PrOpenDoc(macTPrint, NULL, NULL);
			
			for (page = firstPage; page <= lastPage; page++) {
				if (PrError() == noErr) {
					printPageOpen = TRUE;
					PrOpenPage(macPrintPort, NULL);
					
					if (PrError() == noErr) {
						itsDocument->PrintPageOfDoc(page);
					}
					
					printPageOpen = FALSE;
					PrClosePage(macPrintPort);
				}
			}
			
			printDocOpen = FALSE;
			PrCloseDoc(macPrintPort);
			macPrintPort = NULL;
			
			if ( ((**macTPrint).prJob.bJDocLoop == bSpoolLoop)  &&
				 (PrError() == noErr) ) {
				 
				PrPicFile(macTPrint, NULL, NULL, NULL, &printStatus);
			}
			
			lastError = PrError();
			if (lastError != iPrAbort) {	/* Don't report error if the user	*/
											/*   aborted the printing			*/
				FailOSErr( lastError);
			}
		}
		
		ClosePrintMgr();					/* Close the printer driver			*/
	
		PrSetError(lastError);				/* Reset error code to what it was	*/
											/*   before closing printer in case	*/
											/*   Document needs it				*/
		itsDocument->DonePrinting();
	}
	CATCH
	{
		ClosePrintMgr();
		itsDocument->DonePrinting();
	}
	ENDTRY;
}

/******************************************************************************
 SetPrintDir
 
 	Set the direction of printing when there is more than one
 	horizontal and vertical strip.
 	
 ******************************************************************************/

void CPrinter::SetPrintDir( tPrintDirection aPrintDir)
{
	printDirection = aPrintDir;

}	/* CPrinter::SetPrintDir */

/******************************************************************************
 ResetPagination
 
 	Clear any current pagination, setting the number of horizontal and vertical
 	strips to zero.
 ******************************************************************************/

void CPrinter::ResetPagination( void)
{
	itsStripWidths->DeleteAll();
	itsStripHeights->DeleteAll();

}	/* CPrinter::ResetPagination */

/******************************************************************************
 HavePagination
 
 	Return true if there is any horizontal and vertical pagination info
 ******************************************************************************/

Boolean CPrinter::HavePagination( void)
{
	Boolean haveHPage, haveVPage;
	
	haveHPage = itsStripWidths && (itsStripWidths->GetNumItems() > 0);
	haveVPage = itsStripHeights && (itsStripHeights->GetNumItems() > 0);
	
	return haveHPage && haveVPage;

}	/* CPrinter::HavePagination */

/******************************************************************************
 SetStrips
 
 	
 ******************************************************************************/

void CPrinter::SetStrips( short numHStrips, short numVStrips)
{
	THPrint	printRec;
	Rect	pageRect;
	
	ResetPagination();
	
	printRec = GetPrintRecord();
	
	ASSERT( printRec != NULL);
	
	if (printRec)
	{
		pageRect = (**printRec).prInfo.rPage;
		itsStripWidths->InsertValue( 1, pageRect.right - pageRect.left, numHStrips);
		itsStripHeights->InsertValue( 1, pageRect.bottom - pageRect.top, numVStrips);
	}
	
}	/* CPrinter::SetStrips */

/******************************************************************************
 SetHorizPageBreak
 
 	Set a page break at a horizontal position in a client's Frame coordinates
 ******************************************************************************/

void CPrinter::SetHorizPageBreak( short vStripNum, long hPos)
{
	long prevBreak;
	
	ASSERT( (vStripNum >= 1) && (vStripNum <= itsStripWidths->GetNumItems() + 1));
	
	prevBreak = itsStripWidths->SumRange( 1, vStripNum - 1);
		
	SetStripWidth( vStripNum, hPos - prevBreak);
	
	
}	/* CPrinter::SetHorizPageBreak */

/******************************************************************************
 SetVertPageBreak
 
 	Set a page break at a vertical position in a client's Frame coordinates
 ******************************************************************************/

void CPrinter::SetVertPageBreak( short hStripNum, long vPos)
{
	long	prevBreak;
	
	ASSERT( (hStripNum >= 1) && (hStripNum <= itsStripHeights->GetNumItems() + 1));

	prevBreak = itsStripHeights->SumRange( 1, hStripNum - 1);
		
	SetStripHeight( hStripNum, vPos - prevBreak);

}	/* CPrinter::SetVertPageBreak */

/******************************************************************************
 SetAllStripWidths
 
 	Sets the width all vertical page strips.
 ******************************************************************************/

void CPrinter::SetAllStripWidths( short aStripWidth)
{
	short numStrips = itsStripWidths->GetNumItems();
	
	itsStripWidths->DeleteAll();
	itsStripWidths->InsertValue( 1, aStripWidth, numStrips);

}	/* CPrinter::SetAllStripWidths */

/******************************************************************************
 SetStripWidth
 
 	Sets the width of one vertical page strip
 ******************************************************************************/
 
void CPrinter::SetStripWidth( short pageNum, short aStripWidth)
{
	ASSERT( aStripWidth > 0);	/* must have some pixels on a page	*/
	
	if (pageNum <= itsStripWidths->GetNumItems())
		itsStripWidths->SetValue( pageNum, aStripWidth);
	else
		itsStripWidths->InsertValue( pageNum, aStripWidth, 1);
		

}	/* CPrinter::SetStripWidth */

/******************************************************************************
 SetAllStripHeights
 
 	Sets the height of all horizontal page strips
 ******************************************************************************/
void CPrinter::SetAllStripHeights( short aStripHeight)
{
	short numStrips = itsStripHeights->GetNumItems();
	
	itsStripHeights->DeleteAll();
	itsStripHeights->InsertValue( 1, aStripHeight, numStrips);

}	/* CPrinter::SetAllStripHeights */

/******************************************************************************
 SetStripHeight
 
 	Sets the height of one horizontal page strip.
 ******************************************************************************/

void CPrinter::SetStripHeight( short pageNum, short aStripHeight)
{
	ASSERT( aStripHeight > 0);	/* must have some pixels on a page	*/

	if (pageNum <= itsStripHeights->GetNumItems())
		itsStripHeights->SetValue( pageNum, aStripHeight);
	else
		itsStripHeights->InsertValue( pageNum, aStripHeight, 1);
		

}	/* CPrinter::SetStripHeight */

/******************************************************************************
 GetStripCount
 
 	Return the number of horizontal and vertical page strips
 ******************************************************************************/

void CPrinter::GetStripCount( short *hStrips, short *vStrips)
{
	*hStrips = itsStripHeights? itsStripHeights->GetNumItems() : 0;
	*vStrips = itsStripWidths? itsStripWidths->GetNumItems() : 0;

}	/* CPrinter::GetStripCount */

/******************************************************************************
 PageNumToStrips
 
 	Given a page number, return the corresponding strip numbers
 ******************************************************************************/

void CPrinter::PageNumToStrips( short pageNum, short *hStrip, short *vStrip)
{
	short	hStrips, vStrips;
	
	ASSERT( HavePagination());

	hStrips = itsStripHeights->GetNumItems();
	vStrips = itsStripWidths->GetNumItems();
	ASSERT( pageNum <= hStrips * vStrips);

	if (printDirection == printHoriz)
	{
		*hStrip = (pageNum - 1) / vStrips + 1;
		*vStrip = (pageNum - 1) % vStrips + 1;
	}
	else
	{
		*vStrip = (pageNum - 1)/ hStrips + 1;
		*hStrip = (pageNum - 1) % hStrips + 1;
	}	

}	/* CPrinter::PageNumToStrips */

/******************************************************************************
 GetPageStart
 
 	Get the starting position of a page in the Frame coordinates of a client.
 ******************************************************************************/

void CPrinter::GetPageStart( short pageNum, LongPt *startPos)
{
	short	theHStrip, theVStrip;
	
	PageNumToStrips( pageNum, &theHStrip, &theVStrip);
	
	startPos->h = itsStripWidths->SumRange( 1, theVStrip-1);
	startPos->v = itsStripHeights->SumRange( 1, theHStrip-1);
	

}	/* CPrinter::GetPageStart */

/******************************************************************************
 GetPageStart
 
 	Get the starting position of a page in the Frame coordinates of a client.
 ******************************************************************************/

void CPrinter::GetPageArea( short pageNum, LongRect *pageArea)
{
	short	theHStrip, theVStrip;
	
	PageNumToStrips( pageNum, &theHStrip, &theVStrip);
	
	GetPageStart( pageNum, &topLeftL( *pageArea));
	pageArea->right = pageArea->left + itsStripWidths->GetValue( theVStrip);
	pageArea->bottom = pageArea->top + itsStripHeights->GetValue( theHStrip);

}	/* CPrinter::GetPageArea */
