/******************************************************************************
 CRunArray.c

		CRunArray is a dynamic array optimized to conserve space where
		consecutive entries have the same value. Such a sequence of
		identical entries is called a run. Rather than storing N
		consecutive, identical entries, CRunArray stores the value once, 
		along with the length of the run. CRunArray only stores long integer
		quantities.
		
		The hItems handle is used to store a dynamic array of runs. Runs
		are specified using the tRun structure. The CArray::numItems
		instance variable therefore gives the number of runs, not the
		number of items. The number of items is indicated by CRunArray::itemCount.
		GetNumItems is overridden to return itemCount.
		
		The CArray methods Get, Set, InsertAtIndex, DeleteItem, etc. all
		manipulate the runs array, not individual elements. Use the
		CRunArray methods InsertValue, SetValue,GetValue, DeleteValue, etc.
		to manipulate entries.
		
	SUPERCLASS = CArray
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CRunArray.h"
#include "Global.h"

	/* macro for range checking	*/

#define ASSERT_INDEX( index)	ASSERT( (index > 0)&&(index <= itemCount))

/******************************************************************************
 IRunArray
 
 	Initialize a CRunArray. The initial number of items and runs is zero.
******************************************************************************/

void CRunArray::IRunArray( void)
{
	CArray::IArray( sizeof(tRun));
	itemCount = 0;
	hRuns = (tRunHndl) hItems;	

}	/* CRunArray::IRunArray */

/******************************************************************************
 GetNumItems {OVERRIDE}
 
 	Return the number of items in the array.
 	
******************************************************************************/

long CRunArray::GetNumItems( void)
{
	return itemCount;
	
}	/* CRunArray::GetNumItems */

/******************************************************************************
 InsertValue
 
 	Insert a run of values into the array.
 		item: one-based index indicating where to insert the run. If item 
 			  is >= itemCount + 1, then the value is inserted at the
 			  end of the array.
 		value: the value for the new item
 		count: length of the run.
 		
******************************************************************************/

void CRunArray::InsertValue( long item, long value, long count)
{
	long 	runIndex, firstInRun, newCount;
	
	ASSERT( item > 0);
	ASSERT( count > 0);
	
	/* first insert the value once	*/
	
	FindRun( item, &runIndex, &firstInRun);
	
	if (runIndex == BAD_INDEX)
	{
		/* Index is out of range. first we clip it to numItems + 1	*/
		/* then, if this is not the first run added, we check if 	*/
		/* has the same value as the last run if not we have to add	*/
		/* a new run												*/
		
		item = itemCount + 1;
		if ((itemCount > 0) && (value == GetValue( itemCount)))
		{
			/* insert to last run, just bump runLength */
			(*hRuns)[numItems-1].runLength++;
		}
		else
			InsertRun( numItems+1, 1, value);
		itemCount++;
	}
	else
	{	
		/* Rather than duplicate SetValue here, we just insert it to	*/
		/* the existing run for the given index, i.e. just bump			*/
		/* the runLength and then call SetValue, which will handle		*/
		/* the case where a new run might have to be created			*/
		
			/* clip item index to size + 1 */
		item = Min( itemCount+1, item);

		(*hRuns)[runIndex].runLength++;
		itemCount++;
		SetValue( item, value);
	}
	
	/* now insert the rest, if any, once we know the runs won't change	*/
	if (count > 1)
	{
		--count;
		FindRun( item, &runIndex, &firstInRun);
		ASSERT( runIndex != BAD_INDEX);	/* it must exist	*/
		
		(*hRuns)[runIndex].runLength += count;
		itemCount += count;
	}
	
	newCount = itemCount;
	BroadcastChange( runArraySizeChanged, &newCount);
		

}	/* CRunArray::InsertValue */

/******************************************************************************
 SetValue
 
 	Set the value of the item at index to the new value. This
 	operation may split or consolidate existing runs.
 	
******************************************************************************/

void CRunArray::SetValue( long index, long value)
{
	tRunPtr		runs = *hRuns;
	long 		runIndex, firstInRun, lastInRun;
	Boolean		fFirstRun, fLastRun;
	
	ASSERT_INDEX( index);
	
	FindRun( index, &runIndex, &firstInRun);
	
	/* NOTE: runIndex is zero-based, firstInRun, lastInRun, and index	*/
	/* are all one-based.												*/

	if (runs[runIndex].value != value)
	{		
		lastInRun = firstInRun + runs[runIndex].runLength - 1;
		fFirstRun = runIndex == 0;
		fLastRun = runIndex == (numItems-1);

		/* see if the item can go in previous or next run */
		
		if (!fFirstRun && (index == firstInRun) && (runs[runIndex-1].value == value))
		{
			runs[runIndex-1].runLength++;	/* bump the count */
			if ( runs[runIndex].runLength == 1) 
			{
				/* we no longer need this run */
				DeleteRun( runIndex + 1);
			}
			else runs[runIndex].runLength--; // TCL 1.1.2 DLP 12/9/91
		}
		else if (!fLastRun && (index == lastInRun) && (runs[runIndex+1].value == value))
		{
			runs[runIndex+1].runLength++;
			
			if ( runs[runIndex].runLength == 1) 
			{
				/* we no longer need this run */
				DeleteRun( runIndex + 1);
			}
			else runs[runIndex].runLength--; // TCL 1.1.2 DLP 12/9/91
		}
		else if (runs[runIndex].runLength == 1)
		{
			/* we already know that we couldn't merge the new value */
			/* into the previous or next run, so we just change		*/
			/* the value for this run								*/
			
			runs[runIndex].value = value;
		}

		else
		{	/* We have to break the run into two or more runs.		*/
			/* At this point we know the run size is greater than	*/
			/* one and if the item is at the start or end of the	*/
			/* run, we can't merge it with the previous or next run.*/
			/* Therefore the item itself starts a new run, and there*/
			/* may be a new run preceeding or following the item	*/
			
			long runSize;
			tRun newRun;
			
			/* check preceeding	*/
			
			if (index != firstInRun)
			{				
				InsertRun( runIndex+1, index - firstInRun, runs[runIndex].value);
				runIndex++; /* everything just moved down */
			}
			
			if (index != lastInRun)
			{
				InsertRun( runIndex + 2, lastInRun - index, runs[runIndex].value);
			}
			
			/* The item now has the run all to itself. Update the value	*/
			/* and length.												*/
			
			runs[runIndex].value = value;
			runs[runIndex].runLength = 1;	
		}
	}

}	/* CRunArray::SetValue */

/******************************************************************************
 GetValue
 
 	Return the value of an item
 	
******************************************************************************/

long CRunArray::GetValue( long index)
{	
	long runIndex, firstInRun;
	
	ASSERT_INDEX( index);
	
	FindRun( index, &runIndex, &firstInRun);
	
	return (*hRuns)[ runIndex].value;

}	/* CRunArray::GetValue */

/******************************************************************************
 DeleteValue
 
 	Delete a item from the array.
******************************************************************************/

void CRunArray::DeleteValue( long index)
{
	long runIndex, firstInRun, newCount;
	
	ASSERT_INDEX( index);
	
	FindRun( index, &runIndex, &firstInRun);
	
	if ((*hRuns)[runIndex].runLength == 1)
	{
		DeleteRun( runIndex+1);
	}
	else
		(*hRuns)[runIndex].runLength--;
		
	--itemCount;
	
	newCount = itemCount;
	BroadcastChange( runArraySizeChanged, &newCount);
	
}	/* CRunArray::DeleteValue */

/******************************************************************************
 DeleteAll
 
 	Delete all elements and runs
******************************************************************************/

void CRunArray::DeleteAll( void)
{
	long newCount;
	
	Resize( 0);
	numItems = 0;
	itemCount = 0;
	
	newCount = 0;
	BroadcastChange( runArraySizeChanged, &newCount);

}	/* CRunArray::DeleteAll */

/******************************************************************************
 SumRange
 
 	Returns the sum of all items beginning at startIndex up to and
 	including endIndex.
******************************************************************************/

long CRunArray::SumRange( long startIndex, long endIndex)
{
	register long sum, runIndex, itemIndex, lastThisRun;
	long	firstInRun, firstRun;
	register tRunPtr currRun;
	
	/* clip start and end */
	startIndex = Max( 1, startIndex);
	endIndex = Min( itemCount, endIndex);
	
	if (endIndex < startIndex) return 0;

	FindRun( startIndex, &firstRun, &firstInRun);
	currRun = *hRuns + firstRun;
	if (firstInRun + currRun->runLength > endIndex)
	{
			/* start and end are in the same run	*/
		sum = (endIndex - startIndex + 1) * currRun->value;
	}
	else
	{
			/* add in the this run from startIndex down	*/
			/* and then scan thru runs until endIndex	*/
		lastThisRun = firstInRun + currRun->runLength - 1;
		sum = (lastThisRun - startIndex +1) * currRun->value;
		currRun++;
		for (;;currRun++)
		{
			lastThisRun += currRun->runLength;
			if (lastThisRun >= endIndex)
			{
				/* we've found endIndex		*/
				sum += currRun->value * (currRun->runLength - (lastThisRun - endIndex));
				break;
			}
			else
			{
				sum += currRun->value * currRun->runLength;
			}
			ASSERT( lastThisRun < itemCount);
		}
	}			
	
	return sum;

}	/* CRunArray::SumRange */

/******************************************************************************
 FindSum
 
 	Returns the index of the first item which, by summing items 1..itemIndex,
 	has a sum >= aSum. BAD_INDEX is returned if aSum is too large or there
 	are no items in the array.
*******************************************************************************/

long CRunArray::FindSum( long aSum)
{
	register tRunPtr	currRun = *hRuns;
	register long		currSum = 0;
	register long		currIndex = 1;
	register long		runsLeft = numItems;
	
	while (--runsLeft >= 0)
	{
		currSum += currRun->value * currRun->runLength;
		currIndex += currRun->runLength;
		if (currSum >= aSum)
		{
			currIndex -= (currSum - aSum -1) / currRun->value + 1;
			return currIndex;
		}
		currRun++;
	}
	/* falling out of the loop means aSum was too large or zero runs */
	
	return BAD_INDEX;
	
}	/* CRunArray::FindSum */

/******************************************************************************
 FindRun
 
 	Return the index of the run containing the item with index itemIndex.
 	firstInRun is updated with the index of the first item in the run.
 	BAD_INDEX is returned if the itemIndex is not in the array.
 	
 	itemIndex and firstInRun are one based
 	theRunIndex is zero based
******************************************************************************/

void CRunArray::FindRun( long itemIndex, long *theRunIndex, long *firstInRun)
{
	register long runIndex;
	register tRunPtr runs = *hRuns;
	register long currItemIndex;
	
	if ((itemIndex > 0) && (itemIndex <= itemCount))
	{
		--itemIndex;						/* switch to zero-based offsets */
		currItemIndex = runIndex = 0;
		while (runIndex < numItems)
		{
			currItemIndex += runs->runLength;
			if (currItemIndex > itemIndex)
				break;
			runIndex++;
			runs++;
		}
		*theRunIndex = runIndex;
		*firstInRun = currItemIndex - (*hRuns)[runIndex].runLength + 1;
	}
	else
	{
		*theRunIndex = BAD_INDEX;
		*firstInRun = BAD_INDEX;
	}
	

}	/* CRunArray::FindRun */

/******************************************************************************
 InsertRun
 
 	Insert a new run into the array.
******************************************************************************/

void CRunArray::InsertRun( long index, long runLength, long value)
{
	tRun	newRun;
	
	newRun.runLength = runLength;
	newRun.value = value;
	
	InsertAtIndex( &newRun, index);

}	/* CRunArray::InsertRun */

/******************************************************************************
 DeleteRun
 
 	Delete a run from the array. This can result in the merging of the
 	two adjacent runs if they have the same value.
******************************************************************************/

void CRunArray::DeleteRun( long runIndex)
{
	DeleteItem( runIndex);
	
	/* if the deleted run was between two runs with the same value	*/
	/* then we can merge them.										*/
	
	if ((runIndex > 1) && (runIndex <= numItems))
	{
		runIndex--;
		if ((*hRuns)[runIndex-1].value == (*hRuns)[runIndex].value)
		{
			(*hRuns)[runIndex-1].runLength += (*hRuns)[runIndex].runLength;
			DeleteItem( runIndex+1);
		}
	}

}	/* CRunArray::DeleteRun */

/******************************************************************************
 Copy {OVERRIDE}

	Duplicate a CRunArray. This method calls CArray::Copy and the updates
	the hRuns instance variable, an alias to CArray's hItems handle.
 ******************************************************************************/

CObject* CRunArray::Copy( void)
{
	CRunArray *theCopy;
	
	theCopy = (CRunArray*) inherited::Copy();
	theCopy->hRuns = (tRunHndl) theCopy->hItems;
	
	return theCopy;
}