/******************************************************************************
 CFWDesktop.c

							The FWDesktop Class
		
		A Desktop which support floating windows. Floating windows are
		always active and remain above all "regular" windows.
	
	SUPERCLASS = CDesktop
	
	Copyright  1989 Symantec Corporation. All rights reserved.

 ******************************************************************************/

#include "Global.h"
#include "Constants.h"
#include "CFWDesktop.h"
#include "CDirector.h"
#include "TBUtilities.h"
#include "CList.h"
#include "CWindow.h"

/*** Class Constants ***/

#define		NO_DRAG			0x80008000	/* Flag signalling an aborted drag	*/

/******************************************************************************
 IFWDesktop

		Initialize a FWDesktop object
 ******************************************************************************/

void	CFWDesktop::IFWDesktop(
	CBureaucrat		*aSupervisor)		/* Supervisor in chain of command	*/
{
	CDesktop::IDesktop(aSupervisor);	/* Initialize superclass			*/

	itsFloats = new(CList);				/* Initialize our list of floating	*/
	itsFloats->IList();					/*   windows						*/
	topFloat = NULL;
}
	

/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a FWDesktop
 ******************************************************************************/

void	CFWDesktop::Dispose()
{
	itsFloats->DisposeAll();			/* Get rid of floating	windows		*/
	itsFloats = NULL;
		
	inherited::Dispose();				/* Pass message on to superclass	*/
}


/******************************************************************************
 AdjustCursor {OVERRIDE}

		Mouse is located where no other object wants to set the cursor shape.
		Adjust the shape to the standard arrow cursor
 ******************************************************************************/

		/*
		 * Subtract content region of floating window from mouseRgn
		 */
		 
		static void	Window_DiffContRgn(
			CWindow		*theWindow,
			RgnHandle	mouseRgn)
		{
			if (theWindow->IsVisible()) {
												/* TCL 1.1.1 DLP 9/18/91	*/
				UnionRgn(mouseRgn,((WindowPeek)theWindow->GetMacPort())->strucRgn,
					 mouseRgn);
				DiffRgn(mouseRgn,
						((WindowPeek)theWindow->GetMacPort())->contRgn,
						mouseRgn);
			}
		}

void	CFWDesktop::AdjustCursor(
	Point		where,					/* Mouse location in global coords	*/
	RgnHandle	mouseRgn)				/* Region containing the mouse		*/
{
	inherited::AdjustCursor(where, mouseRgn);
	
	if (topFloat != NULL) {
		itsFloats->DoForEach1( (EachFunc1) Window_DiffContRgn, (long) mouseRgn);
	}
}

/******************************************************************************
 AddWind {OVERRIDE}

		Add a window to the desktop
 ******************************************************************************/

void	CFWDesktop::AddWind(
	CWindow		*theWindow)				/* Window object to add				*/
{
	if (theWindow->IsFloating()) {
		itsFloats->Prepend(theWindow);
		
	} else {
		itsWindows->Add(theWindow);
	}
}
	

/******************************************************************************
 RemoveWind {OVERRIDE}

		Remove a window from the desktop
 ******************************************************************************/

void	CFWDesktop::RemoveWind(
	CWindow		*theWindow)				/* Window object to remove			*/
{
	if (theWindow->IsVisible()) {
		HideWind(theWindow);			/* Hide window before removing it	*/
	}
	
	if (theWindow->IsFloating()) {
		itsFloats->Remove(theWindow);
	} else {
		itsWindows->Remove(theWindow);
	}
}
	

/******************************************************************************
 SelectWind {OVERRIDE}

		Select a window by bringing it to the front and making it active
 ******************************************************************************/

void	CFWDesktop::SelectWind(
	CWindow		*theWindow)				/* Window to select					*/
{
	Boolean		aDAIsActive;			/* TRUE if a DA is the front window	*/
	CWindow		*aWindow;				/* Window object					*/
	CWindow		*oldTop;				/* Former top regular window		*/
	WindowPtr	macWPtr;				/* Ptr to Toolbox window record		*/
	WindowPtr	theDA;					/* Ptr to window record for a DA	*/
	
										/* Check if a DA is running			*/
	aDAIsActive = IsSystemWindow((WindowPeek)FrontWindow());
	
										/* Do nothing if the window is at	*/
										/*   the top of its layer and a DA	*/
										/*   is not active					*/
	if ( ((theWindow == topWindow) || (theWindow == topFloat))
			&& !aDAIsActive )
		return;
		
	if (aDAIsActive) {					/* A DA is currently running. Need	*/
										/*   to activate our application.	*/
										
										/* Find our bottom-most window		*/
		aWindow = (CWindow*) itsWindows->LastItem();
		macWPtr = (aWindow == NULL) ? NULL : aWindow->macPort;
		
										/* Put all DAs behind our bottom	*/
										/*   window, maintaining the		*/
										/*   ordering of the DAs			*/
		while (IsSystemWindow((WindowPeek) (theDA = FrontWindow()))) {
			SendBehind(theDA, macWPtr);
			macWPtr = theDA;
		}
	}
	
	if (theWindow->IsModal())
	{
		inherited::SelectWind( theWindow);
		return;
	}
	else if (theWindow->IsFloating()) {		/* Select a floating window			*/
		if (theWindow != topFloat) {	/* If this isn't the top float ...	*/
										/*   Make it the front window		*/
			BringToFront(theWindow->macPort);
										/*   Update our window list			*/
			itsFloats->BringFront(theWindow);
			theWindow->Show();			/*   Make it visible, if necessary	*/
			topFloat = theWindow;		/*   Mark it as the top float		*/
		}
	
	} else if (theWindow != topWindow) {
										/* If this isn't the top regular...	*/
		oldTop = topWindow;				/*   Save the current top regular	*/
		
		aWindow = (CWindow*) itsFloats->LastItem();
		if (aWindow == NULL) {			/*   Put it in front if there are	*/
										/*     no floating windows			*/
			BringToFront(theWindow->macPort);
		} else {						/*   Otherwise, put it behind the	*/
										/*     bottom floating window		*/
			BringBehind(theWindow->macPort, aWindow->macPort);
		}
										/*   Update our window list			*/
		itsWindows->BringFront(theWindow);
		
		theWindow->ShowOrHide(TRUE);	/*   Make it visible, if necessary	*/
				
		topWindow = theWindow;			/*   Mark it as the top regular		*/
		
		if (aDAIsActive) {
			if (oldTop != NULL) {
				HiliteWindow(oldTop->macPort, FALSE);
				((CDirector*)oldTop->itsSupervisor)->active = FALSE;
			}
			((CDirector*)topWindow->itsSupervisor)->active = TRUE;
			
		} else {
				
			if (oldTop != NULL) {		/*   Deactivate the former top		*/
										/*     regular window				*/
				oldTop->Deactivate();
			}
										
			topWindow->Activate();		/*   Activate the new top window	*/
		}
	}
	
	CurActivate = NULL;					/* Nuke any pending activate event	*/
}
	

/******************************************************************************
 ShowWind {OVERRIDE}

		Make a window visible on the desktop
 ******************************************************************************/

void	CFWDesktop::ShowWind(
	CWindow		*theWindow)				/* Window to show					*/
{
	if (theWindow->IsFloating()) {
		ShowHide(theWindow->macPort, TRUE);
		theWindow->Activate();
		topFloat = (CWindow*) itsFloats->FirstSuccess( (TestFunc) Window_IsVisible);
		
	} else {
		inherited::ShowWind(theWindow);
	}
}
	

/******************************************************************************
 HideWind {OVERRIDE}

		Make a window invisible on the desktop
 ******************************************************************************/

void	CFWDesktop::HideWind(
	CWindow		*theWindow)				/* Window to hide					*/
{
	if (theWindow->IsFloating()) {
		ShowHide(theWindow->macPort, FALSE);
		if (theWindow == topFloat) {
			topFloat = (CWindow*) itsFloats->FirstSuccess( (TestFunc) Window_IsVisible);
		}
	} else {
		inherited::HideWind(theWindow);
	}
}
	

/******************************************************************************
 DragWind {OVERRIDE}

		Drag a window on the desktop
 ******************************************************************************/

void	CFWDesktop::DragWind(
	CWindow		*theWindow,				/* Window to drag					*/
	EventRecord	*macEvent)				/* Associated mouse down event		*/
{
	GrafPtr				savePort;		/* Current drawing port				*/
	register WindowPtr	macWindow;		/* Toolbox window pointer			*/
	Point				corner;			/* Top left corner of window		*/
	Rect				dragBox;		/* Where window may be dragged		*/
	extern RgnHandle	gUtilRgn;		/* The region to drag				*/
	register long		newLoc;			/* New location after drag			*/
	short				hDrag;			/* Distance dragged horizontally	*/
	short				vDrag;			/* Distance dragged vertically		*/
	
	if ((theWindow != topWindow)  &&  (theWindow != topFloat)  &&
			!(macEvent->modifiers & cmdKey)) {
										/* Drag for an underlying window	*/
										/*   without the command key being	*/
										/*   pressed. Select window before	*/
										/*   performing the drag.			*/
		SelectWind(theWindow);
	}
	
	if (!StillDown()) {					/* Take a quick exit if the mouse	*/
		return;							/*   has already been released		*/
	}
	
	GetPort(&savePort);					/* Save the current port			*/
	
										/* Find Global coords of the top	*/
										/*   left corner of the window		*/
	macWindow = (WindowPtr) theWindow->GetMacPort();
	SetPort(macWindow);

	 corner = topLeft(macWindow->portRect);
	LocalToGlobal(&corner);
	
	
	SetPort(macPort);					/* Use desktop's port				*/
		
	PenNormal();						/* Use standard pen characteristics	*/
	
	dragBox = bounds;					/* Don't allow drag too close to	*/
										/*   the edge of the screen			*/
	InsetRect(&dragBox, DRAG_MARGIN, DRAG_MARGIN);
	
										/* Drag dotted outline of window	*/
	CopyRgn(((WindowPeek)macWindow)->strucRgn, gUtilRgn);
	newLoc = DragGrayRgn(gUtilRgn, macEvent->where, &dragBox, &dragBox,
							0, NULL);
	hDrag = LoShort(newLoc);
	vDrag = HiShort(newLoc);
	
										/* If a drag actually occurred, put	*/
										/*   window at new location			*/
	if ( (newLoc != NO_DRAG) && ( (hDrag != 0) || (vDrag != 0) ) ) {
		MoveWindow(macWindow, hDrag + corner.h,
							  vDrag + corner.v, FALSE);
	}
		
	SetPort(savePort);					/* Restore the original port		*/
	ForceNextPrepare();
}
	

/******************************************************************************
 Show {OVERRIDE}

		Make a Desktop visible by showing all its windows
 ******************************************************************************/

void	CFWDesktop::Show()
{
	if (!visible) {						/* Do nothing if already visible	*/
		visible = TRUE;
		itsFloats->DoForEach((EachFunc) Window_Show);
		itsWindows->DoForEach((EachFunc) Window_Show);
	}
}
	

/******************************************************************************
 Hide {OVERRIDE}

		Make a Desktop invisible by hiding all its windows
 ******************************************************************************/

void	CFWDesktop::Hide()
{
	if (visible) {						/* Do nothing if already invisible	*/
		visible = FALSE;
		itsWindows->DoForEach((EachFunc) Window_Hide);
		itsFloats->DoForEach((EachFunc) Window_Hide);
	}
}
	

/******************************************************************************
 Activate {OVERRIDE}

		Activate a desktop by activating all floating windows and the
		top regular window
 ******************************************************************************/

		/*
		 *	Activate a window
		 */
		static void	Window_Activate(
			CWindow		*theWindow)
		{
			theWindow->Activate();
		}

void	CFWDesktop::Activate()
{
	if (!active) {						/* Do nothing if already active		*/
		active = TRUE;
		itsFloats->DoForEach((EachFunc) Window_Activate);
		if (topWindow != NULL)
			topWindow->Activate();
	}
}
	

/******************************************************************************
 Dectivate {OVERRIDE}

		Deactivate a desktop by deactivating all floating windows and the
		top regular window
 ******************************************************************************/

		/*
		 *	Deactivate a window
		 */
		static void	Window_Deactivate(
			CWindow		*theWindow)
		{
			theWindow->Deactivate();
		}

void	CFWDesktop::Deactivate()
{
	if (active) {						/* Do nothing if already inactive	*/
		active = FALSE;
		itsFloats->DoForEach((EachFunc) Window_Deactivate);
		if (topWindow != NULL)
			topWindow->Deactivate();
	}
}



/******************************************************************************
 CalcTopFloat

        Set topFloat to the first visible window in itsFloats (if any).
 ******************************************************************************/

void    CFWDesktop::CalcTopFloat()
{
    topFloat = (CWindow *) itsFloats->FirstSuccess((TestFunc) Window_IsVisible);
}
