/******************************************************************************
 OSChecks.c
 
 						Useful Operating System Utilities
 						
 	Functions for checking features of the Mac OS which depend the machine
 	type and the System/Finder configuration.
 							
 	Exports:
 		Boolean		TrapAvailable(trapNum);
 		Boolean		WNEIsImplemented();
 		Boolean		TempMemCallsAvailable();
 		Boolean		ColorQDIsPresent();
 		
 	Copyright  1989 Symantec Corporation. All rights reserved.
 	
 	TCL 1.1 CHANGES
	[
		- changed TrapAvailable to use new approved method, removed
		  the trap type parameter because the function can determine
		  the type by itself.
	]	
 		
 ******************************************************************************/

#include "OSChecks.h"
#include "Traps.h"

/******************************************************************************
 TrapAvailable
 
 	Check whether a certain trap exists on this machine. This function uses
 	the new method as per IM VI, p. 3-8.
 	
 ******************************************************************************/

Boolean TrapAvailable( short theTrap)
{
	TrapType tType;
	short    numToolBoxTraps;
	
				// first determine the trap type
				
	tType = (theTrap & 0x800) > 0 ? ToolTrap : OSTrap;
	
				// next find out how many traps there are
				
	if (NGetTrapAddress( _InitGraf, ToolTrap) == NGetTrapAddress( 0xAA6E, ToolTrap))
		numToolBoxTraps = 0x200;
	else
		numToolBoxTraps = 0x400;
	
				// check if the trap number is too big for the
				// current trap table
				
	if (tType == ToolTrap)
	{
		theTrap &= 0x7FF;
		if (theTrap >= numToolBoxTraps)
			theTrap = _Unimplemented;
	}
	
				// the trap is implemented if its address is
				// different from the unimplemented trap
				
	return (NGetTrapAddress( theTrap, tType) != 
			NGetTrapAddress(_Unimplemented, ToolTrap));
}

/******************************************************************************
 WNEIsImplemented
 
 	Check whether the WaitNextEvent is implemented or not
 ******************************************************************************/

Boolean		WNEIsImplemented()
{
	SysEnvRec	theWorld;				/* System environment				*/
	
	SysEnvirons(1, &theWorld);			/* Check environment				*/
	
	if (theWorld.machineType < 0) 		/* Old ROMs, definitely not present	*/
		return(FALSE);
		
	else								/* Check for WNE trap				*/
		return(TrapAvailable(_WaitNextEvent));
}


/******************************************************************************
 TempMemCallsAvailable
 
 	Check whether the MultiFinder temporary memory calls are available
 ******************************************************************************/

Boolean		TempMemCallsAvailable()
{
										/* Check for the OSDispatch trap	*/                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              
	return(TrapAvailable(_OSDispatch));
}


/******************************************************************************
 ColorQDIsPresent
 
 	Check whether Color QuickDraw is present
 ******************************************************************************/

Boolean		ColorQDIsPresent()
{
	SysEnvRec	theWorld;				/* System environment				*/
	
	SysEnvirons(1, &theWorld);			/* Check environment				*/

	return(theWorld.hasColorQD);		/* Return environment field			*/
}

/******************************************************************************
 FlushCache		TCL 1.1.1 DLP 9/27/91
 
 	Flush the CPU cache(s). This is required on the 68040 after modifying
 	code. 
 ******************************************************************************/

#define _CacheFlushTrap		0xA0BD

void FlushCache(void)
{
	if (TrapAvailable( _CacheFlushTrap))
		asm
		{
			dc.w _CacheFlushTrap
		}
}
