/******************************************************************************
 CArrayPane.c

		A Pane that displays the contents of an array object in a scrolling list. 
		
		This is an abstract class that keeps the display synchronized with the 
		array contents. It must be subclassed to draw the array elements. 
		To accomplish this, override either the DrawCell or GetCellText methods.
		Both methods are defined in CTable.
		
	SUPERCLASS = CTable
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

 ******************************************************************************/

#include "CArrayPane.h"
#include "CArray.h"

/******************************************************************************
 IArrayPane
 
 	Initialize an ArrayPane. The itsArray instance variable is initialized
 	to NULL. Use the SetArray method to install an array.
******************************************************************************/

void CArrayPane::IArrayPane( CView *anEnclosure, CBureaucrat *aSupervisor,
						short aWidth, short aHeight,
						short aHEncl, short aVEncl,
						SizingOption aHSizing, SizingOption aVSizing)
{
	itsArray = NULL;
	ownsArray = FALSE;

	CTable::ITable( anEnclosure, aSupervisor, aWidth, aHeight,
				aHEncl, aVEncl, aHSizing, aVSizing);
				
	// by default, set up the table to have a single column, and to 
	// draw cells in column major order.
	
	SetDrawOrder( tblCol);
	
	AddCol( 1, -1);

}	/* CArrayPane::IArrayPane */

/******************************************************************************
 IViewTemp
 
 	Initialize an ArrayPane from a resource template.
******************************************************************************/

void CArrayPane::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
							Ptr viewData)
{
	itsArray = NULL;
	ownsArray = FALSE;
	
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	
	SetDrawOrder( tblCol);
}

/******************************************************************************
 Dispose
 
 	Dispose of the ArrayPane. Only dispose of itsArray if it owns the array.
******************************************************************************/

void CArrayPane::Dispose( void)
{
	if (ownsArray)
		ForgetObject( itsArray);
	inherited::Dispose();
	
}	/* CArrayPane::Dispose */

/******************************************************************************
 SetArray
 
 	Install a new array. The previous array, if any, is disposed of if
 	the ArrayPane owns it. The fOwnership flags specifies whether the
 	ArrayPane owns the new array.
******************************************************************************/

void CArrayPane::SetArray( CArray *anArray, Boolean fOwnership)
{
	short	deltaRows;
	
	ASSERT( anArray != NULL);
	
	if (itsArray)			/* TCL 1.1.1 DLP 10/28/91 */
	{
		if (ownsArray)
			itsArray->Dispose();
		else
			CancelDependency( itsArray);
	}
	itsArray = anArray;
	ownsArray = fOwnership;
	
	DependUpon( itsArray);
	
		// make sure table has same number of rows as the new array
		// has elements.
	
	deltaRows = itsArray->GetNumItems() - tableBounds.bottom;
	
	if (deltaRows > 0)
		AddRow( deltaRows,  tableBounds.bottom-1);
	else if (deltaRows < 0)
	{
		deltaRows = -deltaRows;
		DeleteRow( deltaRows, tableBounds.bottom - deltaRows);
	}
	Refresh(); /* TCL 1.1.2 DLP 12/13/91 */
		
}	/* CArrayPane::SetArray */

/******************************************************************************
 GetArray
 
 	Return reference to the itsArray.
******************************************************************************/

CArray *CArrayPane::GetArray( void)
{
	return itsArray;

}	/* CArrayPane::GetArray */

/******************************************************************************
 ProviderChanged
 
 	Respond to notification that itsArray has changed.
******************************************************************************/

void CArrayPane::ProviderChanged( CCollaborator *aProvider, long reason, void* info)
{
	tMovedElementInfo	*moveInfo;
	Rect	cellRect;
	long	infoIndex = *(long*) info;
	
	if (aProvider == itsArray)
	{
		switch( reason)
		{
			case arrayInsertElement:	/* new element inserted, add a new row */
				AddRow( 1, infoIndex - 2);
				break;
				
			case arrayDeleteElement:	/* element was deleted, delete its row	*/
				DeleteRow( 1, infoIndex-1);
				break;
				
			case arrayMoveElement:		/* an element was moved, redraw all 	*/
										/* elements encompassing its old and	*/
										/* new positions						*/
				moveInfo = (tMovedElementInfo *) info;
				
				cellRect.left = 0;
				cellRect.right = tableBounds.right;
				cellRect.top = Min( moveInfo->originalIndex, moveInfo->newIndex) -1;
				cellRect.bottom = Max( moveInfo->originalIndex, moveInfo->newIndex);
				RefreshCellRect( &cellRect);
				break;
				
			case arrayElementChanged:	/* an element was changed, redraw its	*/
										/* row									*/
					
				cellRect.left = 0;
				cellRect.right = tableBounds.right;
				cellRect.top = infoIndex-1;
				cellRect.bottom = infoIndex;
				RefreshCellRect( &cellRect);
				break;	
			
			default:
				inherited::ProviderChanged( aProvider, reason, info);
				break;			
		}
	}
	else
		inherited::ProviderChanged( aProvider, reason, info);

}	/* CArrayPane::ProviderChanged */

/******************************************************************************
 ChangeSize
 
 	If there is only one column, respond to a size change by setting its
	width to the width of the frame.
	
******************************************************************************/

void CArrayPane::ChangeSize( Rect *delta, Boolean redraw)
{
	
	inherited::ChangeSize( delta, redraw);
	
	if (tableBounds.right == 1)
		SetColWidth( 0, width);

}	/* CArrayPane::ChangeSize */
