/******************************************************************************
 CTable.c
		
	A class for displaying information in a row and column format.
		
	Required Non-core Classes:
	
		CTableDragger
		CTextEnvirons
	
	SUPERCLASS = CPanorama
	
	Copyright  1991 Symantec Corporation. All rights reserved.
	

*******************************************************************************/

#include "CTable.h"
#include "CTableDragger.h"
#include "CTextEnvirons.h"
#include "CScrollPane.h"
#include "CPaneBorder.h"
#include "Constants.h"
#include "Commands.h"
#include "CBartender.h"
#include "TBUtilities.h"
#include "CRunArray.h"
#include "CPrinter.h"
#include <Packages.h>

		/* macro evauluates to TRUE if a region is rectangular and non-empty */
		
#define IsRectRgn(rgn)	\
	((rgn) && ((**rgn).rgnSize == sizeof(Region)) && !EmptyRgn( rgn))
	
#define kDefaultHIndent		3	/* default horizontal cell indent	*/
#define kDefaultColWidth	50	/* default column width				*/

#define kBorderMargin		1	// margin of active border
#define kBorderPen			2	// pen size of active border


extern short		gClicks;
extern RgnHandle	gUtilRgn;
extern CBartender	*gBartender;
extern CBureaucrat	*gGopher;

RgnHandle CTable::cDeselection;
RgnHandle CTable::cNewSelection;
RgnHandle CTable::cCurrHilite;

/******************************************************************************
 ITable
 	
 	Initialize a CTable pane.
******************************************************************************/

void CTable::ITable(CView *anEnclosure, CBureaucrat *aSupervisor,
				short aWidth, short aHeight,
				short aHEncl, short aVEncl,
				SizingOption aHSizing, SizingOption aVSizing)
{

	/* call superclass's initializer */
	
	CPanorama::IPanorama( anEnclosure, aSupervisor, aWidth, aHeight,
			aHEncl, aVEncl, aHSizing, aVSizing);
			
	SetWantsClicks( TRUE);		/* we want to be clicked!		*/

	ITableX();	
	
}	/* CTable::ITable */

/******************************************************************************
 IViewTemp
 	
 	Initialize a CTable pane from a resource template.
******************************************************************************/

void CTable::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
							Ptr viewData)
{
	tTableTempP template = (tTableTempP) viewData;
	PatHandle	patH = NULL;
	Pattern		pat;
	
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	ITableX();
	
	if (template->rowHeight > 0)		/* TCL 1.1.1 DLP 9/19/91 */
		defRowHeight = template->rowHeight;
	if (template->colWidth > 0)
		defColWidth = template->colWidth;
	
	if (template->numRows > 0)
		AddRow( template->numRows, -1);
		
	if (template->numCols > 0)
		AddCol( template->numCols, -1);
		
	SetSelectionFlags( template->selectionFlags);
	SetDblClickCmd( template->dblClickCmd);	/* TCL 1.1.1 DLP 12/13/91 */
	SetDrawActiveBorder( template->drawActiveBorder != 0);
	
	if (template->rowBorderThickness > 0)
	{
		if (template->rowBorderPatID > 0)
		{
			patH = GetPattern( template->rowBorderPatID);
			FailNILRes( patH);
			BlockMove( *patH, pat, sizeof(Pattern));
			ForgetResource( patH);
		}
		else BlockMove( black, pat, sizeof(Pattern));
		
		SetRowBorders( template->rowBorderThickness, template->rowBorderPenMode,
						pat);
	}
	
	if (template->colBorderThickness > 0)
	{
		if (template->colBorderPatID > 0)
		{
			patH = GetPattern( template->colBorderPatID);
			FailNILRes( patH);
			BlockMove( *patH, pat, sizeof(Pattern));
			ForgetResource( patH);
		}
		else BlockMove( black, pat, sizeof(Pattern));
		
		SetColBorders( template->colBorderThickness, template->colBorderPenMode,
						pat);
	}	
}

/******************************************************************************
 ITableX
 
 	Perform common initialization for the table.
******************************************************************************/

void CTable::ITableX( void)
{
	short			rowHeight;
	FontInfo		fInfo;
	LongRect		interior;
	
	/* make sure class variables are initialized */
	
	if (!cDeselection)
	{
		cDeselection = NewRgn();
		cNewSelection = NewRgn();
		cCurrHilite = NewRgn();
	}

	usingLongCoord = TRUE;
	
	SetRect( &tableBounds, 0, 0, 0, 0);	/* start with no cells */	
	itsSelection = NewRgn();		/* selection is empty */
	
	itsRows = new( CRunArray);
	itsRows->IRunArray();
	itsCols = new( CRunArray);
	itsCols->IRunArray();
	
		
	drawOrder = tblRow;			/* default to row major drawing */
	clipToCells = FALSE;		/* don't clip cell drawing to cell rects */
	selectionFlags = 0;			/* no special flags				*/
	SetCanBeGopher( TRUE);
	
	drawActiveBorder = FALSE;
	dblClickCmd = cmdNull;
						
	CreateTextEnvironment();	/* setup the desired text characteristics */
	Prepare();					/* setup the port and restore the text 	  */
								/* environment */
	
	
	GetFontInfo( &fInfo);
	fontInfo = fInfo;			/* set the instance variable	*/
	rowHeight = fInfo.ascent + fInfo.descent + fInfo.leading;
	
	/* add a pixel extra on the top and bottom of rows, to look nice */
	rowHeight += 2;
	
	SetPt( &indent, kDefaultHIndent, fInfo.ascent+1);
	
	SetDefaults( width? width : kDefaultColWidth, rowHeight);
	
	SetRowBorders( 0, srcCopy, black);
	SetColBorders( 0, srcCopy, black);
	
	GetInterior( &interior);
	LongToQDPt( &topLeftL(interior), &topLeftIndent);

}	/* CTable::ITableX */

/******************************************************************************
 CreateTextEnvironment

	Create and initialize the text environment used for drawing text.
	it will also be used for calculating the default indent and row height.
******************************************************************************/

void CTable::CreateTextEnvironment( void)
{
	CTextEnvirons *textEnvirons;
	TextInfoRec	textInfo;

	textEnvirons = new( CTextEnvirons);
	itsEnvironment = textEnvirons;
	textEnvirons->ITextEnvirons();
	
	textInfo.fontNumber = applFont;
	textInfo.theSize = 12;
	textInfo.theStyle = 0;
	textInfo.theMode = srcOr;
	
	textEnvirons->SetTextInfo( &textInfo);

}	/* CTable::CreateTextEnvironment */

/******************************************************************************
 SetDefaults

	Set default row and column size. Negative value signals not to change
	a default. These values are used to set the scroll pane steps and overlap,
	and as the row and column heights used to insert rows or cols into an
	empty table.

******************************************************************************/

void CTable::SetDefaults( short colWidth, short rowHeight)
{
	if (rowHeight > 0) defRowHeight = rowHeight;
	if (colWidth > 0) defColWidth = colWidth;
	
	/* if we are embedded in a scroll pane, then use the default row and column
		sizes as the scroll step and overlap */
	if (itsScrollPane)
	{
		itsScrollPane->SetSteps( defColWidth, defRowHeight);
		itsScrollPane->SetOverlaps( defColWidth, defRowHeight);
	}

}	/* CTable::SetDefaults */

/******************************************************************************
 SetScrollPane {OVERRIDE}

	Override SetScrollPane to ensure that we have the scroll
	steps and scroll overlap setup properly.

******************************************************************************/

void CTable::SetScrollPane( CScrollPane *aScrollPane)
{
	inherited::SetScrollPane( aScrollPane);
		
	itsScrollPane->SetSteps( defColWidth, defRowHeight);
	itsScrollPane->SetOverlaps( defColWidth, defRowHeight);

}	/* CTable::SetScrollPane */

/******************************************************************************
 SetDrawOrder

	Drawing can occur in row or column major order, whichever
	is more convenient for the subclass. When drawOrder == tblRow,
	DrawRow is called to draw rows. When drawOrder == tblCol, DrawCol
	is called to draw each column. Database programs might
	use row major order, and override DrawRow to set up the
	current record for each row. A table that displays vectors of numbers
	as columns might override DrawCol to setup the current vector for each
	column.

******************************************************************************/

void CTable::SetDrawOrder( tTblDrawOrder aDrawOrder)
{
	drawOrder = aDrawOrder;

}	/* CTable::SetDrawOrder */

/******************************************************************************
 MakeMouseTask
 
	A separate method is used to make a mouse task in order that
	subclasses may override MakeMouseTask to make a custom type of task,
	for example to make dragging move cells around instead of selecting.
	
******************************************************************************/

CMouseTask *CTable::MakeMouseTask( short modifiers)
{
	CTableDragger	*tblTask;
	
	tblTask = new( CTableDragger);
	tblTask->ITableDragger( this, modifiers, selectionFlags);
	return tblTask;
	
}	/* CTable::MakeMouseTask */

/******************************************************************************
 DoClick {OVERRIDE}
 
	Override of DoClick. Three types of clicks are recognized. A click
	in the content (in any cell) is classified into a single or
	double click. A click within the view frame but outside any cells
	is treated differently.

******************************************************************************/

void CTable::DoClick( Point hitPt, short modifierKeys, long when)
{
	CMouseTask	*mouseTask;
	LongRect	pinRect;
	Cell	hitCell;
	LongPt	pt;
		
	QDToFrame( hitPt, &pt);
	
	if (PtInLongRect( &pt, &bounds))
	{
		if (gClicks == 2)	/* a double click? */
		{
			FindHitCell( &pt, &hitCell);
			DoDblClick( hitCell, modifierKeys, when);
		}
		else	/* a single click, make mouse task and pass to TrackMouse */
		{
			mouseTask = MakeMouseTask( modifierKeys);
			if (mouseTask)
			{
				pinRect = bounds;
				
				InsetLongRect( &pinRect, 1, 1);
				TrackMouse( mouseTask, &pt, &pinRect);
				ForgetObject( mouseTask);	/* TCL 1.1.1 DLP 9/18/91 */
			}
		}
	}
	else ClickOutsideBounds( hitPt, modifierKeys, when);

}	/* CTable::DoClick */

/******************************************************************************
 SetDblClickCmd
 
 	Specify the command sent when a cell is double-clicked.
******************************************************************************/

void CTable::SetDblClickCmd( long aCmd)
{
	dblClickCmd = aCmd;

}	/* CTable::SetDblClickCmd */

/******************************************************************************
 DoDblClick
 
 	Respond to a cell double-click. This method sends itself the current
 	dblClickCmd, if any. The command is sent to self so that CTable subclasses
 	may do something with the command. If they don't, the command goes to
 	the supervisor anyway.
 	
******************************************************************************/

void CTable::DoDblClick( Cell hitCell, short modifierKeys, long when)
{
	if (dblClickCmd != cmdNull)
		DoCommand( dblClickCmd);
		
}	/* CTable::DoDblClick */
/******************************************************************************
 HitSamePart {OVERRIDE}
 
	HitSamePart is called by the TCL to determine two clicks can be considered
	to be in "the same part". This is used to recognize double-clicks.
	For a table, we consider a click to be in the same part if both are
	in the same cell.
 
******************************************************************************/

Boolean CTable::HitSamePart(Point pointA, Point pointB)
{
	Cell	cellA, cellB;
	LongPt	ptA, ptB;
	
	WindToFrame( pointA, &ptA);	/* points are in window coords, convert to local coords */
	WindToFrame( pointB, &ptB);
	
							/* find the cell hit by each point						*/
	FindHitCell( &ptA, &cellA);
	FindHitCell( &ptB, &cellB);
	
	/* before we test the cells for equality, we must make sure that that			*/
	/* both points were in the table in the first place.							*/
	
	if (!PtInRect( cellA, &tableBounds) || !PtInRect( cellB, &tableBounds))
		return FALSE;
	else return EqualPt( cellA, cellB);

}	/* CTable::HitSamePart */

/******************************************************************************
 AddRow
 
 	AddRow inserts one or more rows into the table. The rows are inserted
	following afterRow or before row 0 if afterRow is negative.

******************************************************************************/

void CTable::AddRow( short numRows, short afterRow)
{
	short	rowHeight;
	Rect	cellRect;
	
	if (numRows > 0)
	{
		/* clip afterRows to actual row count, remember row numbers zero-based */
		afterRow = Min( afterRow, tableBounds.bottom-1);
		afterRow = Max( afterRow, -1);
		
		/* get the height for the new row. Normally use the height of	*/
		/* afterRow, if it exists. If we are inserting before row 0		*/
		/* use that, otherwise use the default value					*/
		
		if (afterRow >= 0)
			rowHeight = GetRowHeight( afterRow);
		else
			rowHeight = GetRowHeight( 0);
		if (rowHeight < 0)
			rowHeight = defRowHeight;
		

		afterRow += 2;
		itsRows->InsertValue( afterRow, rowHeight, numRows);
		tableBounds.bottom += numRows;
		
		cellRect.top = afterRow-1;
		cellRect.left = 0;
		cellRect.right = tableBounds.right;
		cellRect.bottom = tableBounds.bottom;
		
		AdjustBounds();	
		RefreshCellRect( &cellRect);		
	}

}	/* CTable::AddRow */

/******************************************************************************
 AddCol

	Inserts one or more columns into the table. The columns are inserted
	following afterCol or before columns 0 if afterCol is negative.
******************************************************************************/

void CTable::AddCol( short numCols, short afterCol)
{
	short	colWidth;
	Rect	cellRect;
	
	if (numCols > 0)
	{
		/* clip afterCols to actual colCount, remember col numbers zero-based */
		afterCol = Min( afterCol, tableBounds.right-1);
		afterCol = Max( afterCol, -1);
		
		/* get the width for the new column. Normally use the width of	*/
		/* afterCol, if it exists. If we are inserting before column 0	*/
		/* use that, otherwise use the default value						*/
		
		if (afterCol >= 0)
			colWidth = GetColWidth( afterCol);
		else
			colWidth = GetColWidth( 0);
		if (colWidth < 0)
			colWidth = defColWidth;
		

		afterCol += 2;
		itsCols->InsertValue( afterCol, colWidth, numCols);

		tableBounds.right += numCols;
		cellRect.top = 0;
		cellRect.bottom = tableBounds.bottom;
		cellRect.left = afterCol-1;
		cellRect.right = tableBounds.right + numCols;

		AdjustBounds();	
		RefreshCellRect( &cellRect);		
	}

}	/* CTable::AddCol */

/******************************************************************************
 DeleteRow

	Delete one or more rows following startRow. Parameters
	are checked to ensure we only attempt to delete existing
	rows.

******************************************************************************/

void CTable::DeleteRow( short numRows, short startRow)
{
	Rect 	r;
	short	i;
	
	if (numRows > 0 && (tableBounds.bottom > 0))
	{
		/* clip startRow and numRows to actual #rows */
		
		startRow = Min( tableBounds.bottom -1, startRow);
		numRows = Min( numRows, tableBounds.bottom - startRow);
		
		/* ensure that cells being deleted are not selected */
		SetRect( &r, 0, startRow, tableBounds.right, startRow + numRows);
		DeselectRect( &r, TRUE);
		
		++startRow;			/* shift to one-based index */
		
		for (i = 0; i < numRows; i++)
			itsRows->DeleteValue( startRow);
		
		tableBounds.bottom -= numRows;
		if (tableBounds.bottom < 0) tableBounds.bottom = 0;
		AdjustBounds();
		Refresh();
	}

}	/* CTable::DeleteRow */

/******************************************************************************
 DeleteCol
 
 	Delete one or more columns following startCol. Parameters
	are checked to ensure we only attempt to delete existing
	cols.

******************************************************************************/

void CTable::DeleteCol( short numCols, short startCol)
{
	Rect 	r;
	short	i;

	if (numCols > 0 && (tableBounds.right > 0))
	{
		/* clip startCol and numCols to actual #cols */
		
		startCol = Min( tableBounds.right -1, startCol);
		numCols = Min( numCols, tableBounds.right - startCol);
		
		/* ensure that cells being deleted are not selected */
		SetRect( &r, startCol, 0, startCol + numCols, tableBounds.bottom);
		DeselectRect( &r, TRUE);
		
		++startCol;			/* shift to one-based index */
		
		for (i = 0; i < numCols; i++)
			itsCols->DeleteValue( startCol);
		
		tableBounds.right -= numCols;
		if (tableBounds.right < 0) tableBounds.right = 0;
		AdjustBounds();
		Refresh();
	}

}	/* CTable::DeleteCol */

/******************************************************************************
 SetRowHeight
	
	Set the height of the given row.
******************************************************************************/

void CTable::SetRowHeight( short rowNum, short newHeight)
{
	Rect	cellRect;
	
	cellRect = tableBounds;
	cellRect.top = rowNum;
	RefreshCellRect( &cellRect);

	itsRows->SetValue( rowNum+1, newHeight);
	AdjustBounds();

	RefreshCellRect( &cellRect);
	
}	/* CTable::SetRowHeight */

/******************************************************************************
 SetColWidth
 
	return size of table in cells.
******************************************************************************/

void CTable::SetColWidth( short colNum, short newWidth)
{
	Rect	cellRect;
	
	cellRect = tableBounds;
	cellRect.left = colNum;
	RefreshCellRect( &cellRect);

	itsCols->SetValue( colNum+1, newWidth);
	AdjustBounds();

	RefreshCellRect( &cellRect);

}	/* CTable::SetColWidth */

/******************************************************************************
 GetTableBounds
 
	return size of table in cells.
******************************************************************************/

void CTable::GetTableBounds( Rect *aTableBounds)
{
	*aTableBounds = tableBounds;

}	/* CTable::GetTableBounds */

/******************************************************************************
 GetRowCount
 
 	return number of rows
******************************************************************************/

short CTable::GetRowCount( void)
{
	return tableBounds.bottom;

}	/* CTable::GetRowCount */

/******************************************************************************
 GetColCount
 
 	return number of columns
******************************************************************************/

short CTable::GetColCount( void)
{
	return tableBounds.right;

}	/* CTable::GetColCount */

/******************************************************************************
 GetRowHeight
 
 	Return height of a row.
******************************************************************************/

short CTable::GetRowHeight( short rowNum)
{
	if ((rowNum >= 0) && (rowNum < tableBounds.bottom))
		return itsRows->GetValue( rowNum+1);
	else
		return -1;
	
}	/* CTable::GetRowHeight */

/******************************************************************************
 GetColWidth
 
 	return width of a column.
******************************************************************************/

short CTable::GetColWidth( short colNum)
{
	if ((colNum >= 0) && (colNum < tableBounds.right))
		return itsCols->GetValue( colNum+1);
	else
		return -1;
	
}	/* CTable::GetColWidth */
		
/******************************************************************************
 GetRowStart
 
 	Return the pixel position of the top of a row
 	Assumes rowNum is in range.
******************************************************************************/

long CTable::GetRowStart( short rowNum)
{
	long	rowStart;
	
	ASSERT(( rowNum >= 0) && (rowNum < tableBounds.bottom));
	
	/* itsRows contains the heights of all rows. Need to total the heights	*/
	/* to figure where a row starts. SumRange totals up to and including	*/
	/* the end index, so do not add 1 and therefore find out where the		*/
	/* previous row ends													*/
	
	if (rowNum > 0)
		rowStart = itsRows->SumRange( 1, rowNum);
	else
		rowStart = 0;
		
	return topLeftIndent.v + rowStart;

}	/* CTable::GetRowStart */

/******************************************************************************
 GetColStart
 
 	Return the pixel position of the left size of a column
 	Assumes colNum is in range.
******************************************************************************/

long CTable::GetColStart( short colNum)
{
	long	colStart;
	
	ASSERT((colNum >= 0) && (colNum < tableBounds.right));
		
	if (colNum > 0)
		colStart = itsCols->SumRange( 1, colNum);
	else
		colStart = 0;
		
	return topLeftIndent.h + colStart;

}	/* CTable::GetColStart */

/******************************************************************************
 AdjustBounds
 
	Update the bounds instance variable (which denotes the extent of the view
	in pixels) to match the tableBounds (gives extent in cells).

******************************************************************************/

void CTable::AdjustBounds( void)
{
	Rect	cellR;
	LongRect pixelR;
	
	cellR = tableBounds;
	CellsToPixels( &cellR, &pixelR);
	
	/* top left of pixelR may be greater than 0,0 because of topLeftIndent */
	pixelR.left = pixelR.top = 0;
	SetBounds( &pixelR);

}	/* CTable::AdjustBounds */


/******************************************************************************
 GetCellText
 
 	Get the text for a cell. Called by the default version of DrawCell
******************************************************************************/

void CTable::GetCellText( Cell aCell, short availableWidth, StringPtr itsText)
{
	unsigned char numBuf[20];
		
	/* make a default string which is row,col */
	
	NumToString( aCell.v, itsText);
	itsText[ ++itsText[0]] = ',';
	NumToString( aCell.h, numBuf);
	ConcatPStrings( itsText, numBuf);

}	/* CTable::AdjustBounds */

/******************************************************************************
 Draw {OVERRIDE}

	Draw is called by the TCL to draw our view. Subclasses only need to
	override DrawCell in order to do cell drawing.
******************************************************************************/

void CTable::Draw( Rect *area)
{
	Cell		topLeftCell, botRightCell, currCell;
	short		i;
	Rect		selectedCells, theCells;
	LongRect 	r;
	Boolean		doCellHilite, reallyActive;
			
	QDToFrameR( area, &r);
	
	if (!SectLongRect( &r, &bounds, &r))
		return;		
	
		/* account for topLeftIndent 			*/
		
	r.left = Max( r.left, topLeftIndent.h);
	r.top = Max( r.top, topLeftIndent.v);	

	r.right--;
	r.bottom--;
		
		/* find bounding cells of redraw rect 	*/
		
	FindHitCell( &topLeftL(r), &topLeftCell);
	FindHitCell( &botRightL(r), &botRightCell);
	
	topLeft(theCells) = topLeftCell;
	botRight( theCells) = botRightCell;
	theCells.bottom++; theCells.right++;
	DrawBorders( &theCells);

	/*  if the selection is rectangular, we can optimize by
		hiliting cells in one fell swoop. Otherwise we
		resort to hiliting one cell at a time. */
		
	reallyActive = IsActive();
	doCellHilite = reallyActive && !IsRectRgn( itsSelection);
	
	FrameToQDR( &aperture, &qdClip);
	
	if (drawOrder == tblRow)
	{
		for ( i = topLeftCell.v; i <= botRightCell.v; i++)
			DrawRow( i, topLeftCell.h, botRightCell.h, doCellHilite);
	}
	else if (drawOrder == tblCol)
	{
		for ( i = topLeftCell.h; i <= botRightCell.h; i++)
			DrawCol( i, topLeftCell.v, botRightCell.v, doCellHilite);	
	}

	if (clipToCells)
	{
		ForceNextPrepare();
		Prepare();
	}
	
	if (reallyActive && !doCellHilite)
	{
		selectedCells = (**itsSelection).rgnBBox;
		if (SectRect( &theCells, &selectedCells, &selectedCells))
			Hilite( &selectedCells, TRUE);
	}
	
		
}	/* CTable::Draw */

/******************************************************************************
 DrawRow

	Called to draw a single row of cells, only when drawOrder == tblRow.
	Subclasses could override to do something at the start of each row,
	such as setup a database record.
******************************************************************************/

void CTable::DrawRow( short rowNum, short firstCol, short lastCol,
				register Boolean doHilite)
{
	Cell		currCell;
	LongRect	cellRect;
	Rect		hiliteRect, qdRect, clipRect;
	
	currCell.v = rowNum;
	hiliteRect.top = rowNum;
	hiliteRect.bottom = rowNum + 1;
	
	for ( currCell.h = firstCol; currCell.h <= lastCol; currCell.h++)
	{
		GetCellRect( currCell, &cellRect);
		FrameToQDR( &cellRect, &qdRect);
		
		if (clipToCells)
		{
			if (!SectRect( &qdRect, &qdClip, &clipRect))
				continue;
				
			ClipRect( &clipRect);
		}
		
		DrawCell( currCell, &qdRect);
		if ((doHilite) && PtInRgn( currCell, itsSelection))
		{
			hiliteRect.left = currCell.h;
			hiliteRect.right = currCell.h + 1;
			Hilite( &hiliteRect, TRUE);
		}
	}	

}	/* CTable::DrawRow */

/******************************************************************************
 DrawCol
 
	Called to draw a single column of cells, onyl when drawOrder == tblCol.
	Subclasses could override to do something at the start of each column,
	such as setup a vector of data for the current column.
******************************************************************************/

void CTable::DrawCol( short colNum, short firstRow, short lastRow,
			register Boolean doHilite)
{
	Cell		currCell;
	LongRect	cellRect;
	Rect		hiliteRect, qdRect, clipRect;

	currCell.h = colNum;
	hiliteRect.left = colNum;
	hiliteRect.right = colNum + 1;

	for ( currCell.v = firstRow; currCell.v <= lastRow; currCell.v++)
	{
		GetCellRect( currCell, &cellRect);
		FrameToQDR( &cellRect, &qdRect);
		
		if (clipToCells)
		{
			if (!SectRect( &qdRect, &qdClip, &clipRect))
				continue;
				
			ClipRect( &clipRect);
		}
		
		DrawCell( currCell, &qdRect);
		if ((doHilite) && PtInRgn( currCell, itsSelection))
		{
			hiliteRect.top = currCell.v;
			hiliteRect.bottom = currCell.v + 1;
			Hilite( &hiliteRect, TRUE);
		}
	}

}	/* CTable::DrawCol */


/******************************************************************************
 DrawCell
 
 	Draw a single cell. The default is text call GetCellText and then
 	draw that text. If your table is textual you could just override
 	GetCellText. If the cells don't have text, or have text plus other
 	graphics, then override DrawCell.
 	cellRect is already translated from long to QD coordinates.
******************************************************************************/

void CTable::DrawCell( Cell theCell, Rect *cellRect)
{
	Str255	cellText;
	short	availWidth;
	
	availWidth = cellRect->right - cellRect->left - indent.h; /* TCL 1.1.1 DLP 9/19/91 */
	
	GetCellText( theCell, availWidth, cellText);
	
	if (cellText[0] > 0)
	{
		MoveTo( cellRect->left + indent.h, cellRect->top + indent.v);
		DrawString( cellText);
	}

}	/* CTable::DrawCell */

/******************************************************************************
 Hilite

	Hilite a rectangular region of cells. Assumes the pane is already
	Prepared().

******************************************************************************/

void CTable::Hilite( Rect *cells, Boolean hiliteFlag)
{
	LongRect	pixels;
	Rect		visPixels;
	
	if (!printing)
	{
		CellsToPixels( cells, &pixels);	/* get the pixels spanning the cells	*/
		if ( SectAperture( &pixels, &visPixels))
		{
			SetHiliteMode();		/* invert with user's hilite color		*/
			InvertRect( &visPixels);
		}
	}

}	/* CTable::Hilite */

/******************************************************************************
 DrawBorders

	Draw the requested row and column borders

******************************************************************************/

void CTable::DrawBorders( Rect *theCells)
{
	Rect	cells;
	LongRect pixels;
	Cell	currCell;
	short	delta;
	short	row, col;
	LongPt	lPt;
	Point	pt;
		
	CellsToPixels( theCells, &pixels);
	
	/* borders are drawn on the tops of rows and left of columns	*/
	/* but we don't want to border row or column zero				*/

	cells = *theCells;
	if (cells.top == 0) cells.top++;
	if (cells.left == 0) cells.left++;
	
	/* draw column borders 			*/
	
	if (colBorders.thickness > 0)
	{
		PenPat( colBorders.pat);
		PenMode( colBorders.penMode);
		PenSize( colBorders.thickness, colBorders.thickness);
		
		delta = pixels.bottom - pixels.top;
		for (col = cells.left; col < cells.right; col++)
		{			
			SetLongPt( &lPt, GetColStart( col), pixels.top);
			FrameToQD( &lPt, &pt);
			MoveTo( pt.h, pt.v);
			Line( 0, delta);
		}
	}
	
	/* draw row borders 			*/
	
	if (rowBorders.thickness > 0)
	{
		PenPat( rowBorders.pat);
		PenMode( rowBorders.penMode);
		PenSize( rowBorders.thickness, rowBorders.thickness);
		
		delta = pixels.right - pixels.left;
		for (row = cells.top; row < cells.bottom; row++)
		{
			SetLongPt( &lPt, pixels.left, GetRowStart( row));
			FrameToQD( &lPt, &pt);
			MoveTo( pt.h, pt.v);
			Line( delta, 0);
		}
	}
	PenNormal();


}	/* CTable::DrawBorders */

/******************************************************************************
 RefreshCell

	Invalidate a cell to force a redraw.

******************************************************************************/

void CTable::RefreshCell( Cell aCell)
{
	LongRect	r;
	
	if (PtInRect( aCell, &tableBounds))
	{
		GetCellRect( aCell, &r);
		RefreshLongRect( &r);
	}

}	/* CTable::RefreshCell */

/******************************************************************************
 RefreshCellRect
 
	Invalidate of rectangular region of cells
******************************************************************************/

void CTable::RefreshCellRect( Rect *cells)
{
	LongRect	cellCoords;
	Rect		visArea;
	
	if (CellsToPixels( cells, &cellCoords))
	{
		if (SectAperture( &cellCoords, &visArea))
			InvalRect( &visArea);
	}

}	/* CTable::RefreshCellRect */

/******************************************************************************
 GetCellRect

	Return the pixels spanning theCell.
******************************************************************************/

void CTable::GetCellRect( Cell theCell, register LongRect *cellRect)
{
	
	/* Check if theCell is in table bounds. Note that this
		also handles case of empty table */
		
	if ( PtInRect( theCell, &tableBounds))
	{		
		cellRect->left = GetColStart( theCell.h);
		cellRect->top = GetRowStart( theCell.v);
		
		cellRect->right = cellRect->left + GetColWidth( theCell.h);
		cellRect->bottom = cellRect->top + GetRowHeight( theCell.v);
	}
	else
		SetLongRect( cellRect, 0, 0, 0, 0);
	
}	/* CTable::GetCellRect */

/******************************************************************************
 FindRow

	Find the row corresponding to a vertical coordinate
******************************************************************************/

short	CTable::FindRow( long vLoc)
{
	short row;
	
	if ((vLoc >= 0) && (vLoc < bounds.bottom))
	{
		vLoc -= topLeftIndent.v;
		row = itsRows->FindSum( vLoc) - 1;
		if (row < 0) row = -1;
	}
	else
		row = -1;
		
	return row;
	
}	/* CTable::FindRow */

/******************************************************************************
 FindCol

	Find the column corresponding to a horizontal coordinate
******************************************************************************/

short	CTable::FindCol( long hLoc)
{
	short col;
	
	if ((hLoc >= 0) && (hLoc < bounds.right))
	{
		hLoc -= topLeftIndent.h;
		col = itsCols->FindSum( hLoc) - 1;
		if (col < 0) col = -1;
	}
	else
		col = -1;
		
	return col;

}	/* CTable::FindCol */

/******************************************************************************
 PixelsToCells

	Convert a normal rectangle expressed in points in the view's
	coordinate system to the cells that intersect it.
	Return true if pixelsRect actually contains any cells.
******************************************************************************/

Boolean CTable::PixelsToCells( LongRect *pixelsRect, Rect *cellsRect)
{
	LongRect	r;

	if (SectLongRect( pixelsRect, &bounds, &r))
	{
		FindHitCell( &topLeftL(r), &topLeft( *cellsRect));
		FindHitCell( &botRightL(r), &botRight( *cellsRect));
		return TRUE;
	}
	else 
	{
		SetRect( cellsRect, 0, 0, 0, 0);
		return FALSE;
	}

}	/* CTable::PixelsToCells */
/******************************************************************************
 CellsToPixels
 
	Gives the pixels spanned by a rectangular area of cells.
	returns true if cellsRect contains any pixels, i.e., cellsRect
	has a nonempty intersection with tableBounds.

******************************************************************************/

Boolean CTable::CellsToPixels( Rect *cellsRect, LongRect *pixelsRect)
{	
	Rect		realCells;
	LongRect	r;

	if ( SectRect( cellsRect, &tableBounds, &realCells))
	{
		GetCellRect( topLeft( realCells), &r);
		topLeftL(*pixelsRect) = topLeftL(r);
		
		realCells.right -= 1; realCells.bottom -=1;
		GetCellRect( botRight( realCells), &r);
		botRightL(*pixelsRect) = botRightL(r);
		
		return TRUE;
	}
	else
	{
		SetLongRect( pixelsRect, 0, 0, 0, 0);
		return FALSE;
	}

}	/* CTable::CellsToPixels */

/******************************************************************************
 FindHitCell

	Return the cell containing the hitPt. (-1,-1) is returned
	if hitPt is not in any cell.

******************************************************************************/

void CTable::FindHitCell( LongPt *hitPt, Cell *hitCell)
{
	hitCell->h = FindCol( hitPt->h);
	hitCell->v = FindRow( hitPt->v);

}	/* CTable::FindHitCell */

/******************************************************************************
 ClickOutsideBounds

	Handle a click within the table frame but outside table bounds.

******************************************************************************/

void CTable::ClickOutsideBounds( Point hitPt, short modifierKeys, long when)

{
	DeselectAll( TRUE);

}	/* CTable::ClickOutsideBounds */

/******************************************************************************
 SelectCell
 
	Select a single cell. Bottlenecks thru SelectRect.
******************************************************************************/

void CTable::SelectCell( Cell aCell, Boolean keepPrevious, Boolean reDraw)
{
	Rect	selectRect;
	
	SetRect( &selectRect, aCell.h, aCell.v, aCell.h +1, aCell.v +1);
	SelectRect( &selectRect, keepPrevious, reDraw);
	

}	/* CTable::SelectCell */
/******************************************************************************
 SelectRect

	Select a rectangular region of cells. If keepPrevious is true, the new selection
	is added to the previous selection, otherwise the new selection replaces the
	previous selection. If redraw is true, then the area is redrawn, otherwise
	it is invalidated.
******************************************************************************/

void CTable::SelectRect( Rect *selectRect, Boolean keepPrevious, Boolean redraw)
{
	Rect	realCells, r, visPixels;
	LongRect pixels;
	Boolean	fForcePrepare = FALSE;
	
	/* first clip selectRect to intersection with tableBounds */
	
	if (SectRect( selectRect, &tableBounds, &realCells))
	{
			
			/* cNewSelection holds the cells to be selected					*/
		RectRgn( cNewSelection, &realCells);
		
			/* Only clear the old selection if it has been requested		*/
			/* and is non-empty												*/
		if ((!keepPrevious) && !EmptyRgn( itsSelection))
		{										
				/* determine the cells that will actually be deselected		*/
				/* not simply deselected and reselected						*/
			DiffRgn( itsSelection, cNewSelection, cDeselection);
						
				/* deselect the deselection									*/
			DiffRgn( itsSelection, cDeselection, itsSelection);
		}
		else
			SetEmptyRgn( cDeselection);
			
			/* Determine the region of cells that will be newly selected	*/
		DiffRgn( cNewSelection, itsSelection, cNewSelection);
		
			/* Before updating itsSelection, copy it so we know what cells	*/
			/* are currently hilited (if redrawing otherwise we don't care)	*/
		CopyRgn( itsSelection, cCurrHilite);
			
			/* union the old selection with the new selection				*/
		UnionRgn( itsSelection, cNewSelection, itsSelection);
			
		if (redraw)
		{
			Rect	origBBox;
			
			Prepare();
			
			/* attempt to make cDeselection and cNewSelection rectangular */
			/* so we can hilite them efficiently						  */
			
			if (!EmptyRgn( cDeselection) && !IsRectRgn( cDeselection))
			{
					/* Try unioning cDeselection and itsSelection. To 		*/
					/* compensate, we would have to	then union itsSelection */
					/* with cNewSelection. If cNewSelection is already 		*/
					/* rectangular, we don't even attempt it, because we 	*/
					/* may be optimizing deselection while simultaneously 	*/
					/* screwing up selection								*/
					
				if (!IsRectRgn( cDeselection) && !IsRectRgn( cNewSelection))
				{
					origBBox = (**cDeselection).rgnBBox;
					
					UnionRgn( cDeselection, itsSelection, gUtilRgn);
					if (IsRectRgn( gUtilRgn))
					{
						CopyRgn( gUtilRgn, cDeselection);
						UnionRgn( cNewSelection, itsSelection, cNewSelection);
					}
				}						
			}
			if (!EmptyRgn( cNewSelection) && !IsRectRgn( cNewSelection))
			{
				if (IsRectRgn( cCurrHilite))
				{
					UnionRgn( cNewSelection, cCurrHilite, gUtilRgn);
					if (IsRectRgn( gUtilRgn))
					{
						CopyRgn( gUtilRgn, cNewSelection);
						/* cCurrHilite was rectangular. Now that we have unioned	*/
						/* it with cNewSelection, we must either unhilite			*/
						/* cCurrHilite before hiliting cNewSelection, or even		*/
						/* better, simply clip it out.								*/
						r = (**cCurrHilite).rgnBBox;
						CellsToPixels( &r, &pixels);
						if (SectAperture( &pixels, &visPixels))
						{
							RectRgn( gUtilRgn, &visPixels);
							DiffRgn( thePort->clipRgn, gUtilRgn, thePort->clipRgn);	
							fForcePrepare = TRUE;				
						}
					}
				}
			}
			HiliteCellRegion( cDeselection, FALSE);
			HiliteCellRegion( cNewSelection, TRUE);
			
			if (fForcePrepare) ForceNextPrepare();
		}
		else /* just invalidate */
		{
			r = (**cDeselection).rgnBBox;
			RefreshCellRect( &r);
			r = (**cNewSelection).rgnBBox;
			RefreshCellRect( &r);			
		}
		SetEmptyRgn( cNewSelection);
		SetEmptyRgn( cDeselection);
		SetEmptyRgn( cCurrHilite);
		SetEmptyRgn( gUtilRgn);
		
		BroadcastChange( tableSelectionChanged, NULL);

	}

}	/* CTable::SelectRect */

/******************************************************************************
 HiliteCellRegion

	Hilite a region of cells. Assumes pane is already Prepared().
	Also assumes itsSelection matches fHilite, i.e. if it must default
	to drawing each cell, the hilite will be drawn correctly.
	Returns TRUE if a Hilite() or nothing was done, FALSE if Draw() was done.
******************************************************************************/

void CTable::HiliteCellRegion( RgnHandle cellRgn, Boolean fHilite)
{
	Rect	cellRect;
	
	if (!EmptyRgn( cellRgn))
	{
		cellRect = (**cellRgn).rgnBBox;
		if (IsRectRgn( cellRgn))	
			Hilite( &cellRect, fHilite);
		else
		{
			Rect r;
			
			for (r.top = cellRect.top; r.top < cellRect.bottom; r.top++)
			{
				r.bottom = r.top + 1;
				for (r.left = cellRect.left; r.left < cellRect.right; r.left++)
				{
					r.right = r.left + 1;
					if (PtInRgn( topLeft( r), cellRgn))
						Hilite( &r, fHilite);
				}
			}					
		}
	}

}	/* CTable::HiliteCellRegion */

/******************************************************************************
 DeselectCell

	Deselect a cell. Bottlenecks thru DeselectRect.
******************************************************************************/

void CTable::DeselectCell( Cell aCell, Boolean redraw)
{
	Rect deselectRect;
	
	topLeft( deselectRect) = aCell;
	deselectRect.right = aCell.h + 1;
	deselectRect.bottom = aCell.v + 1;
	
	/* DeselectRect will validate the input, no reason to do it twice */
	
	DeselectRect( &deselectRect, redraw);

}	/* CTable::DeselectCell */

/******************************************************************************
 DeselectRect

	Deselect a rectangular region of cells.
******************************************************************************/

void CTable::DeselectRect( Rect *deselectRect, Boolean redraw)
{
	Rect	realCells, r, visPixels;
	LongRect pixels;
	
	/* first clip selectRect to intersection with tableBounds */
	
	if (SectRect( deselectRect, &tableBounds, &realCells))
	{
		/* only deselect cells that were actually selected */
		
		RectRgn( gUtilRgn, &realCells);
		SectRgn( itsSelection, gUtilRgn, gUtilRgn);
		if (!EmptyRgn( gUtilRgn))
		{
			DiffRgn( itsSelection, gUtilRgn, itsSelection);
			if (redraw)
			{
				r = (**gUtilRgn).rgnBBox;
				if (IsRectRgn( gUtilRgn))
				{
					/* we need to unhilite a rectangular area,
						don't necessarily need to redraw */
					Prepare();
					Hilite( &r, FALSE);
				}
				else
				{
					SetEmptyRgn( gUtilRgn);
					CellsToPixels( &r, &pixels);
					if (SectAperture( &pixels, &visPixels))
					{
						Prepare();
						EraseRect( &visPixels);
						Draw( &visPixels);
					}
				}
			}
		}
		
		BroadcastChange( tableSelectionChanged, NULL);
	}

}	/* CTable::DeselectRect */
/******************************************************************************
 DeselectAll

	Utility to deselect everything.

******************************************************************************/

void CTable::DeselectAll( Boolean redraw)
{
	Rect	r = (**itsSelection).rgnBBox;
	
	DeselectRect( &r, redraw);

}	/* CTable::DeselectAll */

/******************************************************************************
 IsSelected

	Test if a cell is selected. Since we use a QuickDraw region to maintain
	the selection, we can just use PtInRgn to test if a cell (point) is in
	the selection region.

******************************************************************************/

Boolean CTable::IsSelected( Cell aCell)
{
	return PtInRgn( aCell, itsSelection);

}	/* CTable::IsSelected */

/******************************************************************************
 GetSelection

	Return the Table's region indicating the selected cells.
******************************************************************************/

RgnHandle CTable::GetSelection( void)
{
	return itsSelection;

}	/* CTable::GetSelection */

/******************************************************************************
 NextCell

	Next cell is a utility to support iteration thru all the cells in a row,
	column, or entire table. Given a cell in theCell NextCell will return
	the "next" cell according to the supplied criteria, if one can be found.
	The function result is TRUE if a cell was found, FALSE if none was found.
	
	hNext indicates whether we should move to the right to find the next cell.
	vNext indicates whether we should move down to find the next cell.
	
	Preference is given to hNext, meaning that if theCell given is not in
	the last row or column, then the next cell in the same row is returned.

******************************************************************************/

Boolean CTable::NextCell( Boolean hNext, Boolean vNext,Cell *theCell)
{
	Boolean atLastRow, atLastCol;
	Cell newCell = *theCell;
	
	if (!PtInRect( *theCell, &tableBounds)) return FALSE;
	
	if (!hNext && !vNext) return FALSE;
	
	atLastRow = (theCell->v == tableBounds.bottom-1);
	atLastCol = (theCell->h == tableBounds.right-1);
			
	if (hNext)
	{
		if (atLastCol)
		{
			if (atLastRow) return FALSE;/* no more cells */
			if (!vNext) return FALSE;	/* don't want next row */
			else 
			{
				newCell.h = 0;	/* beginning of next row */
				newCell.v++;	/* we know its not the last row */
			}
		}
		else newCell.h++;				/* next cell in row */
	}	
	else if (vNext)	/* hNext was false */
	{
		if (atLastRow)
			return FALSE;	/* no more rows */
		else	
			newCell.v++;	/* next cell in column */
	}
	
	*theCell = newCell;
	return TRUE;

}	/* CTable::NextCell */

/******************************************************************************
 GetSelect

	GetSelect provides two different behaviors, depending on the value of
	next parameter.
	If next is false, GetSelect tests if theCell is selected.
	If next is true, GetSelect finds the first selected cell starting 
	with or following theCell. True is returned if such a cell is found.
	This provides a convenient way to iterate through all the selected cells.

******************************************************************************/

Boolean CTable::GetSelect( Boolean next, Cell *theCell)
{
	short	currRow, currCol, lastCol, lastRow;
	Cell	currCell;
	
	if (EmptyRgn( itsSelection)) return FALSE;
	if (!PtInRect( *theCell, &tableBounds)) return FALSE;
	if (!next) return IsSelected( *theCell);
	
	currCell = *theCell;
	do
	{
		if (PtInRgn( currCell, itsSelection))
		{
			*theCell = currCell;
			return TRUE;
		}
	
	} while ( NextCell( TRUE, TRUE, &currCell));
	
	return FALSE;

}	/* CTable::GetSelect */

/******************************************************************************
 ScrollToSelection {OVERRIDE}

	ScrollToSelection causes the current selection to be scrolled into view.

******************************************************************************/

void CTable::ScrollToSelection( void)
{
	short	hSpan, vSpan;
	Rect	selRect = (**itsSelection).rgnBBox;
	LongRect	topLeftRect, botRightRect;
	LongPt	selPos;
	
	if (EmptyRect( &selRect) || EmptyLongRect( &aperture)) return;	/* nothing selected or nothing visible*/
	
	selRect.right--; selRect.bottom--;	/* bottom right actual cell of rect */
			
	GetFrameSpan( &hSpan, &vSpan);
	selPos = position; /* init to current scroll position */
	
	/* get pixels of topLeft and botRight cells of the selection */
	
	GetCellRect( topLeft(selRect), &topLeftRect);
	GetCellRect( botRight(selRect), &botRightRect);
	
	/*
		Note that the we always ensure that the full top left or
		bottom right cell is in view. For example, if there is 1 selected
		cell, located in the top left of the frame such that only
		the rightmost edge of the cell is visible, we will scroll
		the cell completely into view.
	*/
	
	/**** calc vertical scroll ****/
	
		/* Is the top of selection below the bottom of frame? */
	
	if (topLeftRect.bottom >= position.v + vSpan)	
	{
		selPos.v = topLeftRect.bottom - vSpan; 
	}
		
		/* Is the bottom of selection above the top of the frame? */
	
	else if (botRightRect.top < position.v)
	{
		selPos.v = botRightRect.top;
	}
		
	/**** calc horizontal scroll ****/
	
		/* Is the left edge of selection past the right edge of frame? */
		
	if (topLeftRect.right >= position.h + hSpan)
	{
		selPos.h = topLeftRect.right - hSpan;
	}
		
		/* Is the right edge of selection before the left edge of frame? */
		 
	else if (botRightRect.left < position.h)
	{
		selPos.h = botRightRect.left;
	}

	if ((selPos.h != position.h) || (selPos.v != position.v))
		ScrollTo( &selPos, TRUE);

}	/* CTable::ScrollToSelection */

/******************************************************************************
 SetSelectionFlags
 
	Set selection flags instance variable.
******************************************************************************/

void CTable::SetSelectionFlags( long selFlags)
{
	selectionFlags = selFlags;

}	/* CTable::SetSelectionFlags */

/******************************************************************************
 GetSelectionFlags

	return the selection flags.
******************************************************************************/

long CTable::GetSelectionFlags( void)
{
	return selectionFlags;

}	/* CTable::GetSelectionFlags */

/******************************************************************************
 DoKeyDown {OVERRIDE}
 
	Handle navigation via arrow keys. Currently only works if zero or one
	cells are selected.

******************************************************************************/

void CTable::DoKeyDown(char theChar, Byte keyCode, EventRecord *macEvent)
{
	short	deltaH = 0, deltaV = 0;
	Cell	currCell;
	Rect	selectRect;
			
	switch (keyCode)
	{
		case KeyLeftCursor:
			deltaH = -1;
			break;
			
		case KeyRightCursor:
			deltaH = 1;
			break;
			
		case KeyUpCursor:
			deltaV = -1;
			break;
			
		case KeyDownCursor:
			deltaV = 1;
			break;
	
		default:
		
			inherited::DoKeyDown( theChar, keyCode, macEvent);
			return;
	}
	Prepare();
	
	if (EmptyRgn( itsSelection))	/* if no selection, put selection at 0,0 */
		SetCell( currCell, 0, 0);
	else if (IsRectRgn( itsSelection))	/* OK, its a rect, check for */
	{									/* single selected cell */
		selectRect = (**itsSelection).rgnBBox;
		if ((selectRect.right == selectRect.left + 1) &&
			(selectRect.bottom == selectRect.top + 1))
		{
			currCell = topLeft( selectRect);
			currCell.h += deltaH;
			currCell.v += deltaV;
		}
		else return;
	}
	else return;

	if (PtInRect( currCell, &tableBounds))	/* moving to a real cell? */
	{
		SelectCell( currCell, FALSE, TRUE);
		ScrollToSelection();
	}

}	/* CTable::DoKeyDown */

/******************************************************************************
 DoAutoKey {OVERRIDE}

	we handle repeated keys the same as regular keypresses.
******************************************************************************/

void CTable::DoAutoKey(char theChar, Byte keyCode, EventRecord *macEvent)

{
	DoKeyDown( theChar, keyCode, macEvent);

}	/* CTable::DoAutoKey */

/******************************************************************************
 DoCommand {OVERRIDE}

	Handle Table commands
******************************************************************************/

void CTable::DoCommand( long aCmd)
{
	Rect r;
	
	switch (aCmd)
	{
		case cmdSelectAll:
			r = tableBounds;
			SelectRect( &r, TRUE, TRUE);
			break;
			
		default:
			inherited::DoCommand( aCmd);
			break;
	}

}	/* CTable::DoCommand */

/******************************************************************************
 UpdateMenus {OVERRIDE}

	Enable Table menu commands
******************************************************************************/

void CTable::UpdateMenus( void)
{
	inherited::UpdateMenus();
	
	if ((selectionFlags & selOnlyOne) == 0)
		gBartender->EnableCmd( cmdSelectAll);

}	/* CTable::UpdateMenus */
/******************************************************************************
 SetRowBorders

	Set the style of row borders. A thickness of zero specifies no
	border.
******************************************************************************/

void CTable::SetRowBorders( short thickness, short penMode, Pattern penPat)
{
	rowBorders.thickness = thickness;
	rowBorders.penMode = penMode;
	BlockMove( penPat, rowBorders.pat, sizeof( Pattern));

}	/* CTable::SetRowBorders */

/******************************************************************************
 SetColBorders

	Set the style of column borders. A thickness of zero specifies no
	border.
******************************************************************************/

void CTable::SetColBorders( short thickness, short penMode, Pattern penPat)
{
	colBorders.thickness = thickness;
	colBorders.penMode = penMode;
	BlockMove( penPat, colBorders.pat, sizeof( Pattern));

}	/* CTable::SetColBorders */

/******************************************************************************
 SetDrawActiveBorder

	Specify whether the table has a border when it is active
******************************************************************************/

void CTable::SetDrawActiveBorder( Boolean fDrawActiveBorder)
{
	drawActiveBorder = fDrawActiveBorder;

}	/* CTable::SetDrawActiveBorder */

/******************************************************************************
 BecomeGopher {OVERRIDE}

	BecomeGopher is called when the table is either becoming the gopher
	or was the gopher but is not any more. If the instance variable 
	drawActiveBorder is TRUE and the table is in a scroll pane, then
	it places a border around the scroll pane, and removes the border
	when the table is not the gopher. This provides the same appearance
	as scrolling lists in System 7.
	
******************************************************************************/

Boolean CTable::BecomeGopher( Boolean fBecoming)
{
	if (!inherited::BecomeGopher( fBecoming))
		return FALSE;

	if (drawActiveBorder && itsScrollPane)
	{
		CPaneBorder *border;
		Rect		margin;
		
		if (fBecoming)
		{
			if (!itsScrollPane->GetBorder())
			{
				border = new( CPaneBorder);
				border->IPaneBorder( kBorderFrame);
				SetRect( &margin, -kBorderMargin, -kBorderMargin, 
							kBorderMargin, kBorderMargin);
				border->SetPenSize( kBorderPen, kBorderPen);
				border->SetMargin( &margin);
				
				itsScrollPane->SetBorder( border);
			}
		}
		else
			itsScrollPane->SetBorder( NULL);
	}
	return TRUE;

}	/* CTable::BecomeGopher */

/******************************************************************************
 Activate {OVERRIDE}

******************************************************************************/

void CTable::Activate(void)
{
	if (!active)
	{
		Prepare();
		HiliteCellRegion( itsSelection, TRUE);
	}

	inherited::Activate();	
	
}	/* CTable::Activate */

/******************************************************************************
 Deactivate {OVERRIDE}

******************************************************************************/

void CTable::Deactivate(void)
{
	if (active)
	{
		Prepare();
		HiliteCellRegion( itsSelection, FALSE);
	}

	inherited::Deactivate();
	
}	/* CTable::Deactivate */

/******************************************************************************
 Paginate {OVERRIDE}

	Paginates the table so that only full cells are printed.
******************************************************************************/

void CTable::Paginate( CPrinter *aPrinter, short pageWidth, short pageHeight)
{
	long	currOffset;
	short	stripNum;
	
	aPrinter->ResetPagination();
	
	if (EmptyRect( &tableBounds)) return;
	
	/****** calc vertical strips ******/
	
	stripNum = 1;
	currOffset = 0;
	do
	{
		/* find the column at right end of strip		*/
		
		long col, colStart, colWidth;
		
		currOffset += pageWidth;

		col = FindCol( currOffset);
		if (col < 0) col = tableBounds.right -1;
		
		/* if column fits completely, break the strip after	*/
		/* the column, otherwise break it before the colum	*/
		
		colStart = GetColStart( col);
		colWidth = GetColWidth( col);
		
		if (colStart + colWidth <= currOffset)
			currOffset = colStart + colWidth;
		else
			currOffset = colStart;
		
		aPrinter->SetHorizPageBreak( stripNum++, currOffset);
		
	}
	while (currOffset < bounds.right);
	
	/****** calc horizontal strips ******/
	
	stripNum = 1;
	currOffset = 0;
	do
	{
		/* find the row at bottom end of strip		*/
		
		long row, rowStart, rowHeight;
		
		currOffset += pageHeight;

		row = FindRow( currOffset);
		if (row < 0) row = tableBounds.bottom -1;
				
		rowStart = GetRowStart( row);
		rowHeight = GetRowHeight( row);
		
		if (rowStart + rowHeight <= currOffset)
			currOffset = rowStart + rowHeight;
		else
			currOffset = rowStart;
		
		aPrinter->SetVertPageBreak( stripNum++, currOffset);
	}
	while (currOffset < bounds.bottom);


}	/* CTable::Paginate */

/******************************************************************************
 Dispose {OVERRIDE}

	Dispose of intance variables owned by the table.
******************************************************************************/

void CTable::Dispose( void)
{
	if (itsSelection)
	{
		DisposeRgn( itsSelection);
		itsSelection = NULL;
	}
	ForgetObject( itsRows);
	ForgetObject( itsCols);
	inherited::Dispose();

}	/* CTable::Dispose */

