
/******************************************************************************
 CAbstractText.c

		CAbstractText is a new class for TCL 1.1. As its name indicates,
		it represents an abstract text-editing pane. The old CEditText
		class is now a subclass of CAbstractText. CAbstractText defines
		a basic protocol that text-editing panes should implement. Some
		of these methods must be overriden by subclasses. These are:
		
			DoClick, SetTextPtr, GetTextHandle, SetFontNumber, SetFontStyle, 
			SetFontSize, SetTextMode, SetAlignCmd, SetAlignCmd, SetSpacingCmd, 
			GetHeight, GetCharOffset, GetCharPoint, GetTextStyle, GetCharStyle, 
			FindLine, GetLength, TypeChar, SetSelection, GetSpacingCmd, GetSelection,
			GetNumLines, CopyTextRange, PerformEditCommand, InsertTextPtr, Dawdle
			
		The remaining CAbstractText methods are implemented on top of the above
		methods and are inherited for free by subclasses.
		
		The following methods call GetTextHandle to examine the text buffer,
		under the assumption that it is cheap call. If the text is not
		stored in a single, contiguous handle buffer then you should also
		override these methods:
		
			GetCharBefore, GetCharAfter
		
		In recognition of the fact that the primary subclass of CAbstractText,
		CEditText, is based on TextEdit, and that most programmers will not be
		subclassing CAbstractText directly, CAbstractText's protocol is 
		based on the TextEdit interface and does not define any behavior
		that cannot be provided by TextEdit.
		
	SUPERCLASS = CPanorama
	
	Copyright  1991 Symantec Corporation. All rights reserved.

	
 ******************************************************************************/

#include "CAbstractText.h"

#include "Global.h"
#include "CAbstractText.h"
#include "Commands.h"
#include "TBUtilities.h"
#include "CBartender.h"
#include "CScrollPane.h"
#include "CClipboard.h"
#include "Constants.h"
#include "CTextEditTask.h"
#include "CTextStyleTask.h"
#include "CPrinter.h"

#include <Packages.h>

/*** Global Variables ***/

extern CBartender		*gBartender;	/* Manages all menus				*/
extern CClipboard		*gClipboard;


short CAbstractText::cFirstTaskIndex = 0;


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IAbstractText

		Initialize an AbstractText object
 ******************************************************************************/

void	CAbstractText::IAbstractText(
	CView			*anEnclosure,
	CBureaucrat		*aSupervisor,
	short			aWidth,
	short			aHeight,
	short			aHEncl,
	short			aVEncl,
	SizingOption	aHSizing,
	SizingOption	aVSizing,
	short			aLineWidth)
{
	// initialize instance variables
	
	itsTypingTask = NULL;
	itsLastTask = NULL;
	lineWidth = aLineWidth;
	wholeLines = FALSE;
	editable = TRUE;	
	lastFontNum = MININT;
	lastFontCmd = cmdNull;
	fixedLineHeights = TRUE;
	stylable = TRUE;
	
	// initialize superclass
	
	CPanorama::IPanorama(anEnclosure, aSupervisor,
					aWidth, aHeight, aHEncl, aVEncl, aHSizing, aVSizing);
					
	// indicate that we accept mouse clicks and can become the gopher
	
	SetWantsClicks( TRUE);
	SetCanBeGopher( TRUE);
}

/******************************************************************************
 IViewTemp

		Initialize an AbstractText object from a resource template
 ******************************************************************************/

void CAbstractText::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
							Ptr viewData)
{
	tAbstractTextTempP	template = (tAbstractTextTempP) viewData;
	
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);

	lineWidth = template->lineWidth;			/* Set instance vars from template	*/
	wholeLines = template->wholeLines != 0;
	
	Specify( template->editable != 0, wantsClicks, template->styleable != 0);

	itsTypingTask = NULL;
	lastFontNum = MININT;
	lastFontCmd = cmdNull;
	fixedLineHeights = TRUE;
}


/**** D I S P L A Y   M E T H O D S ****/
	

/******************************************************************************
 ScrollToSelection {OVERRIDE}

		Scroll the text so that the current selection is visible within
		the frame
 ******************************************************************************/
		
void	CAbstractText::ScrollToSelection()
{
	long	startLine, endLine;
	short	hSpan;
	short	vSpan;
	LongPt	selPos;
	long	selStart, selEnd;
	
	GetSelection( &selStart, &selEnd);
	
	startLine = FindLine(selStart);
	
	if (selStart == selEnd) {
		endLine = startLine;
	} else {
		endLine = FindLine(selEnd);
	}
	
	GetFrameSpan( &hSpan, &vSpan);

		/* Here we assume that the vertical units are either pixels	*/
		/* or lines													*/
		
	if (vScale == 1)
	{
							/* don't try to scroll if 1 line or less	*/
							/* is visible								*/
		if (vSpan <= Get1Height( startLine))
			return;
			
			/* set selStart to the bottom of the first selected line 	*/
		selStart = GetHeight( 0, startLine+1);
		
			/* set selEnd to the top of the last selected line			*/
		selEnd = endLine > 0? GetHeight( 0, endLine): 0;
	}
	else
	{
		if (vSpan <= 1) 	/* don't try to scroll if 1 line or less	*/
			return;			/* is visible								*/

		selStart = startLine;
		selEnd = endLine;
	}
	
	selPos.h = position.h;
	
	if (selStart >= position.v + vSpan) {
		selPos.v = selStart - vSpan + 1;
		ScrollTo( &selPos, TRUE);
		
	} else if (selEnd < position.v) {
		selPos.v = selEnd;
		ScrollTo( &selPos, TRUE);
	}
}

/**** T E X T   S P E C I F I C A T I O N   M E T H O D S ****/


/******************************************************************************
 SetTextString

		Replace all the text with the contents of a Pascal string.
 ******************************************************************************/

void	CAbstractText::SetTextString(
	Str255		textStr)
{
										/* Pascal string, so text starts at	*/
										/*   byte one, and the number of	*/
										/*   characters is in byte zero		*/
	SetTextPtr((Ptr) textStr + 1, (long) textStr[0]);
}


/******************************************************************************
 SetTextHandle

		Replace all the text with the contents of a handle. Copies the
		text, the caller must dispose of the handle.
 ******************************************************************************/

void	CAbstractText::SetTextHandle(
	Handle		textHand)
{
	SignedByte		saveHState;			/* Original handle attributes		*/
					
	saveHState = HGetState(textHand);	/* Save state of text handle		*/
	HLock(textHand);					/* Lock text in place				*/
	
										/* Another method does the work		*/
	SetTextPtr(*textHand, GetHandleSize(textHand));
	
	HSetState(textHand, saveHState);	/* Restore state of text handle		*/	
}


/******************************************************************************
 SetTextPtr

	Replace all the text with the text in the given buffer.
 ******************************************************************************/

void	CAbstractText::SetTextPtr(
	Ptr			textPtr,			/* Pointer to some text					*/
	long		numChars)			/* Number of characters in the text		*/
{
	SubclassResponsibility();
}


/******************************************************************************
 GetTextHandle

	Returns a handle to the text buffer. The default behavior expected of
	subclasses is that they keep all the text in a single handle and
	return that handle, not a copy. Subclasses that store the text in a more
	sophisticated manner may need to make a copy of the text and return that.
	They must also override callers of GetTextHandle to ensure that the
	handle is eventually disposed.
 ******************************************************************************/

Handle	CAbstractText::GetTextHandle()
{
	SubclassResponsibility();
}


/**** T E X T   C H A R A C T E R I S T I C S   M E T H O D S ****/


/******************************************************************************
 SetFontNumber

		Specify the font for text by font number
 ******************************************************************************/

void	CAbstractText::SetFontNumber(
	short		aFontNumber)
{
	SubclassResponsibility();
}
	

/******************************************************************************
 SetFontName

		Specify the font for text by font name
 ******************************************************************************/

void	CAbstractText::SetFontName(
	Str255		aFontName)
{
	short		fontNum;			/* Corresponding font number			*/
	
									/* Call utility to look up number for	*/
									/*   this font							*/
	GetFontNumber(aFontName, &fontNum);
	
	if (fontNum >= 0) {				/* Legal font numbers are positive		*/
		SetFontNumber(fontNum);
	}
}
	

/******************************************************************************
 SetFontStyle

		Specify the style, such as bold or italic, for text
 ******************************************************************************/

void	CAbstractText::SetFontStyle(
	short		aStyle)
{
	SubclassResponsibility();
}
	

/******************************************************************************
 SetFontSize

		Specify the point size of text
 ******************************************************************************/

void	CAbstractText::SetFontSize(
	short		aSize)
{
	SubclassResponsibility();
}
	

/******************************************************************************
 SetTextMode

		Specify the transfer mode used for drawing text
 ******************************************************************************/

void	CAbstractText::SetTextMode(
	short		aMode)
{
	SubclassResponsibility();
}
	

/******************************************************************************
 SetAlignCmd

		Specify the alignment for text. For simple text, it should set
		the global alignment. For more complex text, it could set
		the alignment of the selection.
 ******************************************************************************/

void	CAbstractText::SetAlignCmd(
	long anAlignCmd)
{
	SubclassResponsibility();
}
	
/******************************************************************************
 GetAlignCmd

	Return the current alignment. For simple text, this would be the global
	alignment. For more complex text editors, this could be the alignment
	of the selection. In this case the alignment may not be continuous over
	the selection and cmdNull would be returned.
 ******************************************************************************/

long CAbstractText::GetAlignCmd( void)
{
	SubclassResponsibility();
}

/******************************************************************************
 SetSpacingCmd

		Specify the vertical spacing between lines of text.  For simple text, 
		it should set the global spacing. For more complex text, it could set
		the spacing of the selection.
 ******************************************************************************/

void	CAbstractText::SetSpacingCmd(
	long		aSpacingCmd)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetHeight

	Return the height of the indicated lines of text
 ******************************************************************************/

long CAbstractText::GetHeight( long startLine, long endLine)
{
	SubclassResponsibility();
}

/******************************************************************************
 Get1Height

	Return the height of a single line of text
 ******************************************************************************/

short CAbstractText::Get1Height( long aLineNum)
{
	return (short) GetHeight( aLineNum, aLineNum);
}

/******************************************************************************
 GetCharOffset

	Return the offset into the text buffer of the character position at
	a point in Frame coordinates.
 ******************************************************************************/

long CAbstractText::GetCharOffset( LongPt *aPt)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetCharPoint

	Return the offset into the text buffer of the character position at
	a point in Frame coordinates.
 ******************************************************************************/

void CAbstractText::GetCharPoint( long offset, LongPt *aPt)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetTextStyle

	Return current style information. whichAttributes is a set of flags
	indicating which text attributes are requested, and upon return, 
	indicates which of the requested attributes are valid in the
	TextStyle record. If the text supports multiple styles, GetTextStyle
	should return information about which of the requested attributes
	are continous over the selection.
	If the text is single style, this can be the global style for all of 
	the text.
	The attributes flags and TextStyle record are defined by TextEdit.
	
******************************************************************************/

void CAbstractText::GetTextStyle( short *whichAttributes, TextStyle *aStyle)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetCharStyle

	Return style information about the charactor at charOffset	
******************************************************************************/

void CAbstractText::GetCharStyle( long charOffset, TextStyle *theStyle)
{
	SubclassResponsibility();
}


/**** C A L I B R A T I N G   M E T H O D S ****/

/******************************************************************************
 ResizeFrame {OVERRIDE}
		
	Called by the TCL when the pane is resized. After resizing to the
	desired size, we may need to ensure that no partial lines are
	displayed at the bottom of the frame.
 ******************************************************************************/

void CAbstractText::ResizeFrame( Rect *deltaRect)
{
	inherited::ResizeFrame( deltaRect);
	SetWholeLines( wholeLines);
	
}

/******************************************************************************
 SetWholeLines

		Specify whether or not the frame should be adjusted so that
		lines are not cut off vertically. The Pane is NOT redrawn.
		
 ******************************************************************************/

void	CAbstractText::SetWholeLines( Boolean aWholeLines)
{
	LongPt		charPos;
	long		charOffset, topLine, bottomLine, totalHeight;
		
	wholeLines = aWholeLines;
	
	frame.bottom = frame.top + height;
	if (wholeLines)
	{
		if (fixedLineHeights)
		{
			// use the height of the first line as the standard height
			
			short lineHeight = Get1Height(1);	
			
			if (lineHeight > 0)
				frame.bottom = frame.top + lineHeight * (height / lineHeight);
		}
		else
		{
			/* find the line at top of frame	*/
			
			SetLongPt( &charPos,  frame.left + 1, frame.top + 1);
			charOffset = GetCharOffset( &charPos);
			topLine = FindLine( charOffset);
			
			/* find the line at bottom of frame	*/
			
			SetLongPt( &charPos, frame.left + 1, frame.top + height - 1);
			charOffset = GetCharOffset( &charPos);
			if (charOffset > 0)
				bottomLine = FindLine( charOffset);	
			else
				bottomLine = GetNumLines();
				
			if (bottomLine > topLine)
			{
				totalHeight = GetHeight( topLine, bottomLine+1);
				
				while ((totalHeight > height) && (bottomLine > topLine))		
				{
					totalHeight = GetHeight( topLine, bottomLine--);
				}
		
				/* we now know how many full lines will fit. resize the frame,	*/
				/* while ensuring that at least one line is visible				*/
	
				if (totalHeight < height)
				{
					frame.bottom = frame.top + totalHeight;
				}
			}
		}
	}
	CalcAperture();	
}

/******************************************************************************
 GetWholeLines

	Return instance variable.
		
 ******************************************************************************/

Boolean CAbstractText::GetWholeLines( void)
{
	return wholeLines;
}

/******************************************************************************
 FindLine

		Return the line number containing the specified character position.
		Both lines and character positions are numbered starting from zero.
		If the character position is before the start of the text (i.e., a
		negative number), a value of zero is returned; if it is beyond the
		end of the text, the number of the last line is returned.
 ******************************************************************************/

long	CAbstractText::FindLine( long charPos)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetLength
 
 	Return the length in bytes of the text buffer
******************************************************************************/

long CAbstractText::GetLength( void)
{
	SubclassResponsibility();
}

/**** C O M M A N D   M E T H O D S ****/
	

/******************************************************************************
 DoCommand {OVERRIDE}

	Handle Font, Size, and Style menu selections, including spacing and
	alignment. Also handles Cut, Copy, Paste, Clear, and Select All commands.
	
 ******************************************************************************/

void	CAbstractText::DoCommand(
	long		theCommand)
{
	Str255			itemName;
	long			number;
	CTextEditTask	*editTask = NULL;
	CTextStyleTask	*styleTask = NULL;
	
	Boolean		makeStyleTask = FALSE;

	if (theCommand < 0) {
	
		switch (HiShort(-theCommand)) {
		
			case MENUfont:				
			case MENUsize:
				makeStyleTask = TRUE;
				break;
				
			default:
				inherited::DoCommand(theCommand);
				break;
		}
		
	} else {

		switch (theCommand) {
				
			case cmdCut:
			case cmdPaste:				
			case cmdClear:
				if (!editable)
				{
					SysBeep(3);
					break;
				}
				/* else fall thru to next case	*/
				
			case cmdCopy:				
				
				ASSERT( wantsClicks == TRUE);
				
				itsTypingTask = NULL;
				editTask = MakeEditTask( theCommand);
				itsLastTask = editTask;
				itsSupervisor->Notify( editTask);
				editTask->Do();
				break;
				
			case cmdSelectAll:
				Prepare();
				SelectAll( TRUE);
				SelectionChanged();
				break;
				
			case cmdPlain:				
			case cmdBold:
			case cmdItalic:
			case cmdUnderline:
			case cmdOutline:
			case cmdShadow:
			case cmdCondense:
			case cmdExtend:			
			case cmdAlignLeft:
			case cmdAlignCenter:
			case cmdAlignRight:
			case cmdSingleSpace:
			case cmd1HalfSpace:
			case cmdDoubleSpace:
				makeStyleTask = TRUE;
				break;
				

			default:
				inherited::DoCommand(theCommand);
				break;
		}
	}
	
		if (makeStyleTask && stylable)
		{
			itsTypingTask = NULL;
			
			styleTask = MakeStyleTask( theCommand);
			itsLastTask = styleTask;
			itsSupervisor->Notify( styleTask);
			styleTask->Do();
		}
}
	

/******************************************************************************
 UpdateMenus {OVERRIDE}

		Perform menu management tasks
 ******************************************************************************/
 
 	static long FindItemCmd( short menuID, StringPtr itemText)
 	{
 		short itemNum = gBartender->FindItemText( menuID, itemText);
 		
 		return itemNum > 0? gBartender->FindCmdNumber( menuID, itemNum) :
 							cmdNull;
 	}

void CAbstractText::UpdateMenus()
{
	MenuHandle		macMenu;
	long			selStart, selEnd, cmdNum;
	TextStyle		style;
	short			styleFlags, fontItem;
	Str255			itemText;

	inherited::UpdateMenus();
	
	if (wantsClicks) gBartender->EnableCmd( cmdSelectAll);

	GetSelection( &selStart, &selEnd);
	
	if (selStart != selEnd) 
	{
		gBartender->EnableCmd(cmdCopy);
		if (editable)
		{
			gBartender->EnableCmd(cmdCut);
			gBartender->EnableCmd(cmdClear);
		}
	}

	if (editable && gClipboard->DataSize('TEXT') > 0)
	{
		gBartender->EnableCmd(cmdPaste);
	}
	
	// The remaining menu updating involves style and formatting
	// commands. Don't do it if the text is not stylable.
	
	if (!stylable) return;
	
	/* Get selection returns the "current" text attributes. For 	*/
	/* single-style text these are the global attributes and are	*/
	/* by definition always continous. For multi-style text, it		*/
	/* returns the attributes of the selections, and the returned	*/
	/* flags indicate which are continuous							*/
	
	styleFlags = doFont | doFace | doSize;
	GetTextStyle (&styleFlags,  &style);
	
	if (styleFlags & doFont)
	{
		/* lastFontNum is the last font number seen from UpdateMenus */
		/* lastFontCmd is the cmd # corresponding to lastFontNum	 */
		/* if lastFontNum == style.tsFont, then we don't need to	 */
		/* lookup the font name in the menu again					 */
		
		if (lastFontNum != style.tsFont)
		{
			lastFontNum = style.tsFont;
			GetFontName( style.tsFont, itemText);
			lastFontCmd = FindItemCmd( MENUfont, itemText);	
		}
		if (lastFontCmd != cmdNull)
			gBartender->CheckMarkCmd( lastFontCmd, TRUE);
	}
	
	if (styleFlags & doSize)
	{
		// Font size of zero means the system default size,
		// but we need a real number for the size menu.
		
		if (style.tsSize == 0)
			style.tsSize = GetDefFontSize();
			
		/* lastTextSize is the last text size seen from UpdateMenus	*/
		/* lastSizeCmd is the cmd # corresponding to lastFontNum	*/
		/* if lastTextSize == style.tsSize, then we don't need to	*/
		/* lookup the size in the menu again						*/
		
		if (lastTextSize != style.tsSize)
		{
			lastTextSize = style.tsSize;
			NumToString( style.tsSize, itemText);
			lastSizeCmd = FindItemCmd( MENUsize, itemText);	
		}
		if (lastSizeCmd != cmdNull)
			gBartender->CheckMarkCmd( lastSizeCmd, TRUE);
	}
		
	if (styleFlags & doFace)
	{
		if (style.tsFace == NOTHING) {
			gBartender->CheckMarkCmd(cmdPlain, TRUE);
		} else {
			if (style.tsFace & bold) {
				gBartender->CheckMarkCmd(cmdBold, TRUE);
			}
			if (style.tsFace & italic) {
				gBartender->CheckMarkCmd(cmdItalic, TRUE);
			}
			if (style.tsFace & underline) {
				gBartender->CheckMarkCmd(cmdUnderline, TRUE);
			}
			if (style.tsFace & outline) {
				gBartender->CheckMarkCmd(cmdOutline, TRUE);
			}
			if (style.tsFace & shadow) {
				gBartender->CheckMarkCmd(cmdShadow, TRUE);
			}
			if (style.tsFace & condense) {
				gBartender->CheckMarkCmd(cmdCondense, TRUE);
			}
			if (style.tsFace & extend) {
				gBartender->CheckMarkCmd(cmdExtend, TRUE);
			}
		}
	}
	
	cmdNum = GetAlignCmd();
	if (cmdNum != cmdNull)
		gBartender->CheckMarkCmd( cmdNum, TRUE); 
		
	cmdNum = GetSpacingCmd();
	if (cmdNum != cmdNull)
		gBartender->CheckMarkCmd( cmdNum, TRUE); 
		
}


/******************************************************************************
 DoKeyDown {OVERRIDE}

	Handles keystrokes as follows:
	
		1. The Home, Page up, Page down, and all command keys are passed
		   to the superclass, where they can travel up the command chain.
		2. The End key handling is overriden to scroll to the bottom left
		   of the text.
		3. Cursor keys are passed to the superclass, and the current typing
		   task, if any, is notified of the change in selection.
		4. All other keystrokes are handled by the current typing task.
 ******************************************************************************/

void	CAbstractText::DoKeyDown(
	char		theChar,
	Byte		keyCode,
	EventRecord	*macEvent)
{
    
    LongPt   thePosition;
    long    theHExtent, theVExtent;
    CTextEditTask *typingTask;
        
    		/* Don't try to insert and command keys into the text,	*/
    		/* pass them along the command chain					*/
    		
    if (macEvent->modifiers & cmdKey)
    {
    	inherited::DoKeyDown( theChar, keyCode, macEvent);
    	return;
    }
    
    switch (keyCode) {
    
        case KeyHome:
        case KeyPageUp:
        case KeyPageDown:
             inherited::DoKeyDown(theChar, keyCode, macEvent);
             break;
             
        case KeyEnd:
            if (itsScrollPane != NULL) {
                GetExtent(&theHExtent, &theVExtent);
                thePosition.h = 0;
                thePosition.v = Max(0, theVExtent - itsScrollPane->vSpan);
                ScrollTo( &thePosition, TRUE);
                itsScrollPane->Calibrate();
            }
            break;
            
        case KeyLeftCursor:
        case KeyRightCursor:
        case KeyUpCursor:
        case KeyDownCursor:
        	if (editable)	/* TCL 1.1.1 DLP 9/27/91 */
        	{
        		TypeChar( theChar, macEvent->modifiers);
        		SelectionChanged();
        	}
        	break;
            
        default:
        	if (editable)	/* TCL 1.1.1 DLP 9/27/91 */
        	{
	        	if (itsTypingTask && !itsTypingTask->CanStillType())
	        	{
	        		itsSupervisor->Notify( NULL);
	        		itsTypingTask = NULL;
	        		itsLastTask = NULL;
	        	}
	        	if (!itsTypingTask)
	        	{
	        		typingTask = MakeEditTask( cmdNull);
	        		itsLastTask = typingTask;
	        		itsSupervisor->Notify( typingTask);
	        		itsTypingTask = typingTask;
	        	}
	        	itsTypingTask->DoTyping( theChar, keyCode, macEvent);
	        }
            break;
    }
}
	

/******************************************************************************
 DoAutoKey {OVERRIDE}

		Handle a key being held down. Treat this the same as a normal
		keystroke.
 ******************************************************************************/

void	CAbstractText::DoAutoKey(
	char		theChar,
	Byte		keyCode,
	EventRecord	*macEvent)
{
	if (!(macEvent->modifiers & cmdKey)) {
		DoKeyDown(theChar, keyCode, macEvent);
	}
}

/******************************************************************************
 TypeChar

	Process a single keystroke. All undo setup has already been performed,
	and the keystroke should be handled directly.
 ******************************************************************************/

void CAbstractText::TypeChar(char theChar, short theModifers)
{
	SubclassResponsibility();
}	

/******************************************************************************
 BecomeGopher {OVERRIDE}
 
 	If fBecoming is TRUE the pane is becoming the gopher, otherwise
 	it is about to stop being the gopher.
******************************************************************************/

Boolean CAbstractText::BecomeGopher( Boolean fBecoming)
{
	if (!inherited::BecomeGopher( fBecoming))
		return FALSE;
		
	if (fBecoming)
	{
		Activate();
		if (stylable)
		{
			gBartender->EnableMenu( MENUfont);
			gBartender->EnableMenu( MENUsize);
			gBartender->EnableMenu( MENUstyle);
		}
	}
	else
	{
		Deactivate();
		if (stylable)
		{
			gBartender->DisableMenu( MENUfont);
			gBartender->DisableMenu( MENUsize);
			gBartender->DisableMenu( MENUstyle);
		}
	}
	return TRUE;
}

/******************************************************************************
 SetSelection
 
 	Sets the selected text to the range corresponding to character positions
 	selStart through selEnd.
******************************************************************************/

void CAbstractText::SetSelection( long selStart, long selEnd, Boolean fRedraw)
{
	SubclassResponsibility();

}	/* CAbstractText::SetSelection */

/******************************************************************************
 SelectAll
 
 	Selects all the text
******************************************************************************/

void CAbstractText::SelectAll( Boolean fRedraw)
{
	SetSelection( 0, GetLength(), fRedraw);

}	/* CAbstractText::SelectAll */


/******************************************************************************
 AdjustCursor {OVERRIDE}

		Mouse is inside an AbstractText Pane. Use the standard IBeam cursor.
 ******************************************************************************/

void	CAbstractText::AdjustCursor(
	Point		where,					/* Mouse location in Window coords	*/
	RgnHandle	mouseRgn)
{
	extern CursHandle	gIBeamCursor;

	SetCursor(*gIBeamCursor);
}

/******************************************************************************
 GetSpacingCmd

	Return line spacing.
 ******************************************************************************/

long CAbstractText::GetSpacingCmd( void)
{
	SubclassResponsibility();
}

/******************************************************************************
 GetSelection

	Return the start and end of the selection.
 ******************************************************************************/

void CAbstractText::GetSelection( long *startSel, long *endSel)
{
	SubclassResponsibility();

}

/******************************************************************************
 GetNumLines

	Return the number of lines of text.
 ******************************************************************************/

long CAbstractText::GetNumLines( void)
{
	SubclassResponsibility();
}

/******************************************************************************
 Specify

	Specify whether the text is selectable and or editable. Note that
	the text cannot be editable or stylable without also being selectable.
	
 ******************************************************************************/

void CAbstractText::Specify( Boolean fEditable, Boolean fSelectable, Boolean fStylable)
{
	if (!fSelectable) 
	{
		fEditable = fStylable = FALSE;
	}
	
	editable = fEditable;
	SetWantsClicks( fSelectable);
	
	stylable = fStylable;
	
		// if the text is selectable then it can be the gopher
	SetCanBeGopher( fSelectable); /* TCL 1.1.1 DLP 9/27/91 */
}

/******************************************************************************
 GetSpecification

	Return characteristics of the text pane.
		
 ******************************************************************************/

void CAbstractText::GetSpecification( Boolean *fEditable, Boolean *fSelectable, 
	Boolean *fStylable)
{
	*fEditable = editable;
	*fSelectable = wantsClicks;
	*fStylable = stylable;
}

/******************************************************************************
 CopyTextRange

	Return a copy of the currently selected text.
 ******************************************************************************/

Handle CAbstractText::CopyTextRange( long start, long end)
{
	SubclassResponsibility();
}

/******************************************************************************
 PerformEditCommand

	Perform the standard cut, copy, paste, and clear commands on the
	text.
 ******************************************************************************/

void CAbstractText::PerformEditCommand( long theCommand)
{
	SubclassResponsibility();
}

/******************************************************************************
 InsertTextPtr

	Insert a copy of the given text at the start of the selection.
 ******************************************************************************/

void CAbstractText::InsertTextPtr( Ptr text, long length, Boolean fRedraw)
{
	SubclassResponsibility();
}

/******************************************************************************
 InsertTextHandle

	Insert a copy of the given text at the start of the selection.
 ******************************************************************************/

void CAbstractText::InsertTextHandle( Handle text, Boolean fRedraw)
{
	SignedByte state = HGetState( text);
	
	MoveHHi( text);
	HLock( text);
	
	InsertTextPtr( *text, GetHandleSize( text), fRedraw);
	
	HSetState( text, state);
}

/******************************************************************************
 SelectionChanged

	Called after the selection may have changed to notify dependants.
 ******************************************************************************/

void CAbstractText::SelectionChanged( void)
{
	BroadcastChange( textSelectionChanged, NULL);
	if (itsTypingTask)
		itsTypingTask->SelectionChanged();
}

/******************************************************************************
 MakeEditTask

	Create a task to carry out typing, cut, copy, paste, and clear commands.
	Must be a subclass of CTextEditTask.
 ******************************************************************************/

CTextEditTask *CAbstractText::MakeEditTask( long editCmd)
{
	CTextEditTask *newTask = NULL;
	
	TRY
	{
		newTask = new( CTextEditTask);
		
		newTask->ITextEditTask( this, editCmd, cFirstTaskIndex);
	}
	CATCH
	{
		ForgetObject( newTask);
	}
	ENDTRY;
	
	return newTask;
}

/******************************************************************************
 MakeStyleTask

	Create a task to carry out font, style, size, alignment, and spacing
	commands. Must be a subclass of CTextStyleTask.
 ******************************************************************************/

CTextStyleTask *CAbstractText::MakeStyleTask( long styleCmd)
{
	CTextStyleTask *newTask = NULL;
	short			taskIndex;
	
	TRY
	{
		taskIndex = cFirstTaskIndex > 0 ? cFirstTaskIndex + undoFormatting : 0;
		newTask = new( CTextStyleTask);
		
		newTask->ITextStyleTask( this, styleCmd, taskIndex);
	}
	CATCH
	{
		ForgetObject( newTask);
	}
	ENDTRY;
	
	return newTask;
}

/******************************************************************************
 GetCharBefore

	Return the character before aPosition. The character and its size
	are returned in charBuf, and aPosition is updated with the starting
	position of the character. If there is no preceding character, then
	Length(charBuf) is 0, and aPosition is not updated.
	
 ******************************************************************************/

void CAbstractText::GetCharBefore( long *aPosition, tCharBuf charBuf)
{
	Handle	h;
	register Ptr text;
	short	charSize, i;
	long	charPos;
	
					/* nothing in this method moves memory	*/
	h = GetTextHandle();
	text = *h;
	
	if ((*aPosition > 0) && (GetHandleSize( h) > 0))
	{
		charPos = *aPosition - 1;
		
		if (gSystem.scriptsInstalled > 1)
		{
			
			/* must use CharByte to get character.			*/
			/* CharByte does not move memory				*/
				
			while (CharByte( text, charPos--) > 0)
				;
				
			charSize = *aPosition - charPos;
			*aPosition = charPos;
			
			Length(charBuf) = charSize;
			for (i = 1; i <= charSize; i++)
				charBuf[i] = text[charPos++];
		}
		else
		{
			*aPosition = charPos;
			Length(charBuf) = 1;
			charBuf[1] = text[charPos];
		}
	}
	else Length(charBuf) = 0;

}

/******************************************************************************
 GetCharAfter

	Return the character after aPosition. The character and its size
	are returned in charBuf, and aPosition is updated with the starting
	position of the character. If there is no following character, then
	Length(charBuf) is 0, and aPosition is not updated.

 ******************************************************************************/

void CAbstractText::GetCharAfter( long *aPosition, tCharBuf charBuf)
{
	Handle	h;
	register Ptr text;
	long	charPos, textLen;
	
					/* nothing in this method moves memory	*/
	h = GetTextHandle();
	textLen = GetLength();
	text = *h;
	
	if ((textLen > 0) && (*aPosition < textLen))
	{
		charPos = *aPosition;
		
		if (gSystem.scriptsInstalled > 1)
		{
			
			/* must use CharByte to get character.			*/
			/* CharByte does not move memory				*/
			
			Length(charBuf) = 0;
				
			while ((charPos <= textLen) && (CharByte( text, charPos) > 0))
				charBuf[ ++charBuf[0]] = text[charPos++];
				
			*aPosition = charPos;
		}
		else
		{
			*aPosition = charPos;
			Length(charBuf) = 1;
			charBuf[1] = text[charPos];
		}
	}
	else Length(charBuf) = 0;
}

/******************************************************************************
 Paginate
******************************************************************************/

void CAbstractText::Paginate( CPrinter *aPrinter, short pageWidth, short pageHeight)
{
	short	linesPerPage, lineHeight, stripNum;
	long	currLine, numLines, nextPageBreak, totalPix;
	
	if (fixedLineHeights)
	{
		/* force pageHeight to include only full lines	*/
	
				/* use height of line 1 as height of all lines */
		lineHeight = Get1Height( 1);
		
		linesPerPage = pageHeight / lineHeight;
		pageHeight = lineHeight * linesPerPage;
	
		/* CPanorama will paginate into equal sized strips given	*/
		/* the new page height										*/
	
		inherited::Paginate( aPrinter, pageWidth, pageHeight);
	}
	else
	{
		aPrinter->ResetPagination();
		numLines = GetNumLines();
		nextPageBreak = pageHeight;
		stripNum = totalPix = 0;
		
		aPrinter->SetHorizPageBreak( 1, pageWidth);
		
		for( currLine = 1; currLine <= numLines; currLine++)
		{
			lineHeight = Get1Height( currLine);
			totalPix += lineHeight;
			if (totalPix > nextPageBreak)
			{
				aPrinter->SetVertPageBreak( ++stripNum, totalPix - lineHeight);
				totalPix = nextPageBreak + lineHeight;
				nextPageBreak += pageHeight;
			}
		}
		if (totalPix > nextPageBreak - pageHeight)
		{
			aPrinter->SetVertPageBreak( ++stripNum, nextPageBreak);
		}
	}	
}
