/******************************************************************************
 CEditText.c

							The EditText Class
		
	SUPERCLASS = CEditText
	
	Copyright  1989 Symantec Corporation. All rights reserved.

	TCL 1.1 CHANGES
	[
		CEditText is now a subclass of a new class, CAbstractText. Much
		of the old behavior of CEditText/CEditText, as well as new
		behavior, like undo, is implemented by CAbstractText.
		
		Any subclass of CAbstractText, including CEditText, inherits two
		attributes selectable and editable, which are set via the
		CAbstractText::Specify method. The possible combinations and
		behavior are as follows:
		
		Editable	Selectable		Behavior
		--------	----------		--------
		TRUE		TRUE			can type and select, like old CEditText
		TRUE		FALSE			invalid combination
		FALSE		FALSE			no typing or selecting, like old CEditText
		FALSE		TRUE			no typing, but text can be selected and
									copied to the clipboard.
									
		CEditText now supports undo via behavior inherited from CAbstractText.
		
		The SetLineSpacing method has been renamed to SetSpacingCmd in order
		to be consistent with other methods, such as Get/SetAlignCmd.
	]

 ******************************************************************************/


#include "CEditText.h"
#include "CClipboard.h"
#include "Commands.h"
#include "TCLUtilities.h"
#include "Constants.h"
#include "CScrollPane.h"
#include "Global.h"

/*** Global Variables ***/

extern CClipboard	*gClipboard;	/* Copies and Pastes data			*/
extern EventRecord  gLastMouseUp;
extern CBureaucrat	*gGopher;		/* TCL 1.1.1 DLP 9/27/91			*/

CEditText	*gEditText;				/* Current EditText object			*/


/**** C O N S T R U C T I O N / D E S T R U C T I O N   M E T H O D S ****/


/******************************************************************************
 IEditText

		Initialize an EditText object
 ******************************************************************************/

void	CEditText::IEditText(
	CView			*anEnclosure,
	CBureaucrat		*aSupervisor,
	short			aWidth,
	short			aHeight,
	short			aHEncl,
	short			aVEncl,
	SizingOption	aHSizing,
	SizingOption	aVSizing,
	short			aLineWidth)
{
	
	CAbstractText::IAbstractText(anEnclosure, aSupervisor,
					 aWidth, aHeight, aHEncl, aVEncl, aHSizing, aVSizing,
					 aLineWidth);
					
	wholeLines = TRUE;
	IEditTextX();
}

/******************************************************************************
 IViewTemp

		Initialize an EditText object for a template
 ******************************************************************************/

void CEditText::IViewTemp(CView *anEnclosure, CBureaucrat *aSupervisor,
						Ptr viewData)
{
	inherited::IViewTemp( anEnclosure, aSupervisor, viewData);
	IEditTextX();

}	/* CEditText::IViewTemp */

/******************************************************************************
 IEditTextX

		Common initialization for an EditText object
 ******************************************************************************/

void CEditText::IEditTextX(void)
{
	Str255			itemStr;
	FontInfo	macFontInfo;			/* Info about size of font			*/

	ForceNextPrepare();
	SetPort(macPort);
	
	MakeMacTE();						/* Create the TE record */
  
	SetWholeLines(wholeLines);
	
	GetFontInfo( &macFontInfo);			/* OK to just call GetFontInfo here	*/
										/*   since the new TE Record uses	*/
										/*   the same font characteristics	*/
										/*   as the current port			*/
	SetScales(macFontInfo.widMax, (**macTE).lineHeight);
	AdjustBounds();
	
	autoRefresh = lineWidth <= 0;
	spacingCmd = cmdSingleSpace;

	if ((**macTE).txSize == 0) {
		(**macTE).txSize = SYS_FONT_SIZE;
	}
	SetClikLoop((ClikLoopProcPtr) TEClickLoop, macTE);
	CalcTERects();						/* TCL 1.1.2 DLP 12/9/91 */

}	/* CEditText::IEditTextX */	


/**** M O U S E   A N D   K E Y S T R O K E   M E T H O D S ****/


/******************************************************************************
 DoClick {OVERRIDE}

		Respond to a mouse click within the EditText
 ******************************************************************************/

void	CEditText::DoClick(
	Point		hitPt,					/* Mouse location in Frame coords	*/
	short		modifierKeys,			/* State of modifier keys			*/
	long		when)					/* Tick time of mouse click			*/
{ 	
	TEClick(hitPt, (modifierKeys & shiftKey) != NOTHING, macTE);
	
		// CSwitchboard will never see the mouse up that ended
		// the drag, so we stuff gLastMouseUp here to allow
		// multi-click counting to work.
		
	gLastMouseUp.what = mouseUp;
	gLastMouseUp.when = TickCount();
	gLastMouseUp.where = hitPt;
	LocalToGlobal( &gLastMouseUp.where);
	gLastMouseUp.modifiers = modifierKeys;

	SelectionChanged();
	
		/* TCL 1.1.1 DLP 9/27/91 */
		/* For non-editable but selectable text, we want to display	*/
		/* a selection range so the user can copy, but we don't		*/
		/* want to display a caret. In that case, we resign the		*/
		/* gopher to our supervisor, which will deactivate this text*/

	if (!editable && (gGopher == this))
	{
		long selStart, selEnd;
		
		GetSelection( &selStart, &selEnd);	
		if (selStart == selEnd)
			itsSupervisor->BecomeGopher( TRUE);
	}
}
	
/******************************************************************************
 Type

	Process a single keystroke
 ******************************************************************************/

void CEditText::TypeChar(char theChar, short theModifers)
{
	
		// check whether the char will exceed TE capacity
	
	if ((theChar >= ' ') || (theChar == '\t') || (theChar == '\r'))
		CheckInsertion( 1, TRUE);
	
	Prepare();
	TEKey(theChar, macTE);
	AdjustBounds();
	ScrollToSelection();
}

/******************************************************************************
 CheckInsertion

	Ensure that attempt to insert text does not exceed TextEdit's capacity or
	the available memory.
	If useSelection is TRUE, the length of the selection is deducted from the
	total length.
	
 ******************************************************************************/

void CEditText::CheckInsertion( long insertLen, Boolean useSelection)
{
	long newLength = (**macTE).teLength + insertLen;
	long growSize;
	
	if (useSelection)
		newLength -= (**macTE).selEnd - (**macTE).selStart;
	
	if (newLength > kMaxTELength)
		Failure( paramErr, excExceedTELimit);
		
		// check for available memory, but only for insertions
		// of more than one character
		
	growSize = newLength - (**macTE).teLength;
	if (growSize > 0)
	{
		Handle h = NewHandleCanFail( growSize);
		
		FailNIL( h);
		DisposHandle( h);
	}
}

/**** D I S P L A Y   M E T H O D S ****/
	

/******************************************************************************
 Activate {OVERRIDE}

		Activate EditText by hilighting the selection or showing the
		text insertion caret. Also enable applicable menu commands.
 ******************************************************************************/

void	CEditText::Activate()
{
	inherited::Activate();
	
	Prepare();
	TEActivate(macTE);
	gEditText = this;
	
}
	

/******************************************************************************
 Deactivate {OVERRIDE}

		Deactivate EditText by unhiliting the selection or hiding the
		text insertion caret. Also disable the appropriate menu commands.
 ******************************************************************************/

void	CEditText::Deactivate()
{
	inherited::Deactivate();

	Prepare();
	if (macTE) TEDeactivate(macTE);
	gEditText = NULL;
}

/******************************************************************************
 SetSelection
 
 	Sets the selected text to the range corresponding to character positions
 	selStart through selEnd.
******************************************************************************/

void CEditText::SetSelection( long selStart, long selEnd, Boolean fRedraw)
{
	
	if (!fRedraw)
	{
		(**macTE).selStart = (short) selStart;
		(**macTE).selEnd = (short) selEnd;
	}
	else	
	{
		Prepare();	
		TESetSelect( selStart, selEnd, macTE);
	}
	
}	/* CEditText::SetSelection */



/**** P R I N T I N G   M E T H O D S ****/
	

/******************************************************************************
 AboutToPrint {OVERRIDE}

		The specified range of pages is about to be printed.
 ******************************************************************************/

void	CEditText::AboutToPrint(
	short		*firstPage,
	short		*lastPage)
{
	if (active) {
		Prepare();						/* Must make sure that the			*/
		HidePen();						/*   current selection is not		*/
		TEDeactivate(macTE);			/*   hilited on the print out		*/
		ShowPen();
	}

	inherited::AboutToPrint(firstPage, lastPage);
}
	

/******************************************************************************
 DonePrinting {OVERRIDE}

		Print loop has been completed. Re-hilite the active selection.
 ******************************************************************************/

void	CEditText::DonePrinting()
{
	inherited::DonePrinting();

	CalcTERects();	/* TCL 1.1: call CalcTERects rather than setting */
					/* the viewRect directly */
	
	if (active) {						/* Restore active state				*/
		Prepare();
		HidePen();
		TEActivate(macTE);
		ShowPen();
	}
}	

/**** C U R S O R   M E T H O D S ****/


/******************************************************************************
 Dawdle {OVERRIDE}

		The user isn't doing anything, so flash the text insertion cursor
 ******************************************************************************/

void	CEditText::Dawdle(
	long	*maxSleep)
{
	if (editable)	/* TCL 1.1.1 DLP 9/27/91 */
	{
		Prepare();
		TEIdle(macTE);
		*maxSleep = GetCaretTime();
	}
}
	

/*=============================================================================
 TEClickLoop

		Function for auto-scrolling text
 =============================================================================*/
		
pascal Boolean	TEClickLoop()
{
	Point		mouseLoc;
	LongPt		longMouse;

	if (gEditText != NULL) {
		GetMouse(&mouseLoc);
		QDToLongPt( mouseLoc, &longMouse);
		gEditText->AutoScroll( &longMouse);
	}
	return(TRUE);
}
	
/******************************************************************************
 MakeMacTE
 
 	Creates the TextEdit record
******************************************************************************/
void CEditText::MakeMacTE( void)
{
	LongRect	longInterior;
    Rect        tFrame;
    Boolean		savedAlloc;

	GetInterior( &longInterior);
	LongToQDRect( &longInterior, &tFrame);
	
	savedAlloc = SetAllocation( kAllocCanFail);
    macTE = TENew(&tFrame, &tFrame);
    SetAllocation( savedAlloc);
    
    FailNIL( macTE);
}

/******************************************************************************
 Dispose {OVERRIDE}

		Dispose of a EditText by throwing out the Toolbox TextEdit record
 ******************************************************************************/

void	CEditText::Dispose()
{
	TEDispose(macTE);					/* Get rid of TextEdit record		*/
	macTE = NULL;
	inherited::Dispose();				/* Pass message on to superclass	*/
}


/**** D I S P L A Y   M E T H O D S ****/
	

/******************************************************************************
 Draw {OVERRIDE}

		Draw the EditText
 ******************************************************************************/

void	CEditText::Draw(
	Rect		*area)
{
	(**macTE).inPort = thePort; 
	TEUpdate(area, macTE);
	(**macTE).inPort = macPort; 
	
}
	

/******************************************************************************
 Scroll {OVERRIDE}

		Scroll the text within the pane
 ******************************************************************************/

void	CEditText::Scroll(
	long		hDelta,
	long		vDelta,
	Boolean		redraw)
{
	short		hPixels;
	short		vPixels;
	
	hPixels = hDelta * hScale;
	vPixels = vDelta * vScale;

	OffsetRect(&(**macTE).viewRect, hPixels, vPixels);
		
	inherited::Scroll(hDelta, vDelta, redraw);
}


/**** P R I N T I N G   M E T H O D S ****/


/******************************************************************************
 PrintPage {OVERRIDE}

		Print the specified page.
 ******************************************************************************/

void	CEditText::PrintPage(
	short		pageNum,				/* Number of the page to print		*/
	short		pageWidth,				/* Width of printed page in pixels	*/
	short		pageHeight,				/* Height of page in pixels			*/
	CPrinter 	*aPrinter)
{	
	(**macTE).inPort = thePort;			/* Must change the port in the TE	*/
										/*   record to the current port,	*/
										/*   which will be a printer port	*/
	
	if (printClip == clipPAGE) {		/* Expand viewRect to the size of a	*/
										/*   page. It will be restored by	*/
										/*   the DonePrinting() method.		*/
		(**macTE).viewRect.right = (**macTE).viewRect.left + pageWidth;
		
		if (wholeLines) {
			(**macTE).viewRect.bottom = (**macTE).viewRect.top +
											vScale * (pageHeight / vScale);
		} else {
			(**macTE).viewRect.bottom = (**macTE).viewRect.top + pageHeight;
		}
	}

	inherited::PrintPage(pageNum, pageWidth, pageHeight, aPrinter);
			
	(**macTE).inPort = macPort;			/* Restore screen port				*/
}
	

/**** T E X T   S P E C I F I C A T I O N   M E T H O D S ****/

/******************************************************************************
 SetTextPtr

		Pass a Pointer and a character count for text to be used in
		a EditText. Note that TextEdit makes a copy of the text.
 ******************************************************************************/

void	CEditText::SetTextPtr(
	Ptr			textPtr,			/* Pointer to some text					*/
	long		numChars)			/* Number of characters in the text		*/
{
									/* Stuff the text in the Mac TE record	*/
	TESetText(textPtr, numChars, macTE);
	AdjustBounds();
	Refresh();						/* Draw the new text					*/
}


/******************************************************************************
 GetTextHandle

		Retrieve a handle to the text in a EditText Pane
 ******************************************************************************/

Handle	CEditText::GetTextHandle()
{
	return( (Handle) TEGetText(macTE));
}


/**** T E X T   C H A R A C T E R I S T I C S   M E T H O D S ****/


/******************************************************************************
 SetFontNumber

		Specify the font for text by font number
 ******************************************************************************/

void	CEditText::SetFontNumber(
	short		aFontNumber)
{
	FontInfo	macFontInfo;			/* Info about size of font			*/

	(**macTE).txFont = aFontNumber;
	TECalText(macTE);
	
	SetSpacingCmd( spacingCmd);
	
	GetTEFontInfo(&macFontInfo);
	SetScales(macFontInfo.widMax, (**macTE).lineHeight);
	AdjustBounds();
	
	Refresh();
	SetWholeLines(wholeLines);
	Refresh();
}

/******************************************************************************
 SetFontStyle

		Specify the style, such as bold or italic, for text
 ******************************************************************************/

void	CEditText::SetFontStyle(
	short		aStyle)
{
	if (aStyle == NOTHING) {			/* Plain text is the absence of		*/
		(**macTE).txFace = NOTHING;		/*   any style						*/
	
	} else {
		(**macTE).txFace ^= aStyle;		/* Toggle style characteristic		*/
										/*   by XOR'ing proper bit			*/
	}
	
	Refresh();
	SetWholeLines(wholeLines);
	Refresh();
}
	

/******************************************************************************
 SetFontSize

		Specify the point size of text
 ******************************************************************************/

void	CEditText::SetFontSize(
	short		aSize)
{
	(**macTE).txSize = aSize;
	TECalText(macTE);
	SetSpacingCmd(spacingCmd);
}
	

/******************************************************************************
 SetTextMode

		Specify the transfer mode used for drawing text
 ******************************************************************************/

void	CEditText::SetTextMode(
	short		aMode)
{
	(**macTE).txMode = aMode;
	Refresh();
}
	

/******************************************************************************
 SetAlignment

		Specify the alignment for text
 ******************************************************************************/

void	CEditText::SetAlignment(
	short		anAlignment)
{
	TESetJust(anAlignment, macTE);
	Refresh();
}
	
/******************************************************************************
 SetAlignCmd

		Specify the alignment for text
 ******************************************************************************/

void	CEditText::SetAlignCmd( long anAlignCmd)
{
	short teAlign;
	
	alignCmd = anAlignCmd;
	switch( alignCmd)
	{
		case cmdAlignLeft:
			teAlign = teFlushDefault;
			break;
		case cmdAlignCenter:
			teAlign = teCenter;
			break;
		case cmdAlignRight:
			teAlign = teFlushRight;
			break;
		default:
			teAlign = teFlushDefault;
	}
	SetAlignment( teAlign);
} 

/******************************************************************************
 SetSpacingCmd

		Specify the vertical spacing between lines of text
 ******************************************************************************/

void	CEditText::SetSpacingCmd(
	long		aSpacingCmd)
{
	FontInfo	macFontInfo;			/* Info about size of font			*/
	short		extra;					/* Extra space between lines		*/
	
	spacingCmd = aSpacingCmd;			/* Save line spacing setting		*/

	GetTEFontInfo(&macFontInfo);
										/* Set to single spacing			*/
	(**macTE).lineHeight = macFontInfo.ascent + macFontInfo.descent +
							macFontInfo.leading;
	(**macTE).fontAscent = macFontInfo.ascent;
	
	switch (aSpacingCmd) {				/* Figure out how much extra space	*/
										/*   is needed						*/
		case cmdSingleSpace:
			extra = 0;
			break;
			
		case cmd1HalfSpace:
			extra = (**macTE).lineHeight / 2;
			break;
			
		case cmdDoubleSpace:
			extra = (**macTE).lineHeight;
			break;
	}
	
	(**macTE).lineHeight += extra;		/* Add in extra space				*/
	(**macTE).fontAscent += extra;
	
	Refresh();
	SetScales(macFontInfo.widMax, (**macTE).lineHeight);
	SetWholeLines(wholeLines);
	CalcAperture();
	AdjustBounds();
	Refresh();
}

long CEditText::GetSpacingCmd( void)
{
	return spacingCmd;
}

long CEditText::GetAlignCmd( void)
{
	return alignCmd;
}

/******************************************************************************
 GetTEFontInfo

		Return a QuickDraw FontInfo record for the font used by the TE record
		associated with a EditText Pane. This method changes the font
		family, style, and size of the Pane's port to that used by the
		TextEdit record so the Toolbox trap GetFontInfo can be used.
 ******************************************************************************/

void	CEditText::GetTEFontInfo(
	FontInfo		*macFontInfo)
{
	SetPort(macPort);					/* Set font characteristics			*/
	ForceNextPrepare();
	TextFont((**macTE).txFont);
	TextFace((**macTE).txFace);
	TextSize((**macTE).txSize);
	
	GetFontInfo(macFontInfo);			/* Get info about font				*/
}


/**** C A L I B R A T I N G   M E T H O D S ****/


/******************************************************************************
 CalcTERects

		Determine the destination and view rectangles used by TextEdit
		from the values of the EditText's frame and lineWidth
		instance variables.

 ******************************************************************************/

void	CEditText::CalcTERects()
{
	LongRect	interior;	
	short		hDelta;
	short		vDelta;
	
	GetInterior( &interior);

	hDelta = position.h * hScale - interior.left + (**macTE).destRect.left;
	vDelta = position.v * vScale - interior.top + (**macTE).destRect.top;
	
	frame.left += hDelta;
	frame.right += hDelta;
	frame.top += vDelta;
		
	hOrigin += hDelta;
	vOrigin += vDelta;


	(**macTE).destRect.right = (lineWidth > 0) ?
									(**macTE).destRect.left + lineWidth :
									interior.right;
	
	LongToQDRect( &interior, &(**macTE).viewRect);

	TECalText( macTE);
}

	
/******************************************************************************
 ResizeFrame {OVERRIDE}

		Adjust a EditText's frame when the size of the pane changes. The
		delta rectangle parameter specifies the change in location of each
		side of the Pane. Positive values mean a move to the right or down;
		negative values mean a move to the left or up.

		After using the inherited method, we have to adjust the frame so
		that there's no fractional line showing at the bottom of the frame,
		and also adjust the bounds since the text could be set to wrap
		to the frame.
 ******************************************************************************/

void	CEditText::ResizeFrame(
	Rect		*delta)
{
	inherited::ResizeFrame(delta);

	(**macTE).destRect.left += delta->left;
	(**macTE).destRect.top += delta->top;
		
	CalcTERects();	
	AdjustBounds();
}
	
/******************************************************************************
 AdjustBounds

		Adjust the bounds of a EditText to match the associated TextEdit
		record. Send this message whenever something happens which can
		affect the number of lines of text or the line width.
 ******************************************************************************/

void	CEditText::AdjustBounds()
{
	register TEPtr	t;
	
	t = *macTE;
	
	bounds.left = bounds.top = 0;
	bounds.bottom = GetNumLines();

	if (lineWidth > 0) {
		bounds.right = lineWidth;
	} else {
		bounds.right = t->viewRect.right - t->viewRect.left;
	}
        
    bounds.right = (bounds.right - 1) / hScale + 1;

	if (itsScrollPane != NULL) {
		itsScrollPane->AdjustScrollMax();
	}
}
	

/******************************************************************************
 FindLine

		Return the line number containing the specified character position.
		Both lines and character positions are numbered starting from zero.
		If the character position is before the start of the text (i.e., a
		negative number), a value of zero is returned; if it is beyond the
		end of the text, the number of the last line is returned.
 ******************************************************************************/

long	CEditText::FindLine(
	long		charPos)
{
	register TEPtr		t;
	long				lineNum;
	
	t = *macTE;
										/* Check if it's in line zero		*/
	if ( (t->nLines == 0) || (charPos < t->lineStarts[1])) {
		return(0);
	}
										/* Check if it's in the last line	*/
	if (charPos >= t->lineStarts[t->nLines - 1]) {
		if (charPos == t->teLength) {	/* Special case for the last char	*/
			return(bounds.bottom - 1);
		} else {
			return(t->nLines - 1);
		}
	}
										/* Linear search thru lineStarts	*/
										/* +++ Implement binary search		*/
	lineNum = 1;						
	while (charPos >= t->lineStarts[lineNum + 1]) {
		lineNum++;
	}
	return(lineNum);
}

/******************************************************************************
 GetLength
 
 	Return the length in bytes of the TERecord's buffer
******************************************************************************/

long CEditText::GetLength( void)
{
	return (**macTE).teLength;
}

/******************************************************************************
 GetTextStyle
 
 	Returns information about the current style. This is the global style
 	information for standard text edit, and is the style of the selection
 	for styled text.
******************************************************************************/

void CEditText::GetTextStyle( short *whichAttributes, TextStyle *aStyle)
{
	TEContinuousStyle( whichAttributes, aStyle, macTE);
}

/******************************************************************************
 GetCharStyle

	Return style information about the charactor at charOffset	
******************************************************************************/

void CEditText::GetCharStyle( long charOffset, TextStyle *theStyle)
{
	short ignoreLineHeight, ignoreAscent;
	
	TEGetStyle( charOffset, theStyle, &ignoreLineHeight, &ignoreAscent, macTE);
}


/******************************************************************************
 GetSelection
 
 	Return the start and end of the currently selected text.
******************************************************************************/

void CEditText::GetSelection( long *selStart, long *selEnd)
{
	*selStart = (**macTE).selStart;
	*selEnd = (**macTE).selEnd;
}

/******************************************************************************
 GetNumLines
 
 	Return the number of lines of text.
******************************************************************************/

long CEditText::GetNumLines( void)
{
	long numLines = (**macTE).nLines;
	long teLen = (**macTE).teLength;

										/* TE line count is not correct if	*/
										/*   last char is a carriage return	*/
	if ((teLen > 0) && ((*(**macTE).hText)[ teLen - 1] == CARRIAGE_RETURN))
		numLines++;
		
	return numLines;
}

/******************************************************************************
 GetHeight
 
 	Return the height in pixels of the indicated lines of text.
******************************************************************************/

long CEditText::GetHeight( long startLine, long endLine)
{
	long height = TEGetHeight( endLine, startLine, macTE);
	return height;
}

/******************************************************************************
 GetCharOffset
 
 	Return the offset into the text buffer of the character corresponding
 	to the given point in Frame coordinates.
******************************************************************************/

long CEditText::GetCharOffset( LongPt *aPt)
{
	Point	qdPt;
	
	Prepare();
	FrameToQD( aPt, &qdPt);
	
	return TEGetOffset( qdPt, macTE);
	
}

/******************************************************************************
 GetCharPoint
 
 	Return the Frame coordinates of the character at the
 	given offset in the text buffer.
******************************************************************************/

void CEditText::GetCharPoint( long offset, LongPt *aPt)
{
	Point	qdPt;
	
	ASSERT( offset <= MAXINT);
	
	Prepare();
	qdPt = TEGetPoint( (short) offset, macTE);
	QDToFrame( qdPt, aPt);

}

/******************************************************************************
 CopyTextRange
 
 	Return a copy of the currently selected text.
******************************************************************************/

Handle CEditText::CopyTextRange( long start, long end)
{
	Handle	h;
	long	len;
	
	end = Min( end, (**macTE).teLength);
	
	len = Max( end - start, 0);
	h = NewHandleCanFail( len);
	FailNIL( h);
	if (len > 0)
		BlockMove( *(**macTE).hText + start, *h, len);
	
	return h;
	
}

/******************************************************************************
 PerformEditCommand
 
 	Performs the standard cut, copy, paste, and clear operations. Generally
 	called by a CTextEditTask to provide for undo.
 	
******************************************************************************/

void CEditText::PerformEditCommand( long theCommand)
{
	Prepare();

	switch( theCommand)
	{		
		case cmdCut:
			TECut(macTE);
			ZeroScrap();
			TEToScrap();
			gClipboard->UpdateDisplay();
			break;
			
		case cmdCopy:
			TECopy(macTE);
			ZeroScrap();
			TEToScrap();
			gClipboard->UpdateDisplay();			
			break;
			
		case cmdPaste:
			CheckInsertion( gClipboard->DataSize( 'TEXT'), TRUE);

			TEFromScrap();
			TEPaste(macTE);
			break;
			
		case cmdClear:
			TEDelete(macTE);
			break;
			
	}
	AdjustBounds();
	ScrollToSelection();
}

/******************************************************************************
 InsertTextPtr
 
 	Insert the given text just before the selection.
******************************************************************************/

void CEditText::InsertTextPtr( Ptr text, long length, Boolean fRedraw)
{
	CheckInsertion( length, FALSE);
	
	Prepare();
	TEInsert( text, length, macTE);
	AdjustBounds();
	if (fRedraw) Refresh();

}
